Package: mi
Type: Package
Title: Missing Data Imputation and Model Checking
Version: 1.2
Date: 2025-09-01
Authors@R: c(person("Andrew", "Gelman", email = "gelman@stat.columbia.edu", role = "ctb"), 
             person("Jennifer", "Hill", email = "jennifer.hill@nyu.edu", role = "ctb"),
             person("Yu-Sung", "Su", email = "suyusung@tsinghua.edu.cn", role = c("aut")), 
             person("Masanao", "Yajima", email = "my2167@columbia.edu", role = "ctb"), 
             person("Maria", "Pittau", email = "grazia@stat.columbia.edu", role = "ctb"), 
             person("Ben", "Goodrich", email = "benjamin.goodrich@columbia.edu", role = c("cre", "aut")), 
             person("Yajuan", "Si", email = "sophie2012@gmail.com", role = "ctb"),
             person("Jon", "Kropko", email = "jkropko@gmail.com", role = "aut"))
Description: The mi package provides functions for data manipulation, imputing missing values in an approximate Bayesian framework, diagnostics of the models used to generate the imputations, confidence-building mechanisms to validate some of the assumptions of the imputation algorithm, and functions to analyze multiply imputed data sets with the appropriate degree of sampling uncertainty.
VignetteBuilder: knitr
Depends: R (>= 3.0.0), methods, Matrix, stats4
Imports: arm (>= 1.4-11)
Suggests: betareg, lattice, knitr, MASS, nnet, parallel, sn, survival,
        truncnorm, foreign
URL: https://sites.stat.columbia.edu/gelman/
License: GPL (>= 2)
LazyLoad: yes
Author: Andrew Gelman [ctb],
  Jennifer Hill [ctb],
  Yu-Sung Su [aut],
  Masanao Yajima [ctb],
  Maria Pittau [ctb],
  Ben Goodrich [cre, aut],
  Yajuan Si [ctb],
  Jon Kropko [aut]
Maintainer: Ben Goodrich <benjamin.goodrich@columbia.edu>
NeedsCompilation: no
Packaged: 2025-09-01 20:01:42 UTC; ben
Repository: CRAN
Date/Publication: 2025-09-02 07:50:02 UTC
Built: R 4.6.0; ; 2026-04-26 11:28:29 UTC; unix

Package: mi
Type: Package
Title: Missing Data Imputation and Model Checking
Version: 1.3.1
Date: 2026-07-27
Authors@R: c(person("Andrew", "Gelman", email = "gelman@stat.columbia.edu", role = "ctb"), 
             person("Jennifer", "Hill", email = "jennifer.hill@nyu.edu", role = "ctb"),
             person("Yu-Sung", "Su", email = "suyusung@tsinghua.edu.cn", role = c("aut")), 
             person("Masanao", "Yajima", email = "my2167@columbia.edu", role = "ctb"), 
             person("Maria", "Pittau", email = "grazia@stat.columbia.edu", role = "ctb"), 
             person("Ben", "Goodrich", email = "benjamin.goodrich@columbia.edu", role = c("cre", "aut")), 
             person("Yajuan", "Si", email = "sophie2012@gmail.com", role = "ctb"),
             person("Jon", "Kropko", email = "jkropko@gmail.com", role = "aut"))
Description: The mi package provides functions for data manipulation, imputing missing values in an approximate Bayesian framework, diagnostics of the models used to generate the imputations, confidence-building mechanisms to validate some of the assumptions of the imputation algorithm, and functions to analyze multiply imputed data sets with the appropriate degree of sampling uncertainty.
VignetteBuilder: knitr
Depends: R (>= 3.0.0), methods, Matrix, stats4
Imports: arm (>= 1.4-11)
Suggests: betareg, lattice, knitr, MASS, nnet, parallel, sn, survival,
        truncnorm, foreign
URL: https://sites.stat.columbia.edu/gelman/
License: GPL (>= 2)
LazyLoad: yes
Author: Andrew Gelman [ctb],
  Jennifer Hill [ctb],
  Yu-Sung Su [aut],
  Masanao Yajima [ctb],
  Maria Pittau [ctb],
  Ben Goodrich [cre, aut],
  Yajuan Si [ctb],
  Jon Kropko [aut]
Maintainer: Ben Goodrich <benjamin.goodrich@columbia.edu>
NeedsCompilation: no
Packaged: 2026-07-28 02:55:47 UTC; bgoodri
Repository: CRAN
Date/Publication: 2026-07-28 03:30:02 UTC
Built: R 4.5.2; ; 2026-07-28 04:36:32 UTC; unix

#exportPattern("^[[:alpha:]]+")

importFrom(arm, display, fround, pfround, traceplot, bayesglm.fit, bayespolr)
importFrom(graphics, hist)
importFrom(Matrix, image, Matrix)
importFrom(stats4, coef, plot, summary, vcov)
import(methods)

importFrom("grDevices", "colorRampPalette", "grey", "heat.colors", "n2mfrow")
importFrom("graphics", "abline", "axis", "barplot", "lines", "par", "points", "segments", "title")
importFrom("stats", "as.dendrogram", "as.formula", "binomial", "cor",
           "cov", "cov2cor", "dbinom", "dist", "dlogis", "dnorm",
           "gaussian", "hclust", "is.ts", "lm", "lowess",
           "mahalanobis", "make.link", "median", "model.frame",
           "model.matrix", "model.response", "na.omit", "optim",
           "order.dendrogram", "pcauchy", "plogis", "pnorm", "poisson",
           "predict", "pt", "qlogis", "qnorm", "qpois", "quantile",
           "quasibinomial", "quasipoisson", "rbeta", "rbinom",
           "rchisq", "rgamma", "rmultinom", "rnbinom", "rnorm",
           "rpois", "rt", "runif", "sd", "uniroot", "var")
importFrom("utils", "head", "tail", "write.table")


exportClassPattern("^[[:alpha:]]+")
exportMethods(change, change_family, change_link, change_model, change_imputation_method, 
              change_size, change_transformation, change_type,
              complete, fit_model, get_parameters, hist, image, 
              mi, missing_data.frame, missing_variable, plot)
export(display, mi2BUGS, mi2stata, mipply, multinomial, pool, 
       .prune, .possible_missing_variable, Rhats, rdata.frame)

S3method(as.double, missing_variable)
S3method(as.double, categorical)
S3method(as.double, continuous)
S3method(as.double, count)
S3method(as.double, irrelevant)
S3method(as.double, missing_data.frame)
S3method(as.data.frame, missing_data.frame)
S3method(dim, missing_data.frame)
S3method(dimnames, missing_data.frame)
S3method(names, missing_data.frame)
S3method(dim, mi)
S3method(dimnames, mi)
S3method(names, mi)
S3method(is.na, missing_variable)
S3method(is.na, missing_data.frame)
S3method(is.na, mi)
S3method(length, missing_variable)
S3method(length, missing_data.frame)
S3method(length, mi)
S3method(print, mdf_list)
S3method(print, mi_list)
S3method("[", missing_data.frame)
S3method("[<-", missing_data.frame)
S3method("[[", missing_data.frame)
S3method("[[<-", missing_data.frame)
S3method("$", missing_data.frame)
S3method("$<-", missing_data.frame)

<h1 align = "center">:rocket: DataScience :facepunch:</h1>

---
- [mlcrate][https://github.com/mxbi/mlcrate]
<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/README_1200x800.gif">
</p>

<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/coversmall_alt.png">
  <br/>
</p>

# data-science-ipython-notebooks

## Index

* [deep-learning](#deep-learning)
    * [tensorflow](#tensor-flow-tutorials)
    * [theano](#theano-tutorials)
    * [keras](#keras-tutorials)
    * [caffe](#deep-learning-misc)
* [scikit-learn](#scikit-learn)
* [statistical-inference-scipy](#statistical-inference-scipy)
* [pandas](#pandas)
* [matplotlib](#matplotlib)
* [numpy](#numpy)
* [python-data](#python-data)
* [kaggle-and-business-analyses](#kaggle-and-business-analyses)
* [spark](#spark)
* [mapreduce-python](#mapreduce-python)
* [amazon web services](#aws)
* [command lines](#commands)
* [misc](#misc)
* [notebook-installation](#notebook-installation)
* [credits](#credits)
* [contributing](#contributing)
* [contact-info](#contact-info)
* [license](#license)

<br/>
<p align="center">
  <img src="http://i.imgur.com/ZhKXrKZ.png">
</p>

## deep-learning

IPython Notebook(s) demonstrating deep learning functionality.

<br/>
<p align="center">
  <img src="https://avatars0.githubusercontent.com/u/15658638?v=3&s=100">
</p>

### tensor-flow-tutorials

Additional TensorFlow tutorials:

* [pkmital/tensorflow_tutorials](https://github.com/pkmital/tensorflow_tutorials)
* [nlintz/TensorFlow-Tutorials](https://github.com/nlintz/TensorFlow-Tutorials)
* [alrojo/tensorflow-tutorial](https://github.com/alrojo/tensorflow-tutorial)
* [BinRoot/TensorFlow-Book](https://github.com/BinRoot/TensorFlow-Book)

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [tsf-basics](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/1_intro/basic_operations.ipynb) | Learn basic operations in TensorFlow, a library for various kinds of perceptual and language understanding tasks from Google. |
| [tsf-linear](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/2_basic_classifiers/linear_regression.ipynb) | Implement linear regression in TensorFlow. |
| [tsf-logistic](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/2_basic_classifiers/logistic_regression.ipynb) | Implement logistic regression in TensorFlow. |
| [tsf-nn](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/2_basic_classifiers/nearest_neighbor.ipynb) | Implement nearest neighboars in TensorFlow. |
| [tsf-alex](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/3_neural_networks/alexnet.ipynb) | Implement AlexNet in TensorFlow. |
| [tsf-cnn](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/3_neural_networks/convolutional_network.ipynb) | Implement convolutional neural networks in TensorFlow. |
| [tsf-mlp](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/3_neural_networks/multilayer_perceptron.ipynb) | Implement multilayer perceptrons in TensorFlow. |
| [tsf-rnn](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/3_neural_networks/recurrent_network.ipynb) | Implement recurrent neural networks in TensorFlow. |
| [tsf-gpu](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/4_multi_gpu/multigpu_basics.ipynb) | Learn about basic multi-GPU computation in TensorFlow. |
| [tsf-gviz](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/5_ui/graph_visualization.ipynb) | Learn about graph visualization in TensorFlow. |
| [tsf-lviz](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-examples/notebooks/5_ui/loss_visualization.ipynb) | Learn about loss visualization in TensorFlow. |

### tensor-flow-exercises

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [tsf-not-mnist](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-exercises/1_notmnist.ipynb) | Learn simple data curation by creating a pickle with formatted datasets for training, development and testing in TensorFlow. |
| [tsf-fully-connected](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-exercises/2_fullyconnected.ipynb) | Progressively train deeper and more accurate models using logistic regression and neural networks in TensorFlow. |
| [tsf-regularization](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-exercises/3_regularization.ipynb) | Explore regularization techniques by training fully connected networks to classify notMNIST characters in TensorFlow. |
| [tsf-convolutions](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-exercises/4_convolutions.ipynb) | Create convolutional neural networks in TensorFlow. |
| [tsf-word2vec](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-exercises/5_word2vec.ipynb) | Train a skip-gram model over Text8 data in TensorFlow. |
| [tsf-lstm](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/tensor-flow-exercises/6_lstm.ipynb) | Train a LSTM character model over Text8 data in TensorFlow. |

<br/>
<p align="center">
  <img src="http://www.deeplearning.net/software/theano/_static/theano_logo_allblue_200x46.png">
</p>

### theano-tutorials

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [theano-intro](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/theano-tutorial/intro_theano/intro_theano.ipynb) | Intro to Theano, which allows you to define, optimize, and evaluate mathematical expressions involving multi-dimensional arrays efficiently. It can use GPUs and perform efficient symbolic differentiation. |
| [theano-scan](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/theano-tutorial/scan_tutorial/scan_tutorial.ipynb) | Learn scans, a mechanism to perform loops in a Theano graph. |
| [theano-logistic](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/theano-tutorial/intro_theano/logistic_regression.ipynb) | Implement logistic regression in Theano. |
| [theano-rnn](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/theano-tutorial/rnn_tutorial/simple_rnn.ipynb) | Implement recurrent neural networks in Theano. |
| [theano-mlp](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/theano-tutorial/theano_mlp/theano_mlp.ipynb) | Implement multilayer perceptrons in Theano. |

<br/>
<p align="center">
  <img src="http://i.imgur.com/L45Q8c2.jpg">
</p>

### keras-tutorials

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| keras | Keras is an open source neural network library written in Python. It is capable of running on top of either Tensorflow or Theano. |
| [setup](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/0.%20Preamble.ipynb) | Learn about the tutorial goals and how to set up your Keras environment. |
| [intro-deep-learning-ann](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/1.1%20Introduction%20-%20Deep%20Learning%20and%20ANN.ipynb) | Get an intro to deep learning with Keras and Artificial Neural Networks (ANN). |
| [theano](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/1.2%20Introduction%20-%20Theano.ipynb) | Learn about Theano by working with weights matrices and gradients. |
| [keras-otto](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/1.3%20Introduction%20-%20Keras.ipynb) | Learn about Keras by looking at the Kaggle Otto challenge. |
| [ann-mnist](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/1.4%20%28Extra%29%20A%20Simple%20Implementation%20of%20ANN%20for%20MNIST.ipynb) | Review a simple implementation of ANN for MNIST using Keras. |
| [conv-nets](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/2.1%20Supervised%20Learning%20-%20ConvNets.ipynb) | Learn about Convolutional Neural Networks (CNNs) with Keras. |
| [conv-net-1](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/2.2.1%20Supervised%20Learning%20-%20ConvNet%20HandsOn%20Part%20I.ipynb) | Recognize handwritten digits from MNIST using Keras - Part 1. |
| [conv-net-2](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/2.2.2%20Supervised%20Learning%20-%20ConvNet%20HandsOn%20Part%20II.ipynb) | Recognize handwritten digits from MNIST using Keras - Part 2. |
| [keras-models](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/2.3%20Supervised%20Learning%20-%20Famous%20Models%20with%20Keras.ipynb) | Use pre-trained models such as VGG16, VGG19, ResNet50, and Inception v3 with Keras. |
| [auto-encoders](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/3.1%20Unsupervised%20Learning%20-%20AutoEncoders%20and%20Embeddings.ipynb) | Learn about Autoencoders with Keras. |
| [rnn-lstm](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/3.2%20RNN%20and%20LSTM.ipynb) | Learn about Recurrent Neural Networks (RNNs) with Keras. |
| [lstm-sentence-gen](https://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/keras-tutorial/3.3%20%28Extra%29%20LSTM%20for%20Sentence%20Generation.ipynb) |  Learn about RNNs using Long Short Term Memory (LSTM) networks with Keras. |

### deep-learning-misc

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [deep-dream](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/deep-learning/deep-dream/dream.ipynb) | Caffe-based computer vision program which uses a convolutional neural network to find and enhance patterns in images. |

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/scikitlearn.png">
</p>

## scikit-learn

IPython Notebook(s) demonstrating scikit-learn functionality.

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [intro](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scikit-learn/scikit-learn-intro.ipynb) | Intro notebook to scikit-learn.  Scikit-learn adds Python support for large, multi-dimensional arrays and matrices, along with a large library of high-level mathematical functions to operate on these arrays. |
| [knn](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scikit-learn/scikit-learn-intro.ipynb#K-Nearest-Neighbors-Classifier) | Implement k-nearest neighbors in scikit-learn. |
| [linear-reg](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scikit-learn/scikit-learn-linear-reg.ipynb) | Implement linear regression in scikit-learn. |
| [svm](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scikit-learn/scikit-learn-svm.ipynb) | Implement support vector machine classifiers with and without kernels in scikit-learn. |
| [random-forest](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scikit-learn/scikit-learn-random-forest.ipynb) | Implement random forest classifiers and regressors in scikit-learn. |
| [k-means](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scikit-learn/scikit-learn-k-means.ipynb) | Implement k-means clustering in scikit-learn. |
| [pca](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scikit-learn/scikit-learn-pca.ipynb) | Implement principal component analysis in scikit-learn. |
| [gmm](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scikit-learn/scikit-learn-gmm.ipynb) | Implement Gaussian mixture models in scikit-learn. |
| [validation](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scikit-learn/scikit-learn-validation.ipynb) | Implement validation and model selection in scikit-learn. |

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/scipy.png">
</p>

## statistical-inference-scipy

IPython Notebook(s) demonstrating statistical inference with SciPy functionality.

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| scipy | SciPy is a collection of mathematical algorithms and convenience functions built on the Numpy extension of Python. It adds significant power to the interactive Python session by providing the user with high-level commands and classes for manipulating and visualizing data. |
| [effect-size](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scipy/effect_size.ipynb) | Explore statistics that quantify effect size by analyzing the difference in height between men and women.  Uses data from the Behavioral Risk Factor Surveillance System (BRFSS) to estimate the mean and standard deviation of height for adult women and men in the United States. |
| [sampling](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scipy/sampling.ipynb) | Explore random sampling by analyzing the average weight of men and women in the United States using BRFSS data. |
| [hypothesis](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/scipy/hypothesis.ipynb) | Explore hypothesis testing by analyzing the difference of first-born babies compared with others. |

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/pandas.png">
</p>

## pandas

IPython Notebook(s) demonstrating pandas functionality.

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------------------------|-----------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [pandas](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/pandas.ipynb) | Software library written for data manipulation and analysis in Python. Offers data structures and operations for manipulating numerical tables and time series. |
| [github-data-wrangling](https://github.com/donnemartin/viz/blob/master/githubstats/data_wrangling.ipynb) | Learn how to load, clean, merge, and feature engineer by analyzing GitHub data from the [`Viz`](https://github.com/donnemartin/viz) repo. |
| [Introduction-to-Pandas](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.00-Introduction-to-Pandas.ipynb) | Introduction to Pandas. |
| [Introducing-Pandas-Objects](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.01-Introducing-Pandas-Objects.ipynb) | Learn about Pandas objects. |
| [Data Indexing and Selection](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.02-Data-Indexing-and-Selection.ipynb) | Learn about data indexing and selection in Pandas. |
| [Operations-in-Pandas](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.03-Operations-in-Pandas.ipynb) | Learn about operating on data in Pandas. |
| [Missing-Values](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.04-Missing-Values.ipynb) | Learn about handling missing data in Pandas. |
| [Hierarchical-Indexing](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.05-Hierarchical-Indexing.ipynb) | Learn about hierarchical indexing in Pandas. |
| [Concat-And-Append](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.06-Concat-And-Append.ipynb) | Learn about combining datasets: concat and append in Pandas. |
| [Merge-and-Join](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.07-Merge-and-Join.ipynb) | Learn about combining datasets: merge and join in Pandas. |
| [Aggregation-and-Grouping](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.08-Aggregation-and-Grouping.ipynb) | Learn about aggregation and grouping in Pandas. |
| [Pivot-Tables](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.09-Pivot-Tables.ipynb) | Learn about pivot tables in Pandas. |
| [Working-With-Strings](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.10-Working-With-Strings.ipynb) | Learn about vectorized string operations in Pandas. |
| [Working-with-Time-Series](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.11-Working-with-Time-Series.ipynb) | Learn about working with time series in pandas. |
| [Performance-Eval-and-Query](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/pandas/03.12-Performance-Eval-and-Query.ipynb) | Learn about high-performance Pandas: eval() and query() in Pandas. |

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/matplotlib.png">
</p>

## matplotlib

IPython Notebook(s) demonstrating matplotlib functionality.

| Notebook | Description |
|-----------------------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------|
| [matplotlib](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/matplotlib.ipynb) | Python 2D plotting library which produces publication quality figures in a variety of hardcopy formats and interactive environments across platforms. |
| [matplotlib-applied](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/matplotlib-applied.ipynb) | Apply matplotlib visualizations to Kaggle competitions for exploratory data analysis.  Learn how to create bar plots, histograms, subplot2grid, normalized plots, scatter plots, subplots, and kernel density estimation plots. |
| [Introduction-To-Matplotlib](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.00-Introduction-To-Matplotlib.ipynb) | Introduction to Matplotlib. |
| [Simple-Line-Plots](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.01-Simple-Line-Plots.ipynb) | Learn about simple line plots in Matplotlib. |
| [Simple-Scatter-Plots](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.02-Simple-Scatter-Plots.ipynb) | Learn about simple scatter plots in Matplotlib. |
| [Errorbars.ipynb](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.03-Errorbars.ipynb) | Learn about visualizing errors in Matplotlib. |
| [Density-and-Contour-Plots](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.04-Density-and-Contour-Plots.ipynb) | Learn about density and contour plots in Matplotlib. |
| [Histograms-and-Binnings](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.05-Histograms-and-Binnings.ipynb) | Learn about histograms, binnings, and density in Matplotlib. |
| [Customizing-Legends](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.06-Customizing-Legends.ipynb) | Learn about customizing plot legends in Matplotlib. |
| [Customizing-Colorbars](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.07-Customizing-Colorbars.ipynb) | Learn about customizing colorbars in Matplotlib. |
| [Multiple-Subplots](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.08-Multiple-Subplots.ipynb) | Learn about multiple subplots in Matplotlib. |
| [Text-and-Annotation](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.09-Text-and-Annotation.ipynb) | Learn about text and annotation in Matplotlib. |
| [Customizing-Ticks](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.10-Customizing-Ticks.ipynb) | Learn about customizing ticks in Matplotlib. |
| [Settings-and-Stylesheets](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.11-Settings-and-Stylesheets.ipynb) | Learn about customizing Matplotlib: configurations and stylesheets. |
| [Three-Dimensional-Plotting](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.12-Three-Dimensional-Plotting.ipynb) | Learn about three-dimensional plotting in Matplotlib. |
| [Geographic-Data-With-Basemap](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.13-Geographic-Data-With-Basemap.ipynb) | Learn about geographic data with basemap in Matplotlib. |
| [Visualization-With-Seaborn](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/matplotlib/04.14-Visualization-With-Seaborn.ipynb) | Learn about visualization with Seaborn. |

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/numpy.png">
</p>

## numpy

IPython Notebook(s) demonstrating NumPy functionality.

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [numpy](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/numpy.ipynb) | Adds Python support for large, multi-dimensional arrays and matrices, along with a large library of high-level mathematical functions to operate on these arrays. |
| [Introduction-to-NumPy](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.00-Introduction-to-NumPy.ipynb) | Introduction to NumPy. |
| [Understanding-Data-Types](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.01-Understanding-Data-Types.ipynb) | Learn about data types in Python. |
| [The-Basics-Of-NumPy-Arrays](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.02-The-Basics-Of-NumPy-Arrays.ipynb) | Learn about the basics of NumPy arrays. |
| [Computation-on-arrays-ufuncs](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.03-Computation-on-arrays-ufuncs.ipynb) | Learn about computations on NumPy arrays: universal functions. |
| [Computation-on-arrays-aggregates](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.04-Computation-on-arrays-aggregates.ipynb) | Learn about aggregations: min, max, and everything in between in NumPy. |
| [Computation-on-arrays-broadcasting](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.05-Computation-on-arrays-broadcasting.ipynb) | Learn about computation on arrays: broadcasting in NumPy. |
| [Boolean-Arrays-and-Masks](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.06-Boolean-Arrays-and-Masks.ipynb) | Learn about comparisons, masks, and boolean logic in NumPy. |
| [Fancy-Indexing](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.07-Fancy-Indexing.ipynb) | Learn about fancy indexing in NumPy. |
| [Sorting](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.08-Sorting.ipynb) | Learn about sorting arrays in NumPy. |
| [Structured-Data-NumPy](http://nbviewer.jupyter.org/github/donnemartin/data-science-ipython-notebooks/blob/master/numpy/02.09-Structured-Data-NumPy.ipynb) | Learn about structured data: NumPy's structured arrays. |

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/python.png">
</p>

## python-data

IPython Notebook(s) demonstrating Python functionality geared towards data analysis.

| Notebook | Description |
|-----------------------------------------------------------------------------------------------------------------------------------------------|---------------------------------------------------------------------------------------------------------------------------|
| [data structures](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/python-data/structs.ipynb) | Learn Python basics with tuples, lists, dicts, sets. |
| [data structure utilities](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/python-data/structs_utils.ipynb) | Learn Python operations such as slice, range, xrange, bisect, sort, sorted, reversed, enumerate, zip, list comprehensions. |
| [functions](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/python-data/functions.ipynb) | Learn about more advanced Python features: Functions as objects, lambda functions, closures, *args, **kwargs currying, generators, generator expressions, itertools. |
| [datetime](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/python-data/datetime.ipynb) | Learn how to work with Python dates and times: datetime, strftime, strptime, timedelta. |
| [logging](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/python-data/logs.ipynb) | Learn about Python logging with RotatingFileHandler and TimedRotatingFileHandler. |
| [pdb](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/python-data/pdb.ipynb) | Learn how to debug in Python with the interactive source code debugger. |
| [unit tests](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/python-data/unit_tests.ipynb) | Learn how to test in Python with Nose unit tests. |

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/kaggle.png">
</p>

## kaggle-and-business-analyses

IPython Notebook(s) used in [kaggle](https://www.kaggle.com/) competitions and business analyses.

| Notebook | Description |
|-------------------------------------------------------------------------------------------------------------------|----------------------------------------------------------------------------------------------------------------|
| [titanic](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/kaggle/titanic.ipynb) | Predict survival on the Titanic.  Learn data cleaning, exploratory data analysis, and machine learning. |
| [churn-analysis](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/analyses/churn.ipynb) | Predict customer churn.  Exercise logistic regression, gradient boosting classifers, support vector machines, random forests, and k-nearest-neighbors.  Includes discussions of confusion matrices, ROC plots, feature importances, prediction probabilities, and calibration/descrimination.|

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/spark.png">
</p>

## spark

IPython Notebook(s) demonstrating spark and HDFS functionality.

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|--------------------------------------------------------------------------------------------------------------------------------------------|
| [spark](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/spark/spark.ipynb) | In-memory cluster computing framework, up to 100 times faster for certain applications and is well suited for machine learning algorithms. |
| [hdfs](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/spark/hdfs.ipynb) | Reliably stores very large files across machines in a large cluster. |

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/mrjob.png">
</p>

## mapreduce-python

IPython Notebook(s) demonstrating Hadoop MapReduce with mrjob functionality.

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------|--------------------------------------------------------------------------------------------------------------------------------------------|
| [mapreduce-python](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/mapreduce/mapreduce-python.ipynb) | Runs MapReduce jobs in Python, executing jobs locally or on Hadoop clusters. Demonstrates Hadoop Streaming in Python code with unit test and [mrjob](https://github.com/Yelp/mrjob) config file to analyze Amazon S3 bucket logs on Elastic MapReduce.  [Disco](https://github.com/discoproject/disco/) is another python-based alternative.|

<br/>

<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/aws.png">
</p>

## aws

IPython Notebook(s) demonstrating Amazon Web Services (AWS) and AWS tools functionality.


Also check out:

* [SAWS](https://github.com/donnemartin/saws): A Supercharged AWS command line interface (CLI).
* [Awesome AWS](https://github.com/donnemartin/awesome-aws): A curated list of libraries, open source repos, guides, blogs, and other resources.

| Notebook | Description |
|------------------------------------------------------------------------------------------------------------------------------------|-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [boto](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/aws/aws.ipynb#Boto) | Official AWS SDK for Python. |
| [s3cmd](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/aws/aws.ipynb#s3cmd) | Interacts with S3 through the command line. |
| [s3distcp](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/aws/aws.ipynb#s3distcp) | Combines smaller files and aggregates them together by taking in a pattern and target file.  S3DistCp can also be used to transfer large volumes of data from S3 to your Hadoop cluster. |
| [s3-parallel-put](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/aws/aws.ipynb#s3-parallel-put) | Uploads multiple files to S3 in parallel. |
| [redshift](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/aws/aws.ipynb#redshift) | Acts as a fast data warehouse built on top of technology from massive parallel processing (MPP). |
| [kinesis](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/aws/aws.ipynb#kinesis) | Streams data in real time with the ability to process thousands of data streams per second. |
| [lambda](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/aws/aws.ipynb#lambda) | Runs code in response to events, automatically managing compute resources. |

<br/>
<p align="center">
  <img src="https://raw.githubusercontent.com/donnemartin/data-science-ipython-notebooks/master/images/commands.png">
</p>

## commands

IPython Notebook(s) demonstrating various command lines for Linux, Git, etc.

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------------------------------------|-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [linux](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/commands/linux.ipynb) | Unix-like and mostly POSIX-compliant computer operating system.  Disk usage, splitting files, grep, sed, curl, viewing running processes, terminal syntax highlighting, and Vim.|
| [anaconda](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/commands/misc.ipynb#anaconda) | Distribution of the Python programming language for large-scale data processing, predictive analytics, and scientific computing, that aims to simplify package management and deployment. |
| [ipython notebook](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/commands/misc.ipynb#ipython-notebook) | Web-based interactive computational environment where you can combine code execution, text, mathematics, plots and rich media into a single document. |
| [git](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/commands/misc.ipynb#git) | Distributed revision control system with an emphasis on speed, data integrity, and support for distributed, non-linear workflows. |
| [ruby](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/commands/misc.ipynb#ruby) | Used to interact with the AWS command line and for Jekyll, a blog framework that can be hosted on GitHub Pages. |
| [jekyll](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/commands/misc.ipynb#jekyll) | Simple, blog-aware, static site generator for personal, project, or organization sites.  Renders Markdown or Textile and Liquid templates, and produces a complete, static website ready to be served by Apache HTTP Server, Nginx or another web server. |
| [pelican](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/commands/misc.ipynb#pelican) | Python-based alternative to Jekyll. |
| [django](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/commands/misc.ipynb#django) | High-level Python Web framework that encourages rapid development and clean, pragmatic design. It can be useful to share reports/analyses and for blogging. Lighter-weight alternatives include [Pyramid](https://github.com/Pylons/pyramid), [Flask](https://github.com/pallets/flask), [Tornado](https://github.com/tornadoweb/tornado), and [Bottle](https://github.com/bottlepy/bottle).

## misc

IPython Notebook(s) demonstrating miscellaneous functionality.

| Notebook | Description |
|--------------------------------------------------------------------------------------------------------------------------------------------|-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [regex](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/misc/regex.ipynb) | Regular expression cheat sheet useful in data wrangling.|
[algorithmia](http://nbviewer.ipython.org/github/donnemartin/data-science-ipython-notebooks/blob/master/misc/Algorithmia.ipynb) | Algorithmia is a marketplace for algorithms. This notebook showcases 4 different algorithms: Face Detection, Content Summarizer, Latent Dirichlet Allocation and Optical Character Recognition.|

## notebook-installation

### anaconda

Anaconda is a free distribution of the Python programming language for large-scale data processing, predictive analytics, and scientific computing that aims to simplify package management and deployment.

Follow instructions to install [Anaconda](https://docs.continuum.io/anaconda/install) or the more lightweight [miniconda](http://conda.pydata.org/miniconda.html).

### dev-setup

For detailed instructions, scripts, and tools to set up your development environment for data analysis, check out the [dev-setup](https://github.com/donnemartin/dev-setup) repo.

### running-notebooks

To view interactive content or to modify elements within the IPython notebooks, you must first clone or download the repository then run the notebook.  More information on IPython Notebooks can be found [here.](http://ipython.org/notebook.html)

    $ git clone https://github.com/donnemartin/data-science-ipython-notebooks.git
    $ cd data-science-ipython-notebooks
    $ jupyter notebook

Notebooks tested with Python 2.7.x.

## credits

* [Python for Data Analysis: Data Wrangling with Pandas, NumPy, and IPython](http://www.amazon.com/Python-Data-Analysis-Wrangling-IPython/dp/1449319793) by Wes McKinney
* [PyCon 2015 Scikit-learn Tutorial](https://github.com/jakevdp/sklearn_pycon2015) by Jake VanderPlas
* [Python Data Science Handbook](https://github.com/jakevdp/PythonDataScienceHandbook) by Jake VanderPlas
* [Parallel Machine Learning with scikit-learn and IPython](https://github.com/ogrisel/parallel_ml_tutorial) by Olivier Grisel
* [Statistical Interference Using Computational Methods in Python](https://github.com/AllenDowney/CompStats) by Allen Downey
* [TensorFlow Examples](https://github.com/aymericdamien/TensorFlow-Examples) by Aymeric Damien
* [TensorFlow Tutorials](https://github.com/pkmital/tensorflow_tutorials) by Parag K Mital
* [TensorFlow Tutorials](https://github.com/nlintz/TensorFlow-Tutorials) by Nathan Lintz
* [TensorFlow Tutorials](https://github.com/alrojo/tensorflow-tutorial) by Alexander R Johansen
* [TensorFlow Book](https://github.com/BinRoot/TensorFlow-Book) by Nishant Shukla
* [Summer School 2015](https://github.com/mila-udem/summerschool2015) by mila-udem
* [Keras tutorials](https://github.com/leriomaggio/deep-learning-keras-tensorflow) by Valerio Maggio
* [Kaggle](https://www.kaggle.com/)
* [Yhat Blog](http://blog.yhat.com/)

## contributing

Contributions are welcome!  For bug reports or requests please [submit an issue](https://github.com/donnemartin/data-science-ipython-notebooks/issues).

## contact-info

Feel free to contact me to discuss any issues, questions, or comments.

* Email: [donne.martin@gmail.com](mailto:donne.martin@gmail.com)
* Twitter: [@donne_martin](https://twitter.com/donne_martin)
* GitHub: [donnemartin](https://github.com/donnemartin)
* LinkedIn: [donnemartin](https://www.linkedin.com/in/donnemartin)
* Website: [donnemartin.com](http://donnemartin.com)

## license

This repository contains a variety of content; some developed by Donne Martin, and some from third-parties.  The third-party content is distributed under the license provided by those parties.

The content developed by Donne Martin is distributed under the following license:

*I am providing code and resources in this repository to you under an open source license.  Because this is my personal repository, the license you receive to my code and resources is from me and not my employer (Facebook).*

    Copyright 2015 Donne Martin

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


pool <-
  function(formula, data, m = NULL, FUN = NULL, ...) {
    if(is.list(data)) {
      if(all(sapply(data, is, class2 = "mi"))) {
        if(is.null(m)) m <- length(data[[1]]@data)
        else m <- as.integer(m)
        dfs <- complete(data[[1]], m = m, to_matrix = FALSE)
        l <- length(data)
        if(l > 1) for(i in 2:l) {
          temp <- complete(data[[l]], m = m, to_matrix = FALSE)
          for(j in seq_along(temp)) dfs[[j]] <- rbind(dfs[[j]], temp[[j]])
        }
        data <- data[[1]]
      }
      else if(all(sapply(data, is.data.frame))) {
        dfs <- data
        m <- length(dfs) 
      }
      else {
        stop("if 'data' is a list it must be a list of mi objects or data.frames")
      }
    }
    else if(is(data, "mi")) {
      if(is.null(m)) m <- length(data@data)
      else m <- as.integer(m)
      dfs <- complete(data, m = m, to_matrix = FALSE)
    }
    if(!is(formula, "formula")) stop("'formula' must be a formula")
    
    dots <- list(...)
    
    if(is.null(FUN)) {
      if(!is(data, "mi")) stop("if 'data' is not of class 'mi', 'FUN' must be specified")
      yname <- as.character(formula)[2]
      if(!(yname %in% colnames(data@data[[1]]))) {
        stop(paste("no variable called", yname, "possibly due to typo or transformation,",
                   "in which case you need to specify 'FUN' explicitly"))
      }
      else y <- data@data[[1]]@variables[[yname]]
      
      if(!is.method_in_mi("fit_model", y = class(y), data = class(data@data[[1]]))) {
        stop(paste(yname, "seems to have a user-defined 'fit_model' method,",
                   "in which case 'FUN' must be specified explicitly"))
      }
      
      if(is(y, "unordered-categorical")) {
        FUN <- nnet::multinom
        fit <- "multinom"
      }
      else if(is(y, "binary") | is(y, "count") | is(y, "continuous")) {
        FUN <- arm::bayesglm
        fit <- "bayesglm"
        if(!("family" %in% names(dots))) dots$family <- y@family
      }
      else if(is(y, "interval")) {
        FUN <- survival::survreg
        fit <- "survreg"
      }
      else if(is(y, "ordered-categorical")) {
        FUN <- arm::bayespolr
        fit <- "bayespolr"
        if(!("method" %in% names(dots))) dots$method <- if(y@family$link == "logit") "logistic" else y@family$link
      }
    }
    else if(!is(FUN, "function")) stop("'FUN' must be a function or NULL")
    else fit <- deparse(substitute(FUN))
    
    models <- lapply(dfs, FUN = function(d) {
      dots$data <- d
      dots$formula <- formula
      do.call(FUN, args = dots)
    })
    
    summaries <- lapply(models, summary)
    pooled_summary <- summaries[[1]]
    if(is.list(pooled_summary)) for(i in seq_along(pooled_summary)) {
      if(is.numeric(pooled_summary[[i]])) {
        num <- lapply(summaries, FUN = function(x) x[[i]])
        if(is.matrix(pooled_summary[[i]])) {
          mat <- pooled_summary[[i]]
          arr <- array(unlist(num), dim = c(dim(mat), m))
          arr <- apply(arr, 1:2, mean)
          colnames(arr) <- colnames(mat)
          rownames(arr) <- rownames(mat)
          pooled_summary[[i]] <- arr
        }
        else if(length(pooled_summary[[i]]) > 1) {
          arr <- rowMeans(matrix(unlist(num), ncol = m))
          names(arr) <- names(pooled_summary[[i]])
          pooled_summary[[i]] <- arr
        }
        else pooled_summary[[i]] <- mean(unlist(num))
      }
    }
    else {
      pooled_summary <- list()
      warning("could not construct pooled_summary")
    }
    
    coefs <- sapply(models, get_parameters)
    variances <- sapply(models, FUN = function(x) diag(vcov(x)))
    W <- rowMeans(variances)
    B <- apply(coefs, 1, var) 
    ses <- sqrt(W  + B * (1 + 1/m))
    
    if(is(pooled_summary, "summary.glm") | is(pooled_summary, "summary.polr")) {
      pooled_summary$call <- match.call()
      pooled_summary$coefficients[,1:2] <- cbind(rowMeans(coefs), ses)
    }
    else if(is(pooled_summary, "summary.multinom")) {
      pooled_summary$call <- match.call()
      pooled_summary$coefficients <- cbind(coef = rowMeans(coefs), ses, z = NA_real_, p = NA_real_)
    }
    else warning("pooled_summary is probably bogus")
    
    if(ncol(pooled_summary$coefficients) >= 3) {
      if(colnames(pooled_summary$coefficients)[3] == "t value") {
        pooled_summary$coefficients[,3] <- tvalue <- pooled_summary$coefficients[,1] / ses 
        if(TRUE) {
          gamma <- (1 + 1/m) * B / ses^2
          df.r <- pooled_summary$df.residual
          v <- (m - 1) * (1 + m/(m + 1) * W / B)^2
          v_obs <- (1 - gamma) * (df.r + 1) / (df.r + 3) * df.r
          df.star <- 1/(1/v + 1/v_obs)
          if(ncol(pooled_summary$coefficients) == 4) {
            pooled_summary$coefficients[,4] <- 2 * pt(-abs(tvalue), df.star)
          }
          else pooled_summary$coefficients <- cbind(pooled_summary$coefficients, 
                                                    "p-value" = 2 * pt(-abs(tvalue), df.star))
        }
      }
      else {
        pooled_summary$coefficients[,3] <- zvalue <- pooled_summary$coefficients[,1] / ses
        if(ncol(pooled_summary$coefficients) == 4) {
          pooled_summary$coefficients[,4] <- 2 * pnorm(-abs(zvalue))
        }
        else pooled_summary$coefficients <- cbind(pooled_summary$coefficients,
                                                  "p-value" = 2 * pnorm(-abs(zvalue)) )
      }
    }
    kall <- match.call()
    kall[1] <- call(fit)
    out <- new("pooled", formula = formula, fit = fit, models = models,
               coefficients = rowMeans(coefs), ses = ses, pooled_summary = pooled_summary, call = kall)
    return(out)
  }

setMethod("display", signature(object = "pooled"), def =
  function(object, digits = 2, ...) {
    call <- object@call
    summ <- summary(object)
    coef <- object@pooled_summary$coefficients[,1:2]
    colnames(coef) <- c("coef.est", "coef.se")
    n <- summ$df.residual
    k <- summ$df[1]
    k.intercepts <- length(summ$zeta)
    
    print(call)
    pfround(coef, digits)
    if(k.intercepts > 0) {
      cat(paste("n = ", n, ", k = ", k, " (including ", k.intercepts, 
                " intercepts)\nresidual deviance = ", fround(summ$deviance, 1),
                ", null deviance is not computed by polr", "\n", sep = ""))
      return(invisible(NULL))
    }
    cat(paste("n = ", n, ", k = ", k, "\nresidual deviance = ", 
              fround(summ$deviance, 1), ", null deviance = ", fround(summ$null.deviance, 1), 
              " (difference = ", fround(summ$null.deviance - summ$deviance, 1), ")", "\n", sep = ""))
    dispersion <- summ$dispersion 
    if (dispersion != 1) {
      cat(paste("overdispersion parameter = ", fround(dispersion, 1), "\n", sep = ""))
      if (summ$family$family == "gaussian") {
        cat(paste("residual sd is sqrt(overdispersion) = ", 
                  fround(sqrt(dispersion), digits), "\n", sep = ""))
      }
    }
    return(invisible(NULL))
  })

setMethod("show", signature(object = "pooled"), def =
  function(object) {
    display(object)
    return(invisible(NULL))
  })

setMethod("summary", signature(object = "pooled"), def = 
  function(object, ...) {
    return(object@pooled_summary)
  })

setMethod("coef", signature(object = "pooled"), def = 
  function(object, ...) {
    return(object@coefficients)
  })

setMethod("vcov", signature(object = "pooled"), def = 
  function(object, ...) {
    return(object@vcov)
  })

setMethod("residuals", signature(object = "pooled"), def = 
  function(object, ...) {
    return(rowMeans(sapply(object@models, residuals)))
  })

setMethod("fitted", signature(object = "pooled"), def = 
  function(object, ...) {
    return(rowMeans(sapply(object@models, fitted, ...)))
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


setMethod("change_size", signature(data = "missing", y = "missing_variable", to = "integer"), def = 
  function(y, to) {
    n <- to
    if(n <= 0) {
      y@data <- y@data[-y@which_extra]
      y@which_extra <- integer(0)
      y@n_total <- y@n_total - y@n_extra
      y@n_extra <- NA_integer_
      return(y)
    }
    end <- y@n_total
    SEQ <- (end+1):(end+n)
    y@data <- c(y@data, rep(NA, n))
    y@which_extra <- c(y@which_extra, SEQ)
    y@n_extra <- y@n_extra + n
    y@n_total <- y@n_total + n
    return(y)
  })

setMethod("change_size", signature(data = "missing", y = "categorical", to = "integer"), def = 
  function(y, to) {
    n <- to
    if(n <= 0) {
      y@data <- y@data[-y@which_extra]
      y@which_extra <- integer(0)
      y@n_total <- y@n_total - y@n_extra
      y@n_extra <- NA_integer_
      return(y)
    }
    end <- y@n_total
    SEQ <- (end+1):(end+n)
    y@data <- c(y@data, rep(NA, n))
    y@which_extra <- c(y@which_extra, SEQ)
    y@n_extra <- y@n_extra + n
    y@n_total <- y@n_total + n
    return(y)
  })

setMethod("change_size", signature(data = "missing", y = "fixed", to = "integer"), def = 
  function(y, to) {
    n <- to
    if(n <= 0) {
      y@data <- y@data[-y@which_extra]
      y@which_extra <- integer(0)
      y@n_total <- y@n_total - y@n_extra
      y@n_extra <- NA_integer_
      return(y)
    }
    end <- y@n_total
    SEQ <- (end+1):(end+n)
    y@data <- c(y@data, rep(y@data[1], n))
    y@which_extra <- c(y@which_extra, SEQ)
    y@n_extra <- y@n_extra + n
    y@n_total <- y@n_total + n
    return(y)
  })

setMethod("change_size", signature(data = "missing_data.frame", y = "missing", to = "integer"), def = 
  function(data, to) {
    n <- to
    data@variables <- lapply(data@variables, FUN = function(x) change_size(x, n))
    data@DIM[1] <- data@variables[[1]]@n_total
    return(data)
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

.onLoad <- function(lib, pkg) {
  #   library.dynam("mi", pkg, lib)
  return(invisible(NULL))
}

.onUnload <- function(libpath) {
  #   library.dynam.unload("mi", libpath)
  return(invisible(NULL))
}

.onAttach <- function( ... ) {
  miLib <- dirname(system.file(package = "mi"))
  version <- utils::packageDescription("mi", lib.loc = miLib)$Version
  builddate <- utils::packageDescription("mi", lib.loc = miLib)$Packaged
  packageStartupMessage(paste("mi (Version ", version, ", packaged: ", builddate, ")", sep = ""))
  packageStartupMessage("mi  Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University")
  packageStartupMessage("This program comes with ABSOLUTELY NO WARRANTY.")
  packageStartupMessage("This is free software, and you are welcome to redistribute it")
  packageStartupMessage("under the General Public License version 2 or later.")
  packageStartupMessage("Execute RShowDoc('COPYING') for details.")
}

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


setMethod("hist", signature(x = "missing_variable"), def = 
  function(x, ...) {
    y <- x@data
    NAs <- is.na(x)
    
    h_all  <- hist(y, plot = FALSE)
    plot(h_all, border = "lightgray", main = "", xlab = if(x@done) "Completed" else "Observed", axes = FALSE,
         mgp = c(2, 1, 0), tcl = .05, col = if(x@done) "lightgray" else "blue", freq = TRUE, ...)
    axis(1, lwd = 0)
    axis(2)
    if(x@done) {
      h_obs  <- hist(y[!NAs], breaks = h_all$breaks, plot = FALSE)
      h_miss <- hist(y[NAs], breaks = h_all$breaks, plot = FALSE)
      
      segments(h_obs$breaks[1], 0,  y1 = h_obs$counts[1], col = "blue")
      segments(h_miss$breaks[1], 0,  y1 = h_miss$counts[1], col = "red")
      
      segments(h_obs$breaks[1], y0 = h_obs$counts[1], x1  = h_obs$breaks[2], col = "blue")
      segments(h_miss$breaks[1], y0 = h_miss$counts[1], x1 = h_miss$breaks[2], col = "red")
      
      for(i in 2:(length(h_obs$breaks)-1)) {
        segments(x0 = h_obs$breaks[i], y0 = h_obs$counts[i-1], y1 = h_obs$counts[i], col = "blue")
        segments(x0 = h_miss$breaks[i], y0 = h_miss$counts[i-1], y1 = h_miss$counts[i], col = "red")
        
        segments(x0 = h_obs$breaks[i], y0 = h_obs$counts[i], x1 = h_obs$breaks[i+1], col = "blue")
        segments(x0 = h_miss$breaks[i], y0 = h_miss$counts[i], x1 = h_miss$breaks[i+1], col = "red")
      }
      
      segments(x0 = h_obs$breaks[i+1], y0 = h_obs$counts[i], y1 = 0, col = "blue")
      segments(x0 = h_miss$breaks[i+1], y0 = h_miss$counts[i], y1 = 0, col = "blue")
      
      if(.MI_DEBUG) stopifnot(all(h_all$counts == (h_obs$counts + h_miss$counts)))
    }
    return(invisible(NULL))
  })

setMethod("hist", signature(x = "semi-continuous"), def = 
  function(x, ...) {
    con <- complete(x@indicator, 0L) == 0
    y <- x@data[con]
    NAs <- is.na(x)[con]
    
    h_all  <- hist(y, plot = FALSE)
    plot(h_all, freq = TRUE, border = "lightgray", main = "", xlab = if(x@done) "Completed" else "Observed", axes = FALSE,
         mgp = c(2, 1, 0), tcl = .05, col = if(x@done) "lightgray" else "blue", xlim = range(x@data, na.rm = TRUE), ...)
    axis(1, lwd = 0)
    axis(2)
    if(x@done) {
      h_obs  <- hist(y[!NAs], breaks = h_all$breaks, plot = FALSE)
      h_miss <- hist(y[NAs], breaks = h_all$breaks, plot = FALSE)
      
      segments(h_obs$breaks[1], 0,  y1 = h_obs$counts[1], col = "blue")
      segments(h_miss$breaks[1], 0,  y1 = h_miss$counts[1], col = "red")
      
      segments(h_obs$breaks[1], y0 = h_obs$counts[1], x1  = h_obs$breaks[2], col = "blue")
      segments(h_miss$breaks[1], y0 = h_miss$counts[1], x1 = h_miss$breaks[2], col = "red")
      
      for(i in 2:(length(h_obs$breaks)-1)) {
        segments(x0 = h_obs$breaks[i], y0 = h_obs$counts[i-1], y1 = h_obs$counts[i], col = "blue")
        segments(x0 = h_miss$breaks[i], y0 = h_miss$counts[i-1], y1 = h_miss$counts[i], col = "red")
        
        segments(x0 = h_obs$breaks[i], y0 = h_obs$counts[i], x1 = h_obs$breaks[i+1], col = "blue")
        segments(x0 = h_miss$breaks[i], y0 = h_miss$counts[i], x1 = h_miss$breaks[i+1], col = "red")
      }
      
      segments(x0 = h_obs$breaks[i+1], y0 = h_obs$counts[i], y1 = 0, col = "blue")
      segments(x0 = h_miss$breaks[i+1], y0 = h_miss$counts[i], y1 = 0, col = "blue")
      
      NAs <- is.na(x)[!con]
      tab <- table(x@data[!con], NAs)
      for(i in 1:NROW(tab)) {
        segments(x0 = as.numeric(rownames(tab)[i]), y0 = 0, y1 = sum(tab[i,]), col = "lightgray", lty = "dashed")
        segments(x0 = as.numeric(rownames(tab)[i]), y0 = 0, y1 = tab[i,1], col = "blue", lty = "dashed")
        if(ncol(tab) == 2) segments(x0 = as.numeric(rownames(tab)[i]), y0 = 0, y1 = tab[i,2], col = "red", lty = "dashed")
      }
      
      if(.MI_DEBUG) stopifnot(all(h_all$counts == (h_obs$counts + h_miss$counts)))
    }
    else {
      tab <- table(x@data[!con])
      for(i in 1:NCOL(tab)) segments(x0 = as.numeric(names(tab)[i]), y0 = 0, y1 = tab[i], col = "blue", lty = "dashed")
    }
    return(invisible(NULL))
  })

setMethod("hist", signature(x = "categorical"), def = 
  function(x, ...) {
    y <- x@data
    values <- sort(unique(y))
    breaks <- c(min(values) - 0.5, values + 0.5)
    
    values <- unique(y)
    values <- sort(values[!is.na(values)])
    breaks <- c(sapply(values, FUN = function(x) c(x - .25, x + .25)))
    NAs <- is.na(x)
    
    h_all  <- hist(y, breaks, plot = FALSE)
    #   h_all$counts[h_all$counts == 0] <- NA_integer_
    plot(h_all, border = "lightgray", axes = FALSE, main = "", xlab = if(x@done) "Completed" else "Observed", 
         mgp = c(2, 1, 0), tcl = .05, col = if(x@done) "lightgray" else "blue", freq = TRUE, ylim = range(h_all$counts, na.rm = TRUE), ...)
    axis(1, at = values, labels = levels(x@raw_data), lwd = 0)
    axis(2)
    
    if(x@done) {
      h_obs  <- hist(y[!NAs], breaks, plot = FALSE)
      h_miss <- hist(y[NAs], breaks, plot = FALSE)
      
      counts_obs <- h_obs$counts
      counts_obs <- counts_obs
      
      counts_miss <- h_miss$counts
      counts_miss <- counts_miss
      
      segments(breaks[1], 0,  y1 = counts_obs[1], col = "blue")
      segments(breaks[1], 0,  y1 = counts_miss[1], col = "red")
      
      if(counts_obs[1])  segments(breaks[1], y0 = counts_obs[1], x1  = breaks[2], col = "blue")
      if(counts_miss[1]) segments(breaks[1], y0 = counts_miss[1], x1 = breaks[2], col = "red")
      
      for(i in 2:(length(breaks)-1)) {
        segments(x0 = breaks[i], y0 = counts_obs[i-1], y1 = counts_obs[i], col = "blue")
        segments(x0 = breaks[i], y0 = counts_miss[i-1], y1 = counts_miss[i], col = "red")
        
        if(counts_obs[i]) segments(x0 = breaks[i], y0 = counts_obs[i], x1 = breaks[i+1], col = "blue")
        if(counts_miss[i]) segments(x0 = breaks[i], y0 = counts_miss[i], x1 = breaks[i+1], col = "red")
      }
      
      segments(x0 = breaks[i+1], y0 = counts_obs[i], y1 = 0, col = "blue")
      segments(x0 = breaks[i+1], y0 = counts_miss[i], y1 = 0, col = "red")
      
      if(.MI_DEBUG) stopifnot(all(h_all$counts == (h_obs$counts + h_miss$counts)))
    }
    
    return(invisible(NULL))
  })

setMethod("hist", signature(x = "binary"), def = 
  function(x, ...) {
    y <- x@data
    if(max(y, na.rm = TRUE) > 1) y <- y - 1L
    values <- 0:1
    breaks <- c(-.5, .5, 1.5)
    breaks <- c(-.25, .25, .75, 1.25)
    NAs <- is.na(x)
    
    h_all  <- hist(y, breaks, plot = FALSE)
    #   h_all$counts[h_all$counts == 0] <- NA_integer_
    plot(h_all, border = "lightgray", axes = FALSE, main = "", xlab = if(x@done) "Completed" else "Observed", 
         mgp = c(2, 1, 0), tcl = .05, col = if(x@done) "lightgray" else "blue", freq = TRUE, ylim = range(h_all$counts, na.rm = TRUE), ...)
    axis(1, at = values, lwd = 0)
    axis(2)
    
    if(x@done) {
      h_obs  <- hist(y[!NAs], breaks, plot = FALSE)
      h_miss <- hist(y[NAs], breaks, plot = FALSE)
      
      counts_obs <- h_obs$counts
      counts_obs <- counts_obs
      
      counts_miss <- h_miss$counts
      counts_miss <- counts_miss
      
      segments(breaks[1], 0,  y1 = counts_obs[1], col = "blue")
      segments(breaks[1], 0,  y1 = counts_miss[1], col = "red")
      
      if(counts_obs[1])  segments(breaks[1], y0 = counts_obs[1], x1  = breaks[2], col = "blue")
      if(counts_miss[1]) segments(breaks[1], y0 = counts_miss[1], x1 = breaks[2], col = "red")
      
      for(i in 2:(length(breaks)-1)) {
        segments(x0 = breaks[i], y0 = counts_obs[i-1], y1 = counts_obs[i], col = "blue")
        segments(x0 = breaks[i], y0 = counts_miss[i-1], y1 = counts_miss[i], col = "red")
        
        if(counts_obs[i])  segments(x0 = breaks[i], y0 = counts_obs[i], x1 = breaks[i+1], col = "blue")
        if(counts_miss[i]) segments(x0 = breaks[i], y0 = counts_miss[i], x1 = breaks[i+1], col = "red")
      }
      
      segments(x0 = breaks[i+1], y0 = counts_obs[i], y1 = 0, col = "blue")
      segments(x0 = breaks[i+1], y0 = counts_miss[i], y1 = 0, col = "red")
      
      if(.MI_DEBUG) stopifnot(all(h_all$counts == (h_obs$counts + h_miss$counts)))
    }
    
    return(invisible(NULL))
  })

setMethod("hist", signature(x = "missing_data.frame"), def = 
  function(x, ask = TRUE, ...) {
    k <- sum(!x@no_missing)
    if (.Device != "null device" && x@done) {
      oldask <- grDevices::devAskNewPage(ask = ask)
      if (!oldask) on.exit(grDevices::devAskNewPage(oldask), add = TRUE)
      op <- options(device.ask.default = TRUE)
      on.exit(options(op), add = TRUE)
    }
    par(mfrow = n2mfrow(k))
    for(i in 1:x@DIM[2]) {
      if(x@no_missing[i]) next
      hist(x@variables[[i]])
      header <- x@variables[[i]]@variable_name
      if(is(x@variables[[i]], "continuous")) {
        trans <- .show_helper(x@variables[[i]])$transformation[1]
        header <- paste("\n", header, " (", trans, ")", sep = "")
      }
      title(main = header)
    }
    return(invisible(NULL))
  })

setMethod("hist", signature(x = "mdf_list"), def = 
  function(x, ask = TRUE, ...) {
    if (.Device != "null device") {
      oldask <- grDevices::devAskNewPage(ask = ask)
      if (!oldask) on.exit(grDevices::devAskNewPage(oldask), add = TRUE)
      op <- options(device.ask.default = ask)
      on.exit(options(op), add = TRUE)
    }
    sapply(x, FUN = hist, ...)
    return(invisible(NULL))
  })

setMethod("hist", signature(x = "mi"), def = 
  function(x, m = 1:length(x), ask = TRUE, ...) {
    for(i in m) hist(x@data[[i]], ask = ask, ...)
    return(invisible(NULL))
  })

setMethod("hist", signature(x = "mi_list"), def = 
  function(x, m = 1:length(x), ask = TRUE, ...) {
    if (.Device != "null device") {
      oldask <- grDevices::devAskNewPage(ask = ask)
      if (!oldask) on.exit(grDevices::devAskNewPage(oldask), add = TRUE)
      op <- options(device.ask.default = ask)
      on.exit(options(op), add = TRUE)
    }
    sapply(x, FUN = hist, m = m, ask = ask, ...)
    return(invisible(NULL))
  })

stopifnot(require(mi))

x <- rnorm(10)
x[1] <- NA
y <- missing_variable(x, type = "continuous")

y <- missing_variable(x, type = "irrelevant")
x <- rep(1, 10)
y <- missing_variable(x, type = "fixed")
x <- rep(1:5, each = 2)
y <- missing_variable(x, type = "group")

x[1] <- NA
y <- missing_variable(x, type = "unordered-categorical")
y <- missing_variable(x, type = "ordered-categorical")
y <- missing_variable(x, type = "interval")

x <- rbinom(10, size = 1, prob = 0.5)
x[1] <- NA
y <- missing_variable(x, type = "binary")
y <- missing_variable(x, type = "grouped-binary", strata = rep(c("A", "B"), each = 5))

x <- runif(10)
x[1] <- NA
y <- missing_variable(x, type = "bounded-continuous", lower = 0, upper = 1)
y <- missing_variable(x, type = "positive-continuous")
y <- missing_variable(x, type = "proportion")
x[which.min(x)] <- 0
y <- missing_variable(x, type = "nonnegative-continuous")
y <- missing_variable(x, type = "SC_proportion")
x[which.max(x)] <- 1
y <- missing_variable(x, type = "SC_proportion")

x <- rpois(10, lambda = 5)
x[1] <- NA
y <- missing_variable(x, type = "count")

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## like sapply but for objects of mi class
mipply <- ## FIXME: should probably be a generic function instead of poor man's S4
  function(X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE, columnwise = TRUE, to.matrix = FALSE) {
    if(is(X, "mi_list")) {
      out <- lapply(X, mipply, ..., simplify = simplify, USE.NAMES = USE.NAMES, 
                    columnwise = columnwise, to.matrix = to.matrix)
    }
    else if(is(X, "mi")) {
      X <- complete(X, to_matrix = to.matrix)
      if(columnwise) out <- sapply(X, FUN = function(x) apply(x, 2, FUN, ...), 
                                   simplify = simplify, USE.NAMES = USE.NAMES)
      else out <- sapply(X, FUN, ..., simplify = simplify, USE.NAMES = USE.NAMES)
    }
    else if(is(X, "mdf_list")) {
      out <- lapply(X, mipply, ..., simplify = simplify, USE.NAMES = USE.NAMES, 
                    columnwise = columnwise, to.matrix = to.matrix)
    }
    else if(is(X, "missing_data.frame")) {
      if(columnwise) out <- sapply(X, FUN = function(x) apply(x, 2, FUN, ...), 
                                   simplify = simplify, USE.NAMES = USE.NAMES)
      else out <- sapply(X, FUN, ..., simplify = simplify, USE.NAMES = USE.NAMES)
    }
    else if(is(X, "missing_variable")) {
      out <- FUN(X@data, ...)
    }
    else if(is(X, "mi_list")) {
      out <- lapply(X, FUN = mipply, ..., simplify = simplify, USE.NAMES = USE.NAMES, 
                    columnwise = columnwise, to.matrix = to.matrix)
    }
    else stop("'X' must be of class 'mi', 'missing_data.frame', 'missing_variable', or 'mi_list'")
    return(out)
  }

## create a bugs array from an mi object
mi2BUGS <-
  function(imputations, statistic = c("moments", "imputations", "parameters")) {
    if(is(imputations, "mi_list")) return(lapply(imputations, FUN = mi2BUGS, statistic = statistic))
    else if(!is(imputations, "mi")) stop("imputations must be an object of class 'mi' or 'mi_list'")

    statistic <- match.arg(statistic)
    if(statistic == "moments") {
      iterations <- sum(imputations@total_iters)
      mark <- !imputations@data[[1]]@no_missing & 
        !sapply(imputations@data[[1]]@variables, is, class2 = "irrelevant")
      means <- lapply(1:iterations, FUN = function(m) {
        matrices <- lapply(imputations@data, FUN = complete, m = m, to_matrix = TRUE, include_missing = FALSE)
        out <- sapply(matrices, colMeans)[mark,,drop = FALSE]
        return(out)
      })
      sds <- lapply(1:iterations, FUN = function(m) {
        matrices <- lapply(imputations@data, FUN = complete, m = m, to_matrix = TRUE, include_missing = FALSE)
        out <- sapply(matrices, FUN = function(x) apply(x, 2, sd))[mark,,drop = FALSE]
        return(out)
      })
      dims <- dim(means[[1]])
      arr <- array(NA_real_, c(iterations, dims[2], 2 * dims[1]), 
                   list(NULL, NULL, c(paste("mean", rownames(means[[1]]), sep = "_"), 
                                      paste("sd",   rownames(means[[1]]), sep = "_"))))
      for(i in seq_along(means)) for(j in 1:ncol(arr)) {
        arr[i,j,   1:dims[1]] <- means[[i]][,j]
        arr[i,j,-c(1:dims[1])] <- sds[[i]][,j]
      }
    }
    else if(statistic == "imputations") {
      imp_list <- lapply(imputations@data, function(x) lapply(x@variables, function(y) y@imputations))
      n.parameters <- rapply(imp_list, ncol)
      arr <- array(NA_real_, c(sum(imputations@total_iters), length(imp_list), n.parameters)) ## FIXME: names?
      for(i in seq_along(imp_list)) arr[,i,] <- unlist(imp_list[[i]])
    }
    else arr <- get_parameters(imputations)
    return(arr) # compatible with R2WinBUGS
  }

##Outputs completed data in either Stata (.dta) format or comma-separated (.csv) format
mi2stata <- 
  function(imputations, m, file, missing.ind=FALSE, ...) {
    
  if(grepl("\\.csv$", file)) type <- "csv" 
	if(grepl("\\.dta$", file)) type <- "dta" 
	
	else if(!is(imputations, "mi")) stop("imputations must be an object of class 'mi'")
	else if(!is(file, "character")) stop("filename must be specified as a character object")	
	else if(type!="dta" & type!="csv") stop("file type must be 'dta' for stata format or 'csv' for comma-separated format")

	message("Note: after loading the data into Stata, version 11 or later, type 'mi import ice' to register the data as being multiply imputed. 
	For Stata 10 and earlier, install MIM by typing 'findit mim' and include 'mim:' as a prefix for any command using the MI data.")
	unpos <- sum(sapply(imputations@data[[1]]@variables, FUN=function(x){x@n_unpossible}))
	if (unpos>0 & !missing.ind) {
		missing.ind <- TRUE
		warning("There are legitimately skipped values in the data that were not imputed.  Including variables to indicate which missing values
		were imputed.  Values which are still missing but are not indicated are legitimate skips.")
	}
	if (unpos>0 & missing.ind) {
		warning("There are legitimately skipped values in the data that were not imputed. Values which are still missing but are not indicated are legitimate skips.")
	}

	data.list <- complete(imputations, m)
	if (missing.ind) miss.indic <- data.list[[1]][,which(!is.element(colnames(data.list[[1]]), names(imputations@data[[1]]@variables)))]
	vars <- which(is.element(colnames(data.list[[1]]), names(imputations@data[[1]]@variables)))

	stata.data <- data.list[[1]][,vars]
	stata.miss <- sapply(imputations@data[[1]]@variables, FUN=function(x){
		v <- is.element(1:x@n_total, x@which_drawn)
		return(v)
	}, simplify=TRUE)
	is.na(stata.data) <- stata.miss
	if (missing.ind) stata.data <- cbind(stata.data, miss.indic)
	stata.data$mi <- 1:nrow(stata.data); stata.data$mj <- 0	

	for(i in seq_along(data.list)){
		dl <- data.list[[i]]
		if(!missing.ind) dl <- dl[,vars]
		dl$mi <- 1:nrow(dl)
		dl$mj <- i
		stata.data <- rbind(stata.data, dl)
		}
	colnames(stata.data)[which(colnames(stata.data)=="mi")] <- "_mi"
	colnames(stata.data)[which(colnames(stata.data)=="mj")] <- "_mj"

	if(type=="dta") foreign::write.dta(stata.data, file=file, version = 7L, ...)
	else if(type=="csv") write.table(stata.data, file=file, sep=",", col.names=TRUE, row.names=FALSE)
  }

## Returns the Gelman statistic
Rhats <-
  function(imputations, statistic = c("moments", "imputations", "parameters")) {
    BUGS <- mi2BUGS(imputations, statistic)
    
    make_Rhat <- function(x) {
      m <- ncol(x)
      if(m < 2) stop("need at least 2 chains to calculate an R-hat")
      iter <- nrow(x)
      xbars <- colMeans(x)
      variances <- apply(x, MARGIN = 2:3, FUN = sd)^2
      W <- colMeans(variances)
      B <- iter * apply(xbars, MARGIN = 2, FUN = var)
      R <- sqrt( (iter - 1) / iter + 1 / iter * B / W )
      return(R)
    }
    
    if(is(imputations, "mi")) return(make_Rhat(BUGS))
    else return(sapply(BUGS, FUN = make_Rhat))
  }

## tests whether a method is the one defined in my (as opposed to a user-defined method in .GlobalEnv)
is.method_in_mi <-
  function(generic, ...) {
    method <- selectMethod(generic, signature(...))
    return(environmentName(environment(method@.Data)) == "mi")
  }

## cube root transformation
.cuberoot <-
  function(y, inverse = FALSE) {
    if(inverse) y^3
    else        y^(1/3)
  }

.parse_trans <-
  function(trans) {
    if(identical(names(formals(trans)), c("y", "mean", "sd", "inverse"))) return("standardize")
    if(identical(names(formals(trans)), c("y", "a", "inverse"))) return("logshift")
    if(identical(body(trans), body(.squeeze_transform))) return("squeeze")
    if(identical(body(trans), body(.identity_transform))) return("identity")
    if(identical(body(trans), body(log))) return("log")
    if(identical(body(trans), body(sqrt))) return("sqrt")
    if(identical(body(trans), body(.cuberoot))) return("cuberoot")
    if(identical(body(trans), body(qnorm))) return("qnorm")
    return("user-defined")
  }

.prune <- function(class) {
  classes <- names(getClass(class, where = "mi")@subclasses)
  classes <- classes[!sapply(classes, isVirtualClass, where = "mi")]
  if(!isVirtualClass(class, where = "mi")) classes <- c(class, classes)
  return(classes)
}

.possible_missing_variable <-
  function(y) { ## FIXME: update this function whenever you tweak the missing_variable tree
    
    mvs <- .prune("missing_variable")
    
    maybe <- rep(TRUE, length(mvs))
    names(maybe) <- mvs
    
    if(is.factor(y)) y <- factor(y) # to drop unused levels
    vals <- unique(y)
    vals <- sort(vals[!is.na(vals)])
    
    if(length(vals) == 1) {
      maybe[] <- FALSE
      maybe["irrelevant"] <- TRUE
      maybe[.prune("fixed")] <- TRUE
      return(maybe)
    }
    else maybe[.prune("fixed")] <- FALSE
    
    if(!all(table(y) > 1)) maybe[.prune("categorical")] <- FALSE
    
    if(length(vals) == 2) { # permit binary plus children but not other kinds of categorical
      maybe[.prune("categorical")] <- FALSE
      maybe[.prune("binary")] <- TRUE
      maybe[.prune("semi-continuous")] <- FALSE
    }
    else {
      maybe[.prune("binary")] <- FALSE
    }
    
    if(!is.numeric(vals)) {
      maybe[.prune("continuous")] <- FALSE
      maybe[.prune("count")] <- FALSE
      return(maybe)
    }
    
    if(any(vals < 0)) {
      maybe[.prune("nonnegative-continuous")] <- FALSE
      maybe[.prune("positive-continuous")] <- FALSE
      maybe[.prune("count")] <- FALSE
      return(maybe)
    }
    
    if(any(vals == 0)) maybe[.prune("positive-continuous")]    <- FALSE
    else               maybe[.prune("nonnegative-continuous")] <- FALSE # unless SC_proportion
    
    if(!any(vals < 1 && vals > 0)) {
      maybe[.prune("SC_proportion")] <- FALSE
      maybe[.prune("proportion")] <- FALSE
    }
    else if(any(vals >= 1)) {
      maybe[.prune("proportion")] <- FALSE
      if(any(vals > 1)) maybe[.prune("SC_proportion")] <- FALSE
      else              maybe[.prune("SC_proportion")] <- TRUE
    }
    
    if(any(vals != as.integer(vals))) {
      maybe[.prune("count")] <- FALSE
      maybe[.prune("categorical")] <- FALSE
    }
    
    return(maybe)
  }

.cat2dummies <-
  function(y) {
    if(!is(y, "categorical")) stop("must be a categorical variable")
    if(is(y, "binary")) out <- as.matrix(as.integer(y@data == 1))
    else {
      levels <- sort(unique(y@data))
      out <- t(sapply(y@data, FUN = function(x) as.integer(x == levels)[-1]))
    }
    return(out)
  }

setMethod("fitted", signature(object = "RNL"), def = 
  function(object, ...) {
    Pr <- sapply(object, FUN = function(m) {
      eta <- m$x %*% coef(m)
      pred <- m$family$linkinv(eta)
      return(pred)
    })
    Pr <- Pr / rowSums(Pr)
    return(Pr)
  })

setMethod("fitted", signature(object = "clogit"), def = 
  function(object, ...) {
    target <- mean(as.numeric(object$y))
    lp <- object$linear.predictors
    foo <- function(par) {
      intercept <- qlogis(par)
      mean(plogis(intercept + lp)) - target
    }
    opt <- uniroot(foo, lower = 0, upper = 1)
    return(plogis(qlogis(opt$root) + lp))
  })

# Borrowed from library(MCMCpack)
.rdirichlet <-
  function(n, alpha) {
    l <- length(alpha)
    x <- matrix(rgamma(l * n, alpha), ncol = l, byrow = TRUE)
    sm <- rowSums(x)
    return(x / sm)
  }

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


setGeneric("change", def = function(data, y, to, what, ...) standardGeneric("change"))
setGeneric("change_family", def = function(data, y, to, ...) standardGeneric("change_family"))
setGeneric("change_imputation_method", def = function(data, y, to, ...) standardGeneric("change_imputation_method"))
setGeneric("change_link", def = function(data, y, to, ...) standardGeneric("change_link"))
setGeneric("change_model", def = function(data, y, to, ...) standardGeneric("change_model"))
setGeneric("change_size", def = function(data, y, to, ...) standardGeneric("change_size"))
setGeneric("change_transformation", def = function(data, y, to, ...) standardGeneric("change_transformation"))
setGeneric("change_type", def = function(data, y, to, ...) standardGeneric("change_type"))
setGeneric("complete", def = function(y, m, ...) standardGeneric("complete"))
setGeneric("fit_model", def = function(y, data, ...) standardGeneric("fit_model"))
setGeneric("get_parameters", def = function(object, ...) standardGeneric("get_parameters"))
setGeneric("hist", def = function(x, ...) standardGeneric("hist"))
setGeneric("mi", def = function(y, model, ...) standardGeneric("mi"))
setGeneric("missing_variable", def = function(y, type, ...) standardGeneric("missing_variable"))
setGeneric("missing_data.frame", def = function(y, ...) standardGeneric("missing_data.frame"))
## FIXME: acount for the other stuff in the original AllGeneric.R

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## FIXME: Make damn sure .MI_DEBUG is FALSE before pushing to CRAN
.MI_DEBUG <- FALSE 

## FIXME: Also, make all if(.MI_DEBUG) statements one-liners in other files
## and do sed s/if(.MI_DEBUG)/#if(.MI_DEBUG)/g *.R

if(TRUE && .MI_DEBUG) { # define multi-line debugging functions in here
  options(error = recover)
}

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

setMethod("missing_data.frame", signature(y = "data.frame"), def = 
  function(y, subclass = NA_character_, by = NULL, types = NULL, 
           favor_ordered = TRUE, favor_positive = FALSE, threshold = 5, ...) {
    if(!is.na(subclass) && subclass == "allcategorical") threshold <- Inf
    if(!is.null(by)) {
      mdfs <- by(y, lapply(by, FUN = function(b) y[,b]), FUN = function(d) {
        missing_data.frame(d, favor_ordered = favor_ordered, threshold = threshold,
                           favor_positive = favor_positive, subclass = subclass, types = types, ...)
      })
      class(mdfs) <- "mdf_list"
      return(mdfs)
    }
    variables <- vector("list", length = ncol(y))
    if(is.null(types)) for(i in seq_along(variables)) {
      variables[[i]] <- missing_variable(y[,i], favor_ordered = favor_ordered, favor_positive = favor_positive, 
                                         variable_name = colnames(y)[i], threshold = threshold)
    }
    else for(i in seq_along(variables)) {
      variables[[i]] <- new(types[i], variable_name = colnames(y)[i], raw_data = y[,i])
    }
    
    if(is.na(subclass)) new("missing_data.frame", variables = variables, DIMNAMES = dimnames(y), ...)
    else new(paste(subclass, "missing_data.frame", sep = "_"), variables = variables, DIMNAMES = dimnames(y), ...)
  }
          )

setMethod("missing_data.frame", signature(y = "matrix"), def = 
  function(y, ...) {
    return(missing_data.frame(y = as.data.frame(y), ...))
  }
          )

setMethod("missing_data.frame", signature(y = "list"), def = 
  function(y, ...) {
    if(!all(sapply(y, is, class2 = "missing_variable"))) {
      stop("all list elements must inherit from the 'missing_variable' class")
    }
    return(new("missing_data.frame", variables = y))
  }
          )

setAs(from = "data.frame", to = "missing_data.frame", def =
  function(from) {
    missing_data.frame(from)
  }
      )

setAs(from = "matrix", to = "missing_data.frame", def = 
  function(from) {
    missing_data.frame(as.data.frame(from))
  }
      )

setAs(from = "missing_data.frame", to = "data.frame", def =
  function(from) {
    return(complete(from, m = 0L))
  }
      )

setAs(from = "missing_data.frame", to = "matrix", def =
  function(from) {
    return(complete(from, m = 0L, to_matrix = TRUE))
  }
      )

## FIXME: Probably need to add a boatload of methods to mimic the behavior of data.frames

.default_model <-
  function(y, data) {
    if(is(data, "allcategorical_missing_data.frame")) return("Gibbs")
    if(y@all_obs) {
      if(is(y, "semi-continuous")) return(rep(NA_character_, 2))
      else return(NA_character_)
    }
    if(is(y, "irrelevant")) return(NA_character_)
    if(y@imputation_method == "mcar") return(NA_character_)
    if(!is.method_in_mi("fit_model", y = class(y), data = class(data))) {
      if(is(y, "semi-continuous")) return(rep("user-defined", 2))
      else return("user-defined")
    }
    fam <- y@family$family
    link <- y@family$link
    if(is(y, "count")) {
      if(fam == "quasipoisson" && link == "log") return("qpoisson")
      else if(fam == "poisson" && link == "log") return("poisson")
      else return("****")
    }
    else if(is(y, "binary")) {
      if(is(y, "grouped-binary")) return("clogit")
      if(fam == "quasibinomial")  return(paste("q", link, sep = ""))
      else if(fam == "binomial")  return(link)
      else return("****")
    }
    else if(is(y, "interval")) return("survreg")
    else if(is(y, "ordered-categorical"))   return(paste("o", link, sep = ""))
    else if(is(y, "unordered-categorical")) {
      if(fam == "binomial") out <- "RN"
      else out <- "m"
      return(paste(out, link, sep = ""))
    }
    else if(is(y, "proportion")) return(if(fam == "gaussian") "linear" else "betareg")
    else if(is(y, "SC_proportion")) {
      out <- .default_model(y@indicator, data)
      return(c("betareg", out))
    }
    else if(is(y, "semi-continuous")) {
      out <- .default_model(y@indicator, data)
      if(fam == "gaussian") { 
        if(link == "identity") return(c("linear", out))
        else if(link == "log") return(c("loglinear", out))
        else if(link == "inverse") return(c("inverselinear", out))
        else return(c("****", out))
      }
      else if(fam == "Gamma") return(c("****", out))
      else if(fam == "inverse.gaussian") return(c("****", out))
      else if(fam == "quasi") return(c("quasi", out))
      else return(c("****", out))
    }
    else if(is(y, "continuous")) {
      if(fam == "gaussian") { 
        if(link == "identity") return("linear")
        else if(link == "log") return("loglinear")
        else if(link == "inverse") return("inverselinear")
        else return("****")
      }
      else if(fam == "Gamma") return("****")
      else if(fam == "inverse.gaussian") return("****")
      else if(fam == "quasi") return("quasi")
      else return("****")
    }
    else return("user-defined")
  }

setMethod("show", "missing_data.frame", def = 
  function(object) {
    k <- object@DIM[2]
    df <- .show_helper(object@variables[[1]])
    for(i in 2:k) {
      df <- rbind(df, .show_helper(object@variables[[i]]))
    }
    df1 <- cbind(df[,1:3], model = unlist(sapply(object@variables, FUN = function(y) .default_model(y, object))))
    if(is(object, "experiment_missing_data.frame")) df1$concept[names(object@concept)] <- object@concept
    df2 <- df[,-c(1:3)]
    cat("Object of class", class(object), "with", nrow(object), "observations on", ncol(object), "variables\n")
    if(length(object@patterns)) {
      npatterns <- nlevels(object@patterns)
      cat("\nThere are", npatterns, "missing data patterns\n")
      #       print(table(as.integer(object@patterns)))
      #       mat <- as.matrix(levels(object@patterns))
      #       colnames(mat) <- "missing"
      #       print(mat)
      cat("\nAppend '@patterns' to this", class(object), "to access the corresponding pattern for every observation or perhaps use table()\n\n")
    }
    print(df1)
    cat("\n")
    print(df2)
    if(any(df1$model == "****", na.rm = TRUE)) {
      cat("\n**** The model lacks a widely-recognized name but is determined by the chosen type, family, and link.\n")
    }
    return(invisible(NULL))
  })

setMethod("show", "mdf_list", def = 
  function(object) {
    for(i in seq_along(object)) {
      cat("\n", names(object)[i], "\n")
      show(object[[i]])
    }
    return(invisible(NULL))
  })

setMethod("summary", "missing_data.frame", def = 
   function(object) {
     summary(as.data.frame(object))
   })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

.prune_missing_variable <-
  function(y, s) {
    if(!is(y, "missing_variable")) stop("'y' must inherit from the 'missing_variable' class")
    if(!y@all_obs) {
      y@parameters <- y@parameters[1:s,,drop = FALSE]
      y@imputations <- y@imputations[1:s,,drop = FALSE]
    }
    return(y)
  }

.MPinverse <- function(eta, tol = sqrt(.Machine$double.eps)) {
  cov_eta <- cov(eta)
  ev <- eigen(cov_eta, TRUE)
  ev$values <- ifelse(ev$values > tol, 1/ev$values, 0)
  Sigma_inv <- crossprod(sqrt(ev$values)*(t(ev$vectors)))
  return(Sigma_inv)
}

.mi <- function(i, y, verbose, s_start, s_end, ProcStart, max.minutes, parallel, save_models) {
  mdf <- y
  if(verbose) message("Chain ", i, "\n")
  for(s in s_start:s_end) {
    if(verbose) message("Chain ", i, " Iteration ", s, "\n")
    mdf <- fit_model(data = mdf, s = s, verbose = FALSE, warn = s == s_end)
    if(s > 0) {
      pars <- unlist(sapply(mdf@variables, FUN = function(y) {
        if(is(y, "irrelevant")) return(NA_real_)
        else if(y@all_obs) return(NA_real_)
        else return(y@parameters[s,,drop=TRUE])
      }))
      pars <- t(pars[!is.na(pars)])
      fp <- file.path(mdf@workpath, paste0("pars_", i, ".csv"))
      write.table(pars, file = fp, append = TRUE, sep = ",",
                  row.names = FALSE, col.names = FALSE)
      imps <- unlist(sapply(mdf@variables, FUN = function(y) {
        if(is(y, "irrelevant")) return(NA_real_)
        else if(y@all_obs) return(NA_real_)
        else return(y@imputations[s,,drop=TRUE])
      }))
      imps <- t(imps[!is.na(imps)])
      fp <- file.path(mdf@workpath, paste0("imps_", i, ".csv"))
      write.table(imps, file = fp, append = TRUE, sep = ",",
                  row.names = FALSE, col.names = FALSE)
    }
    Time.Elapsed <- proc.time() - ProcStart
    if(((Time.Elapsed)/60)[3] > max.minutes) {
      warning("'max.minutes' threshold exceeded")
      break
    }
  }
  if(((Time.Elapsed)/60)[3] > max.minutes) mdf@variables <- lapply(mdf@variables, .prune_missing_variable, s = s)
  
  if(verbose) message("Estimating models on completed data for chain ", i, "\n")

  mdf@variables <- lapply(mdf@variables, FUN = function(y) {
    if(!y@all_obs & !is(y, "irrelevant")) {
      model <- fit_model(y, mdf, s = s + 1, warn = TRUE)
      y@fitted <- fitted(model)
      if(!isS4(model)) model$x <- model$X <- model$y <- model$model <- NULL
      if(save_models) y@model <- model
    }
    else y@model <- NULL
    return(y)
  })
  
  mdf@done <- TRUE
  if(verbose) message("Done with chain ", i, "\n")
  return(mdf)
}

.mi_split <- function(i, y, data, verbose, s_start, s_end, ProcStart, max.minutes, parallel, save_models) {
  mdf <- y
  if(verbose) message("Chain ", i, "\n")
  data@priors <- mdf@priors
  for(s in s_start:s_end) {
    if(verbose) message("Chain ", i, " Iteration ", s, "\n")
    mdf <- fit_model(mdf, data, s = s, verbose = FALSE, warn = s == s_end)
    Time.Elapsed <- proc.time() - ProcStart
    if(((Time.Elapsed)/60)[3] > max.minutes) {
      warning("'max.minutes' threshold exceeded")
      break
    }
  }
  
  if(((Time.Elapsed)/60)[3] > max.minutes) mdf@variables <- lapply(mdf@variables, .prune_missing_variable, s = s)
  
  if(verbose) message("Estimating models on completed data for chain ", i, "\n")
  
  mdf@variables <- lapply(mdf@variables, FUN = function(y) {
    if(!y@all_obs & !is(y, "irrelevant")) {
      model <- fit_model(y, data, s = s + 1, warn = TRUE)
      y@fitted <- fitted(model)
      if(!isS4(model)) model$x <- model$X <- model$y <- model$model <- NULL
      if(save_models) y@model <- model
    }
    else y@model <- NULL
    return(y)
  })
  
  mdf@done <- TRUE
  if(verbose) message("Done with chain ", i, "\n")
  return(mdf)
}


setMethod("mi", signature(y = "missing_data.frame", model = "missing"), def =
  function(y, n.iter = 30,  n.chains = 4, max.minutes = Inf, seed = NA, verbose = TRUE, 
           save_models = FALSE, parallel = .Platform$OS.type != "Windows")
  {
    call <- match.call()
    if(!is.na(seed)) set.seed(seed)
    if(n.iter < 0) stop(message="number of iterations must be non-negative")
    ProcStart <- proc.time()
    s_start <- 0
    s_end <- n.iter

    Time.Elapsed <- proc.time() - ProcStart
    y@variables <- lapply(y@variables, FUN = function(x) {
      if(!x@all_obs & !is(x, "irrelevant")) {
        x@parameters  <- matrix(NA_real_, nrow = n.iter, ncol = 0)
        x@imputations <- matrix(NA_real_, nrow = n.iter, ncol = x@n_drawn)
        if(is(x, "semi-continuous")) {
          x@indicator@parameters  <- matrix(NA_real_, nrow = n.iter, ncol = 0)
          x@indicator@imputations <- matrix(NA_real_, nrow = n.iter, ncol = x@n_drawn)
        }
      }
      x@done <- TRUE
      return(x)
    })
    if(is(y, "allcategorical_missing_data.frame")) {
      y@latents@imputations <- matrix(NA_integer_, nrow = n.iter, ncol = nrow(y))
      y@latents@levels <- as.character(1:y@Hstar)
    }
    
    if(n.chains <= 0) return(y)
    
    if(is.logical(parallel) && parallel) {
      cores <- getOption("mc.cores", 2L)
      cl <- parallel::makeCluster(cores, outfile = "")
      on.exit(parallel::stopCluster(cl))
    }

    if(!parallel) {
      mdfs <- vector("list", n.chains)
      for(i in seq_along(mdfs)) {
        ProcStart <- proc.time()
        mdfs[[i]] <- .mi(i, y, verbose, s_start, s_end, ProcStart, 
                         max.minutes, parallel, save_models)
      }
    }
    else {
      mdfs <- parallel::parLapply(cl, X = as.list(1:n.chains),
                        fun = function(i) .mi(i, y, verbose, s_start, s_end,
                                              ProcStart, max.minutes, parallel, save_models))
    }
#     
#     else mdfs <- mclapply(as.list(1:n.chains),
#                           FUN = function(i) .mi(i, y, verbose, s_start, s_end,
#                                                 ProcStart, max.minutes, parallel, save_models))
    
    
    names(mdfs) <- paste("chain", 1:length(mdfs), sep = ":")
    
    object <- new("mi", 
                  call        = call,
                  data        = mdfs,
                  total_iters = as.integer(s_end))

    return(object)
  })

setMethod("mi", signature(y = "data.frame", model = "missing"), def =
  function(y, n.iter = 30,  n.chains = 4, max.minutes = Inf, seed = NA, verbose = TRUE, 
           save_models = FALSE, parallel = .Platform$OS.type != "Windows")
  {
    y <- as(y, "missing_data.frame")
    return(mi(y, n.iter = n.iter, n.chains = n.chains, max.minutes = max.minutes, 
              seed = seed, verbose = verbose, save_models = save_models, 
              parallel = parallel))
  })

setMethod("mi", signature(y = "matrix", model = "missing"), def =
  function(y, n.iter = 30,  n.chains = 4, max.minutes = Inf, seed = NA, verbose = TRUE,
           save_models = FALSE, parallel = .Platform$OS.type != "Windows")
  {
    y <- as(y, "missing_data.frame")
    return(mi(y, n.iter, n.chains, max.minutes, seed, verbose, save_models, parallel))
  })

setMethod("mi", signature(y = "mi", model = "missing"), 
          function(y, n.iter = 30,   max.minutes = Inf, seed = NA, verbose = TRUE, 
                   save_models = FALSE, parallel = .Platform$OS.type != "Windows")
              
          {
            call <- match.call()
            
            if(!is.na(seed)) set.seed(seed)
            
            if(n.iter < 1) stop(message="number of iterations must be at least 1")
            
            ProcStart <- proc.time()
            
            total_iters <- y@total_iters
            s_start <- sum(total_iters) + 1
            s_end <- s_start + n.iter - 1
            
            mdfs <- y@data
            n.chains <- length(mdfs)
            for(i in 1:n.chains) {
              y <- mdfs[[i]]
              if(TRUE) y@variables <- lapply(y@variables, FUN = function(x) {                
                if(x@all_obs & is(x, "irrelevant")) return(x)
                  x@imputations <- rbind(x@imputations, matrix(NA_integer_, n.iter, x@n_drawn))
                  x@parameters  <- rbind(x@parameters,  matrix(NA_real_, n.iter, ncol(x@parameters)))
                  if(is(x, "semi-continuous")) {
                    x@indicator@imputations <- rbind(x@indicator@imputations, matrix(NA_integer_, n.iter, x@indicator@n_drawn))
                    x@indicator@parameters  <- rbind(x@indicator@parameters, matrix(NA_real_, n.iter, ncol(x@indicator@parameters)))
                  }
                return(x)
              })
            }
            
            if(is.logical(parallel) && parallel) {
              cores <- getOption("mc.cores", 2L)
              cl <- parallel::makeCluster(cores, outfile = "")
              on.exit(parallel::stopCluster(cl))
            }

            if(!parallel) {
              mdfs <- vector("list", n.chains)
              for(i in seq_along(mdfs)) {
                ProcStart <- proc.time()
                mdfs[[i]] <- .mi(i, y, verbose, s_start, s_end, ProcStart, max.minutes, 
                                 parallel, save_models)
              }
            }
            else {
              mdfs <- parallel::parLapply(cl, as.list(1:n.chains),
                                fun = function(i) .mi(i, y, verbose, s_start, s_end,
                                                      ProcStart, max.minutes, parallel, save_models))
            }
#             else mdfs <- mclapply(as.list(1:n.chains),
#                                   FUN = function(i) .mi(i, y, verbose, s_start, s_end,
#                                                         ProcStart, max.minutes, parallel, save_models))
            
            object <- new("mi", 
                          call        = call,
                          data        = mdfs,
                          total_iters = as.integer(c(total_iters, n.iter)))
            return(object)
          })

setMethod("mi", signature(y = "missing_data.frame", model = "mi"), def =
  function(y, model, n.iter = sum(model@total_iters), max.minutes = 20, seed = NA, 
           verbose = TRUE, save_models = FALSE, 
           parallel = .Platform$OS.type != "Windows")
  {
    n.chains <- length(model)
    call <- match.call()
    if(!is.na(seed)) set.seed(seed)
    
    y <- mi(y, n.chains = 0L, n.iter = n.iter)
    
    ProcStart <- proc.time()
    s_start <- 0
    s_end <- n.iter

    if(is.logical(parallel) && parallel) {
      cores <- getOption("mc.cores", 2L)
      cl <- parallel::makeCluster(cores, outfile = "")
      on.exit(parallel::stopCluster(cl))
    }

    mdfs <- model@data
    if(!parallel) {
      for(i in seq_along(mdfs)) {
        ProcStart <- proc.time()
        mdfs[[i]] <- .mi_split(i, y, mdfs[[i]], verbose, s_start, s_end, ProcStart, 
                         max.minutes, parallel, save_models)
      }
    }
    else {
      mdfs <- parallel::parLapply(cl, as.list(1:n.chains),
                        fun = function(i) .mi_split(i, y, mdfs[[i]], verbose, s_start, s_end,
                                              ProcStart, max.minutes, parallel, save_models))
    }
#     else mdfs <- mclapply(as.list(1:n.chains),
#                           FUN = function(i) .mi_split(i, y, mdfs[[i]], verbose, s_start, s_end,
#                                                 ProcStart, max.minutes, parallel, save_models))
    
    names(mdfs) <- paste("chain", 1:length(mdfs), sep = ":")
    
    to_drop <- 1:ncol(model@data[[1]]@X)
    for(i in 1:n.chains) {
      model@data[[i]]@variables <- c(model@data[[i]]@variables, mdfs[[i]]@variables)
      model@data[[i]]@no_missing <- c(model@data[[i]]@no_missing, mdfs[[i]]@no_missing)
      # leave patterns as is I guess
      model@data[[i]]@DIM[2] <- model@data[[i]]@DIM[2] + mdfs[[i]]@DIM[2]
      model@data[[i]]@DIMNAMES[[2]] <- c(model@data[[i]]@DIMNAMES[[2]], mdfs[[i]]@DIMNAMES[[2]])
      mdfs[[i]]@index <- lapply(mdfs[[i]]@index, FUN = function(x) if(is.null(x)) x else to_drop)
      model@data[[i]]@index <- c(model@data[[i]]@index, mdfs[[i]]@index)
      model@data[[i]]@weights <- c(model@data[[i]]@weights, mdfs[[i]]@weights)
      model@data[[i]]@priors <- c(model@data[[i]]@priors, mdfs[[i]]@priors)
    }
    
    object <- new("mi", 
                  call        = call,
                  data        = model@data,
                  total_iters = as.integer(s_end))
    
    return(object)
    
  })

setMethod("mi", signature(y = "mdf_list", model = "missing"), def =
  function (y, ...) {
    out <- lapply(y, FUN = mi, ...)
    class(out) <- "mi_list"
    return(out)
  })

setMethod("mi", signature(y = "list", model = "missing"), def =
  function (y, ...) {
    if(!all(sapply(y, is, class2 = "mi"))) {
      stop("all elements of 'y' must be mi objects or missing_data.frame objects")
    }
    
    ## FIXME: should probably check that all the mi objects are based on the same missing_data.frame
    mdfs <- lapply(mi, FUN = function(x) return(x@data))
    
    object <- new("mi", 
                  call        = y[[1]]@call,
                  data        = mdfs,
                  total_iters = y[[1]]@total_iters)
    return(object)
  })

setMethod("mi", signature(y = "mdf_list", model = "missing"), 
          function (y, n.iter = 30,  n.chains = 4, max.minutes = Inf, seed = NA, verbose = TRUE, 
                    save_models = FALSE, parallel = .Platform$OS.type != "Windows")
          {
            
            out <- lapply(y, mi, n.iter = n.iter, n.chains = n.chains, max.minutes = max.minutes, 
                          seed = seed, verbose = verbose, save_models = save_models, parallel = parallel)
            class(out) <- "mi_list"
            return(out)
          })

setMethod("mi", signature(y = "mi_list", model = "missing"), def = 
  function (y, ...) {
    out <- lapply(y, FUN = mi, ...)
    class(out) <- "mi_list"
    return(out)
  })

setMethod("show", signature(object = "mi"), def =
  function(object) {
    cat("Object of class", class(object), "with", length(object@data), "chains, each with", 
        sum(object@total_iters), "iterations.\n")
    cat("Each chain is the evolution of an object of", class(object@data[[1]]), "class with",
        nrow(object@data[[1]]), "observations on", ncol(object@data[[1]]), "variables.\n")
    return(invisible(NULL))
  })

setMethod("show", signature(object = "mi_list"), def =
  function(object) {
    sapply(object, show)
    return(invisible(NULL))
  })

setMethod("summary", signature(object = "mi"), def =
  function(object) {
    mdf <- object@data[[1]]
    matrices <- complete(object, to_matrix = TRUE, include_missing = FALSE)
    chains <- length(matrices)
    matrices <- array(unlist(matrices), dim = c(dim(mdf), chains), dimnames = c(dimnames(mdf), NULL))
    out <- vector("list", ncol(mdf))
    names(out) <- colnames(mdf)
    for(i in seq_along(out)) {
      if(mdf@no_missing[i]) {
        if(is(mdf@variables[[i]], "categorical")) {
          mat <- table(matrices[,i,1])
          lev <- mdf@variables[[i]]@levels
          if(length(lev) && length(dim(mat)) > 1) colnames(mat) <- lev
        }
        else mat <- summary(matrices[,i,1])
        out[[i]] <- list(is_missing = "all values observed", observed = mat)
      }
      else if(is(mdf@variables[[i]], "categorical")) {
        mark <- is.na(mdf@variables[[i]])
        mat <- table(c(matrices[,i,]), rep(mark, times = chains))
        lev <- mdf@variables[[i]]@levels
        if(length(lev)) rownames(mat) <- lev
        colnames(mat) <- c("observed", "imputed")
        out[[i]] <- list(crosstab = mat) 
      }
      else {
        missing <- is.na(mdf@variables[[i]]@raw_data)
        out[[i]] <- list(is_missing = table(missing), imputed = summary(c(matrices[missing,i,])), 
                         observed = summary(c(matrices[!missing,i,])))
      }
    }
    return(out)
  })

setMethod("traceplot", signature(x = "mi"), def =
  function(x, ...) {
    traceplot(mi2BUGS, ...)
  })

setMethod("traceplot", signature(x = "mi_list"), def =
  function(x, ...) {
    traceplot(lapply(x, mi2BUGS, ...))
  })


## all the mi() methods below should return the missing_variable after imputing
## need to explicitly write out methods instead of doing poor man's S4
setMethod("mi", signature(y = "missing_variable", model = "ANY"), def = 
  function(y, model, ...) {
    stop("This method should not have been called. You need to define the relevant mi() S4 method")
  })

setMethod("mi", signature(y = "missing_variable", model = "missing"), def = 
  function(y) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    draws <- sample(y@data[y@which_obs], size = y@n_drawn, replace = TRUE)
    y@data[y@which_drawn] <- draws
    return(y)
  })

setMethod("mi", signature(y = "semi-continuous", model = "missing"), def = 
  function(y) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    y@indicator <- mi(y@indicator)
    draws <- sample(y@data[y@which_obs], size = y@n_drawn, replace = TRUE)
    if(is(y, "SC_proportion")) {
      n <- y@n_total
      if(is(y@indicator, "binary")) {
        mark <- which(complete(y@indicator, m = 0L)[y@which_miss] == 1)
        if(any(y@raw_data == 0, na.rm = TRUE)) draws[mark] <- .5 / n
        else                                   draws[mark] <- (n - .5) / n
      }
      else {
        mark <- which(complete(y@indicator, m = 0L)[y@which_miss] != 0)
        draws[mark] <- (draws[mark] * (n - 1) + .5) / n
      }
    }
    else if(is(y, "nonnegative-continuous")) {
      mark <- which(y@indicator@data[y@which_miss] == 1)
      if(length(mark)) draws[mark] <- y@transformation(rep(0, length(mark)))
    }
    else stop("FIXME: semi-continuous is not supported yet")
    y@data[y@which_drawn] <- draws
    return(y)
  })

# setMethod("mi", signature(y = "semi-continuous", model = "missing"), def = 
# function(y) {
#   if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
# 
#   categories <- 1:(ncol(y@indicator@dummies) + 1)
#   draws <- sample(categories, size = y@n_drawn, replace = TRUE)
#   dummies <- t(sapply(draws, FUN = function(x) x == categories))[,-1,drop = FALSE]
#   y@indicator@dummies[y@which_drawn,] <- dummies
#   y@indicator@data[y@which_drawn] <- draws
# 
#   draws <- sample(y@data[y@which_obs], size = y@n_drawn, replace = TRUE)
#   if(is(y, "SC_proportion")) {
#     n <- y@n_total
#     if(is(y@indicator, "binary")) {
#       mark <- which(complete(y@indicator, m = 0L)[y@which_miss] == 1)
#       if(any(y@raw_data == 0, na.rm = TRUE)) draws[mark] <- .5 / n
#       else                                   draws[mark] <- (n - .5) / n
#     }
#     else {
#       mark <- which(complete(y@indicator, m = 0L)[y@which_miss] != 0)
#       draws[mark] <- (draws[mark] * (n - 1) + .5) / n
#     }
#   }
#   else if(is(y, "nonnegative-continuous")) {
#     mark <- which(y@indicator@data[y@which_miss] == 1)
#     if(length(mark)) draws[mark] <- y@transformation(rep(0, length(mark)))
#   }
# 
#   the_range <- range(y@data, na.rm = TRUE)
#   free <- y@data[y@which_obs]
#   free <- free[free != the_range[1] & free != the_range[2]]
#   draws <- sample(free, size = y@n_drawn, replace = TRUE)
#   y@data[y@which_drawn] <- draws
#   return(y)
# })

setMethod("mi", signature(y = "bounded-continuous", model = "missing"), def = 
  function(y) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    a <- if(length(y@lower) == 1) y@lower else y@lower[y@which_drawn]
    a <- ifelse(a == -Inf, min(y@data, na.rm = TRUE), a)
    a <- ifelse(a == Inf,  max(y@data, na.rm = TRUE), a)
    
    b <- if(length(y@upper) == 1) y@upper else y@upper[y@which_drawn]
    b <- ifelse(b == -Inf, min(y@data, na.rm = TRUE), b)
    b <- ifelse(b == Inf,  max(y@data, na.rm = TRUE), b)
    
    draws <- runif(y@n_drawn, min = a, max = b)
    y@data[y@which_drawn] <- draws
    return(y)
  })

setMethod("mi", signature(y = "categorical", model = "missing"), def = 
  function(y) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    draws <- sample(y@data[y@which_obs], size = y@n_drawn, replace = TRUE)
    y@data[y@which_drawn] <- draws
    return(y)
  })

.draw_parameters <-
  function(means, ev) {
    if(any(ev$values <= 0)) return(means)
    else return(means + (ev$vectors %*% (sqrt(ev$values) * rnorm(length(means))))[,1])
  }

.pmm <-
  function(y, eta, Sigma_inv = NULL, strata = NULL) {
    if(is(y, "unordered-categorical")) {
      if(is.null(Sigma_inv)) Sigma_inv <- .MPinverse(eta)
      MD <- mahalanobis(eta, colMeans(eta), Sigma_inv, inverted=TRUE)
      MD_observed <- MD[y@which_obs]
      y_observed <- y@data[y@which_obs]
      draws <- sapply(MD[y@which_drawn], FUN = function(x) {
        mark <- which.min(abs(MD_observed - x))
        drawmark <- c(y_observed[mark], mark)
        return(drawmark)
      })
    }
    else if(is(y, "grouped-binary")) {
      draws <- sapply(y@which_drawn, FUN = function(i) {
        which_same <- which(strata == strata[i])
        candidates <- intersect(which_same, y@which_obs)
        if(length(candidates) == 0) {
          msg <- paste(y@variable_name, ": must have some observed values in each group")
          stop(msg)
        }
        eta_can <- eta[candidates]
        y_can <- y@data[candidates]
        mark <- which.min(abs(eta_can - eta[i]))
        drawmark <- c(y_can[mark], mark)
        return(drawmark)
      })
    }
    else {
      eta_obs <-  eta[y@which_obs]
      y_obs <- y@data[y@which_obs]
      draws <- sapply(eta[y@which_drawn], FUN = function(x) {
        if(is.na(x)) return(NA_real_) # happens with semi-continuous
        mark <- which.min(abs(eta_obs - x))
        drawmark <- c(y_obs[mark], mark)
        return(drawmark)
      })
    }
    return(t(draws))
  }

setOldClass("polr")
setMethod("mi", signature(y = "ordered-categorical", model = "polr"), def = 
  function(y, model, s, ...) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    
    if(!is.element(y@imputation_method,  c("ppd", "pmm"))) badHessian <- FALSE
    else if(is.null(model$Hessian)) badHessian <- FALSE
    else if(!all(is.finite(model$Hessian))) badHessian <- TRUE
    else {
      means <- c(coef(model), model$zeta)
      ev <- eigen(vcov(model), symmetric = TRUE)
      badHessian <- any(ev$values <= 0)
      parameters <- .draw_parameters(means, ev)
      while(!badHessian && 
        any(diff(parameters[-(1:ncol(model$x))]) <= 0)) { # rejection sampling on cutpoints
        parameters <- .draw_parameters(means, ev)
      }
    }
    
    if(badHessian && y@imputation_method == "ppd") {
      warning(paste("predictive mean matching used for", y@variable_name, "on iteration", s,
                    "as a fallback due to Hessian error"))
      
      old_method <- y@imputation_method
      y@imputation_method <- "pmm"
      y <- mi(y, model, s, ...)
      y@imputation_method <- old_method
      return(y)
    }
    else if(y@imputation_method == "ppd") {
      eta <- as.vector(model$x[y@which_drawn,,drop=FALSE] %*% head(parameters, ncol(model$x)))
      pfun <- switch(y@family$link, logit = plogis, probit = pnorm, 
                     cloglog = function(q) exp(-exp(-q)), cauchit = pcauchy)
      zeta <- parameters[-(1:ncol(model$x))]
      draws <- sapply(eta, FUN = function(x) {
        which(rmultinom(1, 1, diff(c(0,pfun(zeta - x),1))) == 1)
      })
    }
    else if(y@imputation_method == "pmm") {
      parameters <- c(coef(model), model$zeta)
      eta <- model$x %*% parameters[1:ncol(model$x)]
      pmm <- .pmm(y, eta)
      draws <- pmm[,1]
      y@fitted[y@which_drawn,] <- y@fitted[y@which_obs,][pmm[,2],]
    }
    else if(y@imputation_method == "median") {
      predictions <- predict(model, type = "class")
      draws <- rep(floor(median(predictions[y@which_obs])), y@n_drawn)
    }
    else if(y@imputation_method == "mode") draws <- predict(model, type = "class")[y@which_drawn]
    else stop("'imputation_method' not recognized")
    
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

setOldClass("multinom")
setMethod("mi", signature(y = "unordered-categorical", model = "multinom"), def = 
  function(y, model, s, ...) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    ev <- eigen(vcov(model), symmetric = TRUE)
    parameters <- .draw_parameters(t(coef(model)), ev)
    if (ncol(model.matrix(model)) != nrow(parameters)) parameters <- t(parameters)
    eta <- model.matrix(model) %*% parameters
    if(y@imputation_method == "ppd") {
      exp_eta <- matrix(pmin(.Machine$double.xmax / ncol(eta), 
                             cbind(1, exp(eta[y@which_drawn,,drop = FALSE]))), 
                        ncol = ncol(eta) + 1)
      denom <- rowSums(exp_eta)
      Pr <- exp_eta / denom
      if (y@use_NA) {
        Pr <- Pr[,-1]/rowSums(Pr[,-1])
        badrows <- apply(is.na(Pr), 1, all)
        if(any(badrows)) {
          warning("Some rows of Pr are all 0 after dropping the missingness category")
          Pr[badrows,] <- 1/(ncol(Pr) - 1)
        }
      }
      draws <- apply(Pr, 1, FUN = function(p) which(rmultinom(1, 1, p) == 1))
    }
    else if(y@imputation_method == "pmm"){
    	pmm <- .pmm(y, eta)
    	draws <- pmm[,1]
    	y@fitted[y@which_drawn,,drop=FALSE] <- y@fitted[y@which_obs,,drop=FALSE][pmm[,2]]
    } 
    else if(y@imputation_method == "mode") draws <- predict(model, type = "class")[y@which_drawn]
    else stop("'imputation_method' not recognized")
    
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

setOldClass("RNL")
setMethod("mi", signature(y = "unordered-categorical", model = "RNL"), def = 
  function(y, model, s, ...) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    if(y@imputation_method == "ppd") { # imputating from the posterior predictive distribution
      Pr <- sapply(model, FUN = function(m) {
      	  ev <- eigen(vcov(m), symmetric = TRUE)
    	  parameters <- .draw_parameters(coef(m), ev)
    	  eta <- m$x[y@which_drawn,,drop=FALSE] %*% parameters
    	  pred <- m$family$linkinv(eta)
    	  return(pred)
  	  })
      if(y@use_NA) {
        Pr <- Pr[,-1]/rowSums(Pr[,-1])
        badrows <- apply(is.na(Pr), 1, all)
        if(any(badrows)) {
          warning("Some rows of Pr are all 0 after dropping the missingness category")
          Pr[badrows,] <- 1/(ncol(Pr) - 1)
        }
      }
      draws <- apply(Pr, 1, FUN = function(p) which(rmultinom(1, 1, p) == 1))
    }
    else if(y@imputation_method == "pmm") {
	  eta <- sapply(model, FUN = function(m) {
        ev <- eigen(vcov(m), symmetric = TRUE)
        parameters <- .draw_parameters(coef(m), ev)
        eta <- m$x %*% parameters
        return(eta)
      })
      pmm <- .pmm(y, eta)
      draws <- pmm[,1]
      y@fitted[y@which_drawn,,drop=FALSE] <- y@fitted[y@which_obs,,drop=FALSE][pmm[,2]]
    }
    else stop("only ppd and pmm are supported imputation methods in the RNL case")
    
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

setOldClass("glm")
setMethod("mi", signature(y = "binary", model = "glm"), def = 
  function(y, model, s, ...) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    if(y@imputation_method == "ppd") {
      ev <- eigen(vcov(model), symmetric = TRUE)
      parameters <- .draw_parameters(coef(model), ev)
      eta <- model$x[y@which_drawn,,drop=FALSE] %*% parameters
      pred <- model$family$linkinv(eta)
      draws <- rbinom(y@n_drawn, 1, pred) + 1L
    }
    else if(y@imputation_method == "pmm") {
      ev <- eigen(vcov(model), symmetric = TRUE)
      parameters <- .draw_parameters(coef(model), ev)
      eta <- model$x %*% parameters
      pmm <- .pmm(y, eta)
      draws <- pmm[,1]
      y@fitted[y@which_drawn] <- y@fitted[y@which_obs][pmm[,2]]
    }
    else if(y@imputation_method == "median") {
      predictions <- predict(model, type = "class")
      draws <- rep(floor(median(predictions[y@which_obs])), y@n_drawn)
    }
    else if(y@imputation_method == "mode") draws <- predict(model, type = "class")[y@which_drawn]
    else if(y@imputation_method == "mean") stop("'mean' is not a supported 'imputation_method' for binary variables")
    else if(y@imputation_method == "expectation") stop("'expectation' is not a supported 'imputation_method' for binary variables")
    else stop("'imputation_method' not recognized")
    
    draws <- as.integer(draws)
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

setOldClass("clogit")
setMethod("mi", signature(y = "grouped-binary", model = "clogit"), def = 
function(y, model, s, ...) {
  if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
  # reconstruc the strata
  Terms <- model$terms
  temp <- survival::untangle.specials(Terms, "strata")
  mf <- model.frame(model)
  strata <- strata(mf[, temp$vars], shortlabel = TRUE)
  if(y@imputation_method == "pmm") {
    ev <- eigen(vcov(model), symmetric = TRUE)
    parameters <- .draw_parameters(coef(model), ev)
    eta <- model$x %*% parameters
    draws <- .pmm(y, eta, strata = strata)[,1]
    #FIXME: haven't adjusted fitted values
  }
  else stop("only 'pmm' is supported for 'grouped-binary' variables")

  draws <- as.integer(draws)
  y@data[y@which_drawn] <- draws
  y@imputations[s,] <- draws
  return(y)
})

setMethod("mi", signature(y = "interval", model = "glm"), def = 
  function(y, model, s, ...) {
    stop("FIXME: write this method")
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    if(y@imputation_method == "ppd") {
      stop("FIXME")
    }
    else stop("only ppd is supported as an imputation method for interval variables")
    
    return(y)
  })

setMethod("mi", signature(y = "categorical", model = "matrix"), def = 
  function(y, model, s, ...) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    if(y@imputation_method != "ppd") stop("only ppd is supported in this case")
    if(nrow(model) != y@n_drawn) stop("matrix of probabilities has the wrong number of rows")
    draws <- apply(model, 1, FUN = function(p) which(rmultinom(1, 1, p) == 1))    
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

## helper function
.mi_continuous <-
  function(y, model) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    if(y@imputation_method == "ppd") {
      ev <- eigen(vcov(model), symmetric = TRUE)
      parameters <- .draw_parameters(coef(model), ev)
      if(model$family$family == "gaussian") {
        eta <- model$x[y@which_drawn,,drop=FALSE] %*% parameters
        pred <- model$family$linkinv(eta)
        if(is(y, "bounded-continuous")) {
          a <- if(length(y@lower) > 1) y@lower[y@which_drawn] else y@lower
          b <- if(length(y@upper) > 1) y@upper[y@which_drawn] else y@upper
          draws <- truncnorm::rtruncnorm(y@n_drawn, mean = pred, 
                                          sd = sqrt(model$dispersion), a = a, b = b)
        }
        else draws <- rnorm(y@n_drawn, pred, sqrt(model$dispersion))
      }
      else {
        eta <- model$x %*% parameters
        model$fitted <- model$family$linkinv(eta)
#         model$dispersion <- parameters@sigma^2
        draws <- y@family$sim(model, nsim = 1)[y@which_drawn]
      }
    }
    else if(y@imputation_method == "pmm") {
      ev <- eigen(vcov(model), symmetric = TRUE)
      parameters <- .draw_parameters(coef(model), ev)
      if(is(y, "semi-continuous")) {
        eta <- rep(NA_real_, y@n_total)
        mark <- complete(y@indicator, 0L) == 0
        eta[mark] <- model$x[mark,] %*% parameters
      }
      else eta <- model$x %*% parameters
      draws <- .pmm(y, eta)[,1]
      #FIXME: haven't adjusted fitted values using pmm for continuous
    }
    else if(y@imputation_method == "mean") {
      eta <- predict(model, type = "response")
      eta_observed <- eta[y@which_obs]
      eta_mean <- mean(eta_observed)
      draws <- rep(eta_mean, y@n_drawn)
    }
    else if(y@imputation_method == "median") {
      eta <- predict(model, type = "response")
      eta_observed <- eta[y@which_obs]
      eta_median <- median(eta_observed)
      draws <- rep(eta_median, y@n_drawn)
    }
    else if(y@imputation_method == "expectation") draws <- predict(model, type = "response")[y@which_drawn]
    else stop("'imputation_method' not recognized")
    return(draws)
  }

setMethod("mi", signature(y = "continuous", model = "glm"), def = 
  function(y, model, s, ...) {
    draws <- .mi_continuous(y, model)
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

# setMethod("mi", signature(y = "censored-continuous", model = "glm"), def = 
# function(y, model, s, ...) {
#   not_obs <- c(y@which_drawn, y@which_censored)
#   if(y@imputation_method == "ppd") {
#     parameters <- arm::sim(model, 1)
#     eta <- model$x[not_obs,,drop=FALSE] %*% parameters@coef[1,]
#     pred <- model$family$linkinv(eta)
#     draws <- rnorm(y@n_drawn, pred, parameters@sigma)
#   }
#   else if(y@imputation_method == "pmm") {
#     eta <- predict(model, type = "link")
#     eta_observed <- eta[y@which_obs]
#     y_observed <- y@data[y@which_obs]
#     draws <- sapply(eta[nob_obs], FUN = function(x) {
#                      mark <- which.min(abs(eta_observed - x))
#                      return(y_observed[mark])
#                    })
#   }
#   else if(y@imputation_method == "mean") {
#     eta <- predict(model, type = "response")
#     eta_observed <- eta[y@which_obs]
#     eta_mean <- mean(eta_observed)
#     draws <- rep(eta_mean, length(not_obs))
#   }
#   else if(y@imputation_method == "median") {
#     eta <- predict(model, type = "response")
#     eta_observed <- eta[y@which_obs]
#     eta_median <- median(eta_observed)
#     draws <- rep(floor(eta_median), length(not_obs))
#   }
#   else if(y@imputation_method == "expectation") draws <- predict(model, type = "response")[not_obs]
#   else stop("'imputation_method' not recognized")
# 
#   y@data[not_obs] <- draws
#   y@imputations[s,] <- draws
#   return(y)
# })

setMethod("mi", signature(y = "semi-continuous", model = "glm"), def = 
  function(y, model, s, ...) {
    stop("the semi-continuous mi() method should not have been called")
  })

setMethod("mi", signature(y = "nonnegative-continuous", model = "glm"), def = 
  function(y, model, s, ...) {
    draws <- .mi_continuous(y, model)
    # now account for the fact that some draws were determined to be 0 in step 1
    mark <- which(complete(y@indicator, 0L)[y@which_miss] == 1)
    if(length(mark)) draws[mark] <- y@transformation(rep(0, length(mark)))
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

## helper function
.mi_proportion <-
  function(y, model) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    
    if(!is.element(y@imputation_method,  c("ppd", "pmm"))) badHessian <- FALSE
    else if(is.null(model$vcov)) badHessian <- FALSE
    else if(!all(is.finite(model$vcov))) badHessian <- TRUE
    else {
      ev <- eigen(vcov(model), TRUE)
      badHessian <- any(ev$values <= 0)
      means <- coef(model)
      parameters <- .draw_parameters(means, ev)
#       while(!badHessian && parameters[length(parameters)] <= 0) {
#         parameters <- .draw_parameters(means, ev)
#       }
    }
    
    if(badHessian && y@imputation_method == "ppd") {
      warning(paste("predictive mean matching used for", y@variable_name,
                    "as a fallback due to Hessian error"))
      
      old_method <- y@imputation_method
      y@imputation_method <- "pmm"
      y <- mi(y, model)
      return(y@data[y@which_miss])
    }    
    else if(y@imputation_method == "ppd") {
      eta <- model$x[y@which_drawn,,drop=FALSE] %*% parameters[1:NCOL(model$x)]
      mu <- model$link$mean$linkinv(eta)
      phi <- model$link$precision$linkinv(parameters[length(parameters)]) ## FIXME: in the parameterized case
      shape1 <- mu * phi
      shape2 <- phi - shape1
      draws <- rbeta(y@n_drawn, shape1, shape2)
    }
    else if(y@imputation_method == "pmm") {
      eta <- model$x %*% parameters[-length(parameters)]
      draws <- .pmm(y, eta)[,1] #FIXME: haven't adjusted fitted values for pmm
    }
    else if(y@imputation_method == "mean") {
      mu <- predict(model)
      mu_observed <- mu[y@which_obs]
      mu_mean <- mean(mu_observed)
      draws <- rep(mu_mean, y@n_drawn)
    }
    else if(y@imputation_method == "median") {
      mu <- predict(model)
      mu_observed <- mu[y@which_obs]
      mu_median <- median(mu_observed)
      draws <- rep(mu_median, y@n_drawn)
    }
    else if(y@imputation_method == "expectation") draws <- predict(model)[y@which_drawn]
    else stop("'imputation_method' not recognized")
    return(draws)
  }

setOldClass("betareg")
setMethod("mi", signature(y = "proportion", model = "betareg"), def = 
  function(y, model, s, ...) {
    draws <- .mi_proportion(y, model)
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

setMethod("mi", signature(y = "proportion", model = "glm"), def = 
  function(y, model, s, ...) {
    draws <- .mi_continuous(y, model)
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

setMethod("mi", signature(y = "SC_proportion", model = "betareg"), def = 
  function(y, model, s, ...) {
    draws <- .mi_proportion(y, model)
    n <- y@n_total
    if(is(y@indicator, "binary")) {
      mark <- which(complete(y@indicator, 0L)[y@which_miss] == 1)
      if(any(y@raw_data == 0, na.rm = TRUE)) draws[mark] <- .5 / n
      else                                   draws[mark] <- (n - .5) / n
    }
    else {
      signs <- complete(y@indicator, 0L)[y@which_drawn]
      draws[signs < 0] <- .5 / n
      draws[signs > 0] <- (n - .5) / n
    }
    
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

## draw from overdispersed Poisson distribution
.rpois.od <- function(n, lambda, dispersion = 1) {
  if (dispersion <= 1) ans <- rpois(n, lambda)
  else {
    B <- 1/(dispersion-1)
    A <- lambda * B
    ans <- rnbinom(n, size= A , mu = lambda)
  }
  return(ans)
}

setMethod("mi", signature(y = "count", model = "glm"), def = 
  function(y, model, s, ...) {
    if(y@n_drawn == 0) stop("'impute' should not have been called because there are no missing data")
    if(y@imputation_method == "ppd") {
      ev <- eigen(vcov(model), symmetric = TRUE)
      parameters <- .draw_parameters(coef(model), ev)
      eta <- model$x[y@which_drawn,,drop=FALSE] %*% parameters
      pred <- model$family$linkinv(eta)
      draws <- .rpois.od(y@n_drawn, pred, model$dispersion)
    }
    else if(y@imputation_method == "pmm") {
      ev <- eigen(vcov(model), symmetric = TRUE)
      parameters <- .draw_parameters(coef(model), ev)
      eta <- model$x %*% parameters
      draws <- .pmm(y, eta)[,1] #FIXME: haven't adjusted fitted values for pmm
    }
    else if(y@imputation_method == "mean") {
      eta <- predict(model, type = "response")
      eta_observed <- eta[y@which_obs]
      eta_mean <- mean(eta_observed)
      draws <- rep(round(eta_mean), y@n_drawn)
    }
    else if(y@imputation_method == "median") {
      eta <- predict(model, type = "response")
      eta_observed <- eta[y@which_obs]
      eta_median <- median(eta_observed)
      draws <- rep(floor(eta_median), y@n_drawn)
    }
    else if(y@imputation_method == "expectation") draws <- round(predict(model, type = "response")[y@which_drawn])
    else stop("'imputation_method' not recognized")
    
    draws <- as.integer(draws)
    y@data[y@which_drawn] <- draws
    y@imputations[s,] <- draws
    return(y)
  })

setMethod("mi", signature(y = "irrelevant", model = "ANY"), def = 
  function(y, model, ...) {
    stop("The mi() method should not have been called on an 'irrelevant' variable")
  })

## FIXME: account for the other stuff at the bottom of the original mi.R file

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

.guess_type <-
  function(y, favor_ordered = TRUE, favor_positive = FALSE, threshold = 5,
           variable_name = deparse(substitute(y))) {
    
    if(!is.null(dim(y))) stop(paste(variable_name, ": must be a vector"))
    if(is.factor(y)) y <- factor(y) # to drop unused levels    
    values <- unique(y)
    values <- sort(values[!is.na(values)])
    len <- length(values)
    if(len == 0) {
      warning(paste(variable_name, ": cannot infer variable type when all values are NA, guessing 'irrelevant'"))
      type <- "irrelevant"
    }
    else if(len == 1)             type <- "fixed"
    else if(grepl("^[[:punct:]]", 
                  variable_name)) type <- "irrelevant"
    else if(identical("id",
                      tolower(variable_name))) type <- "irrelevant"
    else if(len == 2) {
      if(!is.numeric(values))     type <- "binary"
      else if(all(values == 
        as.integer(values))) type <- "binary"
      else if(favor_positive) {
        if(all(values > 0))       type <- "positive-continuous"
        else if(all(values >= 0)) type <- "nonnegative-continuous"
        else                      type <- "continuous"
      }
      else                        type <- "continuous"
    }
    else if(is.ts(y)) {
      if(favor_positive) {
        if(all(values > 0))       type <- "positive-continuous"
        else if(all(values >= 0)) type <- "nonnegative-continuous"
        else                      type <- "continuous"
      }
      else                        type <- "continuous"
    }
    else if(is.ordered(y))        type <-   "ordered-categorical"
    else if(is.factor(y))         type <- "unordered-categorical"
    else if(is.character(y))      type <- "unordered-categorical"
    else if(is.numeric(y)) {
      if(all(values >= 0) && 
        all(values <= 1)) {
        
        if(any(values %in% 0:1))  type <- "SC_proportion"
        else                      type <- "proportion"
      }
      else if(len <= threshold && 
        all(values == as.integer(values)))
        type <- if(favor_ordered) "ordered-categorical" else "unordered-categorical"
      else if(favor_positive) {
        if(all(values > 0))       type <- "positive-continuous"
        else if(all(values >= 0)) type <- "nonnegative-continuous"
        else                      type <- "continuous"
      }
      else                        type <- "continuous"
    }
    else stop(paste("cannot infer variable type for", variable_name))
    
    return(type)
  }

## this constructor largely supplants typecast in previous versions of library(mi)
setMethod("missing_variable", signature(y = "ANY", type = "missing"), def = 
  function(y, favor_ordered = TRUE, favor_positive = FALSE, threshold = 5,
           variable_name = deparse(substitute(y))) {
    
    type <- .guess_type(y, favor_ordered, favor_positive, threshold, variable_name)
    return(missing_variable(y = y, type = type, variable_name = variable_name))
  })

setMethod("missing_variable", signature(y = "ANY", type = "character"), def = 
  function(y, type, variable_name = deparse(substitute(y)), ...) {
    return(new(type, raw_data = y, variable_name = variable_name, ...))
  })

.show_helper <-
  function(object) {
    type <- class(object)
    missingness <- object@n_miss
    meth <- object@imputation_method
    if(object@n_miss) {
      if(is.character(object@family)) {
        fam <- object@family
        link <- NA_character_
      }
      else {
        fam <- object@family$family
        link <- object@family$link
      }
    }
    else fam <- link <- NA_character_
    if(is(object, "continuous")) trans <- .parse_trans(object@transformation)
    else trans <- NA_character_
    
    df <- data.frame(type = type, missing = missingness, method = meth, 
                     family = fam, link = link, transformation = trans)
    rownames(df) <- object@variable_name
    if(is(object, "semi-continuous")) df <- rbind(df, .show_helper(object@indicator))
    return(df)
  }

setMethod("show", signature(object = "missing_variable"), def = 
  function(object) {
    df <- .show_helper(object)
    print(df)
  })

## setAs methods cause subtle problems with auto-coercion
# setAs(from = "unordered-categorical", to = "ordered-categorical", def =
# function(from) {
#   class(from) <- "ordered-categorical"
#   return(from)
# })
# 
# setAs(from = "ordered-categorical", to = "unordered-categorical", def =
# function(from) {
#   class(from) <- "unordered-categorical"
#   return(from)
# })
# 
# setAs(from = "binary", to = "unordered-categorical", def =
# function(from) {
#   stop("not possible or necessary to coerce from 'binary' to 'unordered-categorical'")
# })


# setAs(from = "binary", to = "ordered-categorical", def =
# function(from) {
#   stop("not possible or necessary to coerce from 'binary' to 'unordered-categorical'")
# })

# setAs(from = "nonnegative-continuous", to = "continuous", def = 
# function(from) {
#   mean <- mean(from@raw_data, na.rm = TRUE)
#   sd <- sd(from@raw_data, na.rm = TRUE)
#   from@transformation <- .standardize_transform
#   formals(from@transformation)$mean <- mean
#   formals(from@transformation)$sd <- sd
#   from@inverse_transformation <- .standardize_transform
#   formals(from@inverse_transformation)$mean <- mean
#   formals(from@inverse_transformation)$sd <- sd
#   formals(from@inverse_transformation)$inverse <- TRUE
#   from@data <- from@transformation(from@raw_data)
#   class(from) <- "continuous"
#   return(from)
# })
# 
# setAs(from = "continuous", to = "positive-continuous", def = 
# function(from) {
#   from@transformation <- log
#   from@inverse_transformation <- exp
#   from@data <- from@transformation(from@raw_data)
#   class(from) <- "positive-continuous"  
#   validObject(from)
#   return(from)
# })
# 

## maybe add more methods
## NOTE: If you change something here, look also at the change_type.R file

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


## these extract parameters from an estimated object

setMethod("get_parameters", signature(object = "ANY"), def =
  function(object, ...) {
    return(c(coef(object)))
  })

setOldClass("polr")
setMethod("get_parameters", signature(object = "polr"), def =
  function(object, ...) {
    return(c(coef(object), object$zeta))
  })

setOldClass("multinom")
setMethod("get_parameters", signature(object = "multinom"), def =
  function(object, ...) {
    return(c(t(coef(object))))
  })

setMethod("get_parameters", signature(object = "missing_variable"), def = 
  function(object, latest = FALSE, ...) {
    if(latest) {
      if(is.logical(latest)) {
        mark <- !apply(object@parameters, 1, FUN = function(x) any(is.na(x)))
        mark <- mark[length(mark)]
      }
      else mark <- latest
      return(object@parameters[mark,])
    }
    else return(object@parameters)
  })

setMethod("get_parameters", signature(object = "missing_data.frame"), def = 
  function(object, latest = FALSE, ...) {
    mini_list <- lapply(object@variables, get_parameters, latest = latest, ...)
    out <- matrix(NA_real_, nrow(mini_list[[1]]), ncol = 0)
    for(i in seq_along(mini_list)) out <- cbind(out, mini_list[[i]])
    return(out)
  })

setMethod("get_parameters", signature(object = "mi"), def = 
  function(object, latest = FALSE, ...) {
    mini_list <- lapply(object@data, get_parameters, latest = latest, ...)
    dims <- dim(mini_list)
    out <- array(NA_real_, c(dims[1], length(mini_list), dims[2]), 
                 dimnames = list(NULL, NULL, colnames(mini_list[[1]])))
    for(i in 1:NCOL(out)) out[,i,] <- mini_list[[i]]
    return(out)
  })

setMethod("get_parameters", signature(object = "mi_list"), def = 
  function(object, latest = FALSE, ...) {
    lapply(object, get_parameters, latest = latest, ...)
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


## These functions extract completed data

setMethod("complete", signature(y = "missing_variable", m = "integer"), def = 
  function(y, m, ...) {
    out <- y@data
    if(m > 0 & y@n_drawn) out[y@which_drawn] <- as.numeric(y@imputations[m,])
    return(out)
  })

setMethod("complete", signature(y = "irrelevant", m = "integer"), def = 
  function(y, m, ...) {
    return(y@raw_data)
  })

setMethod("complete", signature(y = "categorical", m = "integer"), def = 
  function(y, m, to_factor = TRUE, ...) {
    out <- y@data
    if(m > 0 & y@n_drawn) out[y@which_drawn] <- as.numeric(y@imputations[m,])
    if(to_factor) {
      out <- factor(out, ordered = is(y, "ordered-categorical"))
      levels(out) <- y@levels
    }
    return(out)
  })

setMethod("complete", signature(y = "binary", m = "integer"), def = 
  function(y, m, to_factor = TRUE, ...) {
    out <- y@data
    if(m > 0 & y@n_drawn) out[y@which_drawn] <- as.numeric(y@imputations[m,])
    if(to_factor) {
      out <- factor(out, ordered = FALSE)
      levels(out) <- y@levels
    }
    return(out)
  })

setMethod("complete", signature(y = "continuous", m = "integer"), def = 
  function(y, m, transform = TRUE, ...) {
    out <- y@data
    if(m > 0 & y@n_drawn) out[y@which_drawn] <- as.numeric(y@imputations[m,])
    if(transform) out <- y@inverse_transformation(out)
    return(out)
  })

setMethod("complete", signature(y = "nonnegative-continuous", m = "integer"), def = 
  function(y, m, transform = TRUE, ...) {
    out <- y@data
    if(m > 0 & y@n_drawn) out[y@which_drawn] <- as.numeric(y@imputations[m,])
    if(transform) {
      out <- y@inverse_transformation(out)
      out[y@raw_data == 0] <- 0
    }
    return(out)
  })

setMethod("complete", signature(y = "SC_proportion", m = "integer"), def = 
  function(y, m, transform = TRUE, ...) {
    out <- y@data
    if(m > 0 & y@n_drawn) out[y@which_drawn] <- as.numeric(y@imputations[m,])
    if(transform) out <- y@inverse_transformation(out)
    out[y@raw_data == 0] <- 0
    out[y@raw_data == 1] <- 1
    return(out)
  })

setMethod("complete", signature(y = "missing_data.frame", m = "integer"), def =
  function(y, m, to_matrix = FALSE, include_missing = TRUE) {
    if(to_matrix) out <- sapply(y@variables, complete, m = m, to_factor = FALSE, transform = FALSE)
    else          out <- as.data.frame(lapply(y@variables, complete, m = m, to_factor = TRUE, transform = TRUE))
    
    if(is(y, "allcategorical_missing_data.frame")) {
      out <- cbind(out, latents = complete(y@latents, m = m, to_factor = !to_matrix))
    }
    if(include_missing) {
      M <- is.na(y)[,!sapply(y@variables, FUN = function(y) y@all_obs), drop = FALSE]
      colnames(M) <- paste("missing", colnames(M), sep = "_")
      out <- cbind(out, M)
    }
    return(out)
  })

setMethod("complete", signature(y = "mi", m = "numeric"), def = 
  function(y, m = length(y), to_matrix = FALSE, include_missing = TRUE) {
    stopifnot(m == as.integer(m))
    m <- as.integer(m)
    l <- length(y@data)
    draws <- sum(y@total_iters)
    
    if(length(m) > 1) out <- lapply(y@data[m], complete, m = 0L, to_matrix = to_matrix, include_missing = include_missing)
    else if(m == 1)   out <- complete(y@data[[1]], m = 0L, to_matrix = to_matrix, include_missing = include_missing) # not a list
    else if(m <= l)   out <- lapply(y@data[1:m], complete, m = 0L, to_matrix = to_matrix, include_missing = include_missing)
    else { # wants more completed datasets than chains
      quotient <- m %/% l
      remainder <- m %% l
      num <- quotient + (1:l <= remainder)
      out <- vector("list", m)
      count <- 1
      for(i in seq_along(y@data)) {
        if(num[i] == 1) {
          out[[count]] <- complete(y@data[[i]], m = 0L, to_matrix = to_matrix, include_missing = include_missing)
          count <- count + 1
        }
        else { # double-dip from a chain
          SEQ <- seq(from = ceiling(draws / 2), to = draws, length.out = num[i])
          temp <- sapply(SEQ, FUN = function(j) complete(y@data[[i]], m = as.integer(j), to_matrix = to_matrix, 
                                                         include_missing = include_missing), simplify = FALSE)
          for(j in seq_along(temp)) {
            out[[count]] <- temp[[j]]
            count <- count + 1
          }
        }
      }
    }
    return(out)
  })

setMethod("complete", signature(y = "mi", m = "missing"), def = 
  function(y, to_matrix = FALSE, include_missing = TRUE) {
    return(complete(y, m = length(y), to_matrix = to_matrix, include_missing = include_missing))
  })

setMethod("complete", signature(y = "mi_list", m = "numeric"), def = 
  function(y, m = length(y[[1]]), to_matrix = FALSE, include_missing = TRUE) {
    temp <- lapply(y, FUN = complete, m = m, to_matrix = to_matrix, include_missing = include_missing)
    dfs <- temp[[1]]
    if(length(m) == 1 && m == 1 && length(temp) > 1) for(i in 2:length(temp)) {
      dfs <- rbind(dfs, temp[[i]])
    }
    else if(length(temp) > 1) for(i in 2:length(temp)) for(j in 1:length(dfs)) {
      dfs[[j]] <- rbind(dfs[[j]], temp[[i]][[j]])
    }
    return(dfs)
  })

setMethod("complete", signature(y = "mi_list", m = "missing"), def = 
  function(y, to_matrix = FALSE, include_missing = TRUE) {
    return(complete(y, m = length(y[[1]]), to_matrix = to_matrix, include_missing = include_missing))
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

setMethod("change", signature(data = "missing_data.frame", y = "ANY", to = "ANY", what = "character"), def =
  function(data, y, to, what, ...) {
    if(length(what) > 1) stop("'what' must have length one")
    
    #   what <- match.arg(what, c("family", "imputation_method", "model", "size", "transformation", 
    #                             "type", "class", "link", "method"))
    if(what == "class")  what <- "type"
    if(what == "method") what <- "imputation_method"
    
    if(is.character(y) && !(what %in% c("family", "imputation_method", "link", "model", "size", "transformation", "type"))) {
      if(length(y) > 1)  stop("'y' must have length one")
      if(length(to) > 1) stop("'to' must have length one")
      if(is.logical(y) | is.numeric(y)) y <- colnames(data)[y]
      
      if(to == "unpossible") {
        mv <- data@variables[[y]]
        unpossible <- which(mv@raw_data == what)
        mv@n_unpossible <- length(unpossible)
        mv@which_unpossible <- unpossible
        
        mv@which_obs <- mv@which_obs[!(mv@which_obs %in% mv@which_unpossible)]
        mv@n_obs <- length(mv@which_obs)
        
        mv@which_miss <- mv@which_miss[!(mv@which_miss %in% mv@which_unpossible)]
        mv@n_miss <- length(mv@which_miss)
        
        data@variables[[y]] <- mv
        
        if(!length(data@weights)) {
          data@weights <- lapply(data@variables, FUN = function(y) {
            if(y@n_unpossible) {
              w <- rep(1, y@n_total)
              w[y@which_unpossible] <- 0
              return(w)
            }
            else return(NULL)
          })
        }
        else data@weights[[y]][mv@which_unpossible] <- 0
        
        return(data)
      }
      
      mv <- data@variables[[y]]
      mv@raw_data[mv@raw_data == what] <- to
      if(is.na(what) | is.na(to)) mv <- new(class(mv), raw_data = mv@raw_data, variable_name = mv@variable_name)
      data@variables[[y]] <- mv
      return(data)
    }
    
    if(what == "family") return(change_family(data = data, y = y, to = to))
    else if(what == "link") return(change_link(data = data, y = y, to = to))
    else if(what == "imputation_method") return(change_imputation_method(data = data, y = y, to = to))
    else if(what == "model") return(change_model(data = data, y = y, to = to))
    else if(what == "size") return(change_size(data = data, n = y))
    else if(what == "transformation") {
      if(missing(to)) return(change_transformation(data = data, y = y))
      else return(change_transformation(data = data, y = y, to = to, ...))
    }
    else if(what == "type") {
      if(missing(to)) return(change_type(data = data, y = y))
      else return(change_type(data = data, y = y, to = to, ...))
    }
    else stop("this should never happen")
  })

setMethod("change", signature(data = "missing_data.frame", y = "ANY", to = "numeric", what = "numeric"), def =
  function(data, y, to, what) {
    if(length(to) > 1)   stop("'to' must be a scalar")
    if(length(what) > 1) stop("'what' must be a scalar")
    if(is.logical(y) | is.numeric(y)) y <- colnames(data)[y]
    
    mv <- data@variables[[y]]
    mv@raw_data[mv@raw_data == what] <- to # NOTE: exception to "never change the raw_data slot rule"
    if(is(mv, "categorical")) {
      values <- unique(mv@raw_data)
      values <- values[!is.na(values)]
      if(length(values) == 2) mv <- new("binary", raw_data = mv@raw_data, variable_name = mv@variable_name)
    }
    else if(is.na(what) | is.na(to)) mv <- new(class(mv), raw_data = mv@raw_data, variable_name = mv@variable_name)
    else if(is(mv, "continuous")) mv@data <- mv@transformation(mv@raw_data)
    data@variables[[y]] <- mv
    return(data) ## FIXME: maybe reinitialize data?
  })

setMethod("change", signature(data = "missing_data.frame", y = "ANY", to = "logical", what = "numeric"), def =
  function(data, y, to, what) {
    change(data = data, y = y, what = what, to = as.numeric(to))
  })

setMethod("change", signature(data = "missing_data.frame", y = "ANY", to = "character", what = "numeric"), def = 
  function(data, y, to, what) {
    if(length(to) > 1)   stop("'to' must be a scalar")
    if(to != "unpossible") stop("'to' must be 'unpossible'")
    if(length(what) > 1) stop("'what' must be have length one")
    if(is.logical(y) | is.numeric(y)) y <- colnames(data)[y]
    
    mv <- data@variables[[y]]
    unpossible <- which(mv@raw_data == what)
    mv@n_unpossible <- length(unpossible)
    mv@which_unpossible <- unpossible
    
    mv@which_obs <- mv@which_obs[!(mv@which_obs %in% mv@which_unpossible)]
    mv@n_obs <- length(mv@which_obs)
    
    mv@which_miss <- mv@which_miss[!(mv@which_miss %in% mv@which_unpossible)]
    mv@n_miss <- length(mv@which_miss)
    
    data@variables[[y]] <- mv
    
    if(!length(data@weights)) {
      data@weights <- lapply(data@variables, FUN = function(y) {
        if(y@n_unpossible) {
          w <- rep(1, y@n_total)
          w[y@which_unpossible] <- 0
          return(w)
        }
        else return(NULL)
      })
    }
    else data@weights[[y]][mv@which_unpossible] <- 0
    return(data)
  })

setMethod("change", signature(data = "missing_data.frame", y = "ANY", to = "logical", what = "character"), def =
  function(data, y, to, what) {
    change(data = data, y = y, what = what, to = as.numeric(to))
  })

setMethod("change", signature(data = "mdf_list", y = "ANY", to = "ANY", what = "ANY"), def =
  function(data, y, to, what, ...) {
    out <- lapply(data, FUN = change, y = y, to = to, what = what, ...)
    class(out) <- "mdf_list"
    return(out)
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## these are convience functions that implicitly change something else by changing the model buzzword

setMethod("change_model", signature(data = "missing", y = "missing_variable", to = "character"), def = 
  function(y, to) {
    switch(to, 
           "logit" =  new("binary", variable_name = y@variable_name, raw_data = y@raw_data,
                          imputation_method = y@imputation_method, family = binomial(link = "logit")), 
           "probit" =  new("binary", variable_name = y@variable_name, raw_data = y@raw_data, 
                           imputation_method = y@imputation_method, family = binomial(link = "probit")),
           "cauchit" = new("binary", variable_name = y@variable_name, raw_data = y@raw_data, 
                           imputation_method = y@imputation_method, family = binomial(link = "cauchit")), 
           "cloglog" = new("binary", variable_name = y@variable_name, raw_data = y@raw_data, 
                           imputation_method = y@imputation_method, family = binomial(link = "cloglog")), 
           
           "qlogit" =  new("binary", variable_name = y@variable_name, raw_data = y@raw_data,
                           imputation_method = y@imputation_method, family = quasibinomial(link = "logit")), 
           "qprobit" =  new("binary", variable_name = y@variable_name, raw_data = y@raw_data, 
                            imputation_method = y@imputation_method, family = quasibinomial(link = "probit")),
           "qcauchit" = new("binary", variable_name = y@variable_name, raw_data = y@raw_data, 
                            imputation_method = y@imputation_method, family = quasibinomial(link = "cauchit")),
           "qcloglog" = new("binary", variable_name = y@variable_name, raw_data = y@raw_data, 
                            imputation_method = y@imputation_method, family = quasibinomial(link = "cloglog")),
           
           "ologit" =  new("ordered-categorical", variable_name = y@variable_name, raw_data = y@raw_data, 
                           imputation_method = y@imputation_method, family = multinomial(link = "logit")), 
           "oprobit" =  new("ordered-categorical", variable_name = y@variable_name, raw_data = y@raw_data,
                            imputation_method = y@imputation_method, family = multinomial(link = "probit")),
           "ocauchit" =  new("ordered-categorical", variable_name = y@variable_name, raw_data = y@raw_data,
                             imputation_method = y@imputation_method, family = multinomial(link = "cauchit")), 
           "ocloglog" =  new("ordered-categorical", variable_name = y@variable_name, raw_data = y@raw_data,
                             imputation_method = y@imputation_method, family = multinomial(link = "cloglog")),
           
           "mlogit" =  new("unordered-categorical", variable_name = y@variable_name, raw_data = y@raw_data,
                           imputation_method = y@imputation_method, family = multinomial(link = "logit")),
           "RNL" = new("unordered-categorical", variable_name = y@variable_name, raw_data = y@raw_data,
                       imputation_method = y@imputation_method, family = binomial(link = "logit")),
           
           "qpoisson" =  new("count", variable_name = y@variable_name, raw_data = y@raw_data, 
                             imputation_method = y@imputation_method, family = quasipoisson(link = "log")),
           "poisson" =  new("count", variable_name = y@variable_name, raw_data = y@raw_data, 
                            imputation_method = y@imputation_method, family = poisson(link = "log")),
           
           "linear" = new("continuous", variable_name = y@variable_name, raw_data = y@raw_data,
                          imputation_method = y@imputation_method, family = gaussian(link = "identity")),
           stop("model not recognized")
           )
  })

setMethod("change_model", signature(data = "missing_data.frame", y = "character", to = "character"), def =
  function(data, y, to) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    if(all(y %in% names(getClass("missing_variable")@subclasses))) {
      classes <- sapply(data@variables, class)
      y <- c(sapply(y, FUN = function(x) {
        names(classes[which(classes == x)])
      }))
      if(is.list(y)) stop(paste("no variables of class", names(y)[1]))
      to <- rep(to[1], length(y))
    }
    y <- match.arg(y, data@DIMNAMES[[2]], several.ok = TRUE)
    check <- FALSE
    for(i in 1:length(y)) {
      categorical <- is(data@variables[[y[i]]], "categorical")
      data@variables[[y[i]]] <- change_model(y = data@variables[[y[i]]], to = to[i])
      if(categorical & !is(data@variables[[y[i]]], "categorical")) check <- TRUE
      if(!categorical & is(data@variables[[y[i]]], "categorical")) check <- TRUE
    }
    if(check) return(new(class(data), variables = data@variables))
    else      return(data)
  })

setMethod("change_model", signature(data = "missing_data.frame", y = "numeric", to = "character"), def =
  function(data, y, to) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    for(i in 1:length(y)) {
      categorical <- is(data@variables[[y[i]]], "categorical")
      data@variables[[y[i]]] <- change_model(y = data@variables[[y[i]]], to = to[[i]])
      if(categorical & !is(data@variables[[y[i]]], "categorical")) check <- TRUE
      if(!categorical & is(data@variables[[y[i]]], "categorical")) check <- TRUE
    }
    if(check) return(new(class(data), variables = data@variables))
    else      return(data)
  })

setMethod("change_model", signature(data = "missing_data.frame", y = "logical", to = "character"), def =
  function(data, y, to) {
    if(length(y) != data@DIM[2]) {
      stop("the length of 'y' must equal the number of variables in 'data'")
    }
    return(change_model(data, which(y), to))
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011 Andrew Gelman
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## Function to draw from the relevant symmetric generalized beta distribution
.rgbeta <- 
  function(num, shape) {
    if(shape > 0)       -1 + 2 * rbeta(num, shape, shape)
    else if(shape == 0) -1 + 2 * rbinom(num, 1, 0.5)
    else stop("shape must be non-negative")
  }

## Function to draw a Cholesky factor of a random correlation matrix 
## as a function of canonical partial correlations (CPCs)
.rcorvine <-
  function(n_full, n_partial, n_cat, eta,
           restrictions, strong, experiment, treatment_cor, last_CPC) {
    nom <- !is.null(n_cat)
    n <- n_full + 2 * n_partial
    if(nom) {
      n <- n + sum(n_cat) - length(n_cat)
      nc <- 2 * n_partial + sum(n_cat) - length(n_cat)
      holder <- matrix(NA_real_, nrow = nc, ncol = nc)
    }
    else holder <- matrix(NA_real_, nrow = 2 * n_partial, ncol = 2 * n_partial)
    count <- 1
    
    if(eta <= 0) stop("'eta' must be positive")
    alpha <- eta + (n - 2) / 2 # if eta == 1, then tcrossprod(L) is uniform over correlation matrices
    L <- matrix(NA_real_, n, n)
    L[upper.tri(L)] <- 0
    L[1,1] <- 1
    L <- .rcorvine_helper(L, holder, n_full, n_partial, n_cat, alpha, restrictions, 
                          strong, experiment, treatment_cor, nom, last_CPC)
    if(restrictions == "MARish") L <- .MAR_opt(L, n_full, n_partial)
    return(L)
  }

.rcorvine_helper <-
  function(L, holder, n_full, n_partial, n_cat, alpha, restrictions, 
           strong, experiment, treatment_cor, nom, last_CPC) {
    n <- nrow(L)
    mark <- is.na(L[,1])
    sum_mark <- sum(mark)
    CPCs <- .rgbeta(sum(mark), alpha)
    count <- 1
    if(experiment) {
      len <- length(treatment_cor)
      if(len == 1 && treatment_cor == 0) treatment_cor <- rep(0, sum_mark)
      else if(len != sum_mark) {
        stop(paste("length of 'treatment_cor' must be", sum_mark))
      }
      treatment_mark <- is.na(treatment_cor)
      treatment_cor[treatment_mark] <- CPCs[treatment_mark]
      CPCs <- treatment_cor # treatment variable is first
    }
    if(n_full == 0 && restrictions != "none") {
      CPCs[sum_mark:(sum_mark - n_partial)] <- 0
      holder[1,mark] <- CPCs
      count <- count + 1
    }
    else if(n_full == 0) {
      holder[1,mark] <- CPCs
      count <- count + 1
    }
    L[mark,1] <- CPCs
    W <- log(1 - CPCs^2)
    
    ## NOTE: order of variables is: 
    ## all fully observed (with the treatment first if applicable)
    ## all partially observed but not nominal variables (if any)
    ## the components of the nominal variable(s) (if any)
    ## all missingness indicators
    
    # fully observed variables have arbitrary CPCs
    start <- 2
    end <- n_full
    if(n_full >= 2) for(i in start:end) {
      L[i,i] <- exp(0.5 * W[i-1])
      gap <- which(is.na(L[,i]))
      gap1 <- gap - 1
      alpha <- alpha - 0.5
      CPCs <- .rgbeta(length(gap), alpha)
      if(restrictions == "MCAR") CPCs[length(gap):(length(gap) - n_partial + 1)] <- 0
      L[gap,i] <- CPCs * exp(0.5 * W[gap1])
      W[gap1] <- W[gap1] + log(1 - CPCs^2)
    }
    
    # partially observed variables have arbitrary CPCs among themselves
    # but are conditionally uncorrelated with all missingness indicators under MAR
    # note: we condition on all fully observed variables and all previous partially observed variables
    # this triangle scheme implies that the errors when predicting the partially observed variables are uncorrelated
    if(n_full >= 2) start <- end + 1
    end <- start + n_partial - 1
    if(nom) end <- end - length(n_cat)
    if(start <= end) for(i in start:end) {
      L[i,i] <- exp(0.5 * W[i-1])
      gap <- which(is.na(L[,i]))
      gap1 <- gap - 1
      alpha <- alpha - 0.5
      CPCs <- .rgbeta(length(gap), alpha)
      if(restrictions %in% c("triangular", "stratified")) {
        CPCs[length(gap):(length(gap) - n_partial + 1)] <- 0
        if(i == end && !is.na(last_CPC)) CPCs[length(CPCs)] <- last_CPC
      }
      else if(restrictions == "MCAR") CPCs[] <- 0
      
      L[gap,i] <- CPCs * exp(0.5 * W[gap1])
      W[gap1] <- W[gap1] + log(1 - CPCs^2)
      holder[count,(count+1):ncol(holder)] <- CPCs
      count <- count + 1
    }
    
    # if there are nominal partially observed variables, make the category residuals uncorrelated (MNL assumption)
    if(nom) {
      #if(n_full >= 2) 
      start <- end + 1
      end <- start + sum(n_cat) - 1
      for(i in start:end) {
        L[i,i] <- exp(0.5 * W[i-1])
        gap <- which(is.na(L[,i]))
        gap1 <- gap - 1
        alpha <- alpha - 0.5
        if(restrictions != "none") CPCs <- rep(0, length(gap))
        else {
          CPCs <- .rgbeta(length(gap), alpha)
          CPCs[-(length(gap):(length(gap) - n_partial + 1))] <- 0
        }
        L[gap,i] <- CPCs * exp(0.5 * W[gap1])
        W[gap1] <- W[gap1] + log(1 - CPCs^2)
        holder[count,(count+1):ncol(holder)] <- CPCs
        count <- count + 1
      }
    }
    
    # missingness indicators can be constructed to be instruments if MAR holds whose strength can be manipulated
    if(n_partial > 1) {
      start <- end + 1
      end <- n - 1
      count <- if(n_full > 0) 1 else 2
      if(start <= end) for(i in start:end) {
        L[i,i] <- exp(0.5 * W[i-1])
        gap <- which(is.na(L[,i]))
        gap1 <- gap - 1
        alpha <- alpha - 0.5
        if(restrictions %in% c("none", "MARish"))          CPCs <- .rgbeta(length(gap), alpha)
        else if(restrictions %in% c("stratified", "MCAR")) CPCs <- rep(0, length(gap))
        else if(strong == 2) CPCs <- holder[count,(count+1):(count + length(gap))]
        else if(strong == 1) CPCs <- .rgbeta(length(gap), alpha)
        else if(strong == 0) CPCs <- rep(0, length(gap))
        L[gap,i] <- CPCs * exp(0.5 * W[gap1])
        W[gap1] <- W[gap1] + log(1 - CPCs^2)
      }
    }
    
    L[n,n] <- exp(0.5 * W[n-1])
    return(L)
  }

## Function to draw a Cholesky factor of a random correlation matrix 
## as a function of canonical partial correlations (CPCs)
.rcorvine_partial <-
  function(Sigma, n_partial, n_cat, eta, restrictions, strong, experiment, treatment_cor) {
    
    n <- nrow(Sigma)
    n_full <- n - n_partial
    ldlt <- LDLt(Sigma)
    U <- t(ldlt$L)
    holder <- matrix(NA_real_, n, n)
    holder[1,-1] <- U[1,-1]
    W <- c(NA_real_, 1 - holder[1,-1]^2)
    for(i in 2:(n-1)) {
      denominator <- W[i]
      gap <- (i+1):n
      temp <- U[i,gap] / sqrt(W[gap] / denominator)
      invalid <- is.na(temp)
      temp[invalid] <- sign(U[i,gap][invalid])
      invalid <- abs(temp) > 1
      temp[invalid] <- sign(temp[invalid])
      holder[i,gap] <- temp
      W[gap] <- W[gap] * (1 - holder[i,gap]^2)
    }
    
    nom <- !is.null(n_cat)
    n <- n + n_partial
    if(nom) {
      n <- n + sum(n_cat) - length(n_cat)
    }
    
    L <- t(U) * sqrt(diag(ldlt$D))
    if(eta <= 0) stop("'eta' must be positive")
    diff <- n - nrow(L)
    alpha <- eta + diff / 2
    
    L <- cbind(L, matrix(0, nrow(L), diff))
    L <- rbind(L, matrix(NA_real_, diff, n))
    L[upper.tri(L)] <- 0
    
    holder <- matrix(NA_real_, nrow = diff, ncol = ncol(L))
    L <- .rcorvine_helper(L, holder, n_full, n_partial, n_cat, alpha, restrictions, strong, experiment, treatment_cor, nom)
    if(restrictions == "MARish") L <- .MAR_opt(L, n_full, n_partial)
    return(L)
    
  }

.MAR_opt <-
  function(L, n_full, n_partial) {
    n_p2 <- n_partial^2
    lowers <- lower.tri(L)
    cell_mark <- tail(which(lowers), n_p2)
    lowers[] <- FALSE
    lowers[cell_mark] <- TRUE
    row_mark <- which(apply(lowers, 1, any))
    diag(L)[row_mark] <- NA_real_
    partials <- (n_full + 1):(n_full + n_partial)
    missingness <- nrow(L):(nrow(L) - n_partial + 1)
    block_mark <- which( row(L) %in% partials & col(L) %in% missingness )
    
    foo <- function(theta) {
      L[cell_mark] <- theta
      diags <- 1 - rowSums(L[row_mark,,drop=FALSE]^2, na.rm = TRUE)
      if(any(diags < 0)) return(NA_real_)
      diag(L)[row_mark] <- sqrt(diags)
      Sigma_inv <- chol2inv(t(L))
      return(c(crossprod(Sigma_inv[block_mark])))
    }
    
    opt <- optim(L[cell_mark], foo, method = "BFGS")
    L[cell_mark] <- opt$par
    diag(L)[row_mark] <- sqrt(1 - rowSums(L[row_mark,,drop=FALSE]^2, na.rm = TRUE))
    return(L)
  }

.NMARness <-
  function(L) {
    xs <- grep("^x_", rownames(L), value = TRUE)
    ys <- grep("^y_", rownames(L), value = TRUE)
    us <- grep("^u_", rownames(L), value = TRUE)
    sapply(ys, FUN = function(y) {
      i <- y
      sapply(us, FUN = function(u) {
        j <- u
        cons <- c(xs, setdiff(us, u))
        mark <- c(i,j,cons) 
        D_ijcons <- det(tcrossprod(L[mark,,drop = FALSE]))
        mark <- cons
        D_cons <- det(tcrossprod(L[mark,,drop = FALSE]))
        mark <- c(i,cons)
        D_icons <- det(tcrossprod(L[mark,,drop = FALSE]))
        mark <- c(j,cons)
        D_jcons <- det(tcrossprod(L[mark,,drop = FALSE]))
        return(1 - D_ijcons * D_cons / (D_icons * D_jcons))
      })
    })
  }

## Function to construct a random data.frame with tunable missingness
rdata.frame <-
  function(N = 1000, restrictions = c("none", "MARish", "triangular", "stratified", "MCAR"), 
           last_CPC = NA_real_, strong = FALSE, pr_miss = .25, 
           Sigma = NULL, alpha = NULL, 
           experiment = FALSE, treatment_cor = c(rep(0, n_full - 1), rep(NA, 2 * n_partial)),
           n_full = 1, n_partial = 1, n_cat = NULL,
           eta = 1, df = Inf, types = "continuous", estimate_CPCs = TRUE) {
    
    if(length(N) != 1) stop("length of 'N' must be 1")
    if(N <= 0) stop("'N' must be positive")
    
    restrictions <- match.arg(restrictions)
    if(strong && restrictions == "none") warning("instruments are not valid unless the MAR assumption is enforced")
    
    if(n_full < 0) stop("'n_full' must be >= 0")
    if(n_partial < 0) stop("'n_partial must be >= 0")
    n <- n_partial + n_full
    if(n == 0) stop("at least one of 'n_full' or 'n_partial' must be positive")
    
    if(length(pr_miss) == 1) pr_miss <- rep(pr_miss, n_partial)
    
    if(any(pr_miss <= 0)) stop("all elements of 'pr_miss' must be > 0")
    if(any(pr_miss >= 1)) stop("all elements of 'pr_miss' must be < 1")
    
    if(length(df) != 1) stop("'df' must be of length 1")
    if(df <= 0) stop("'df' must be a positive")
    
    if(length(types) == 1) types <- rep(types, n)
    types <- match.arg(types, c("continuous", "count", "binary", "treatment", "ordinal", "nominal", "proportion", "positive"), 
                       several.ok = TRUE)
    if(any(types[1:n_full] == "nominal")) {
      warning("fully observed nominal variables not supported, changing them to ordinal without loss of generality")
      types <- ifelse(types == "nominal" & 1:length(types) <= n_full, "ordinal", types)
    }
#     else if(!is.null(n_cat)) types[n:(n - length(n_cat) + 1)] <- "nominal"
    if(all( c("ordinal", "nominal") %in% types[-(1:n_full)] )) {
      stop("including both ordinal and nominal partially observed variables is not supported yet")
    }
    
    if(any(types == "nominal")) {
      has_nominal <- TRUE
      if(is.null(n_cat)) {
        if(types[n] != "nominal") {
          warning("assuming the last partially observed variable is nominal with 3 categories")
          types[n] <- "nominal"
        }
        n_cat <- 3
      }
    }
    else has_nominal <- FALSE
    
    if(has_nominal) {
      if(any(n_cat < 3)) stop("nominal variables must have more than 2 categories")
      types <- c(types[types != "nominal"], types[types == "nominal"])
    }
    
    if(experiment) {
      if(types[1] != "treatment") stop("the first variable must be the treatment variable")
      if(any(types[-1] == "treatment")) stop("only one treatment variable is permitted")
    }
    
    if(is.null(Sigma)) L <- .rcorvine(n_full, n_partial, 
                                      if(has_nominal) n_cat else NULL, 
                                      eta, restrictions, strong, experiment, treatment_cor, last_CPC)
    else {
      if(!isSymmetric(Sigma)) stop("'Sigma' must be symmetric")
      if(ncol(Sigma) != (n_full + 2 * n_partial)) stop("'Sigma' must be of order 'n_full + 2 * n_partial'")
      if(any(types == "nominal")) stop("nominal variables not supported when 'Sigma' is given")
      if(experiment) stop("treatment variables not supported when 'Sigma' is given")
      L <- chol(Sigma)
    }
    
    if(is.null(alpha)) {
      Z <- matrix(rnorm(N * nrow(L)), nrow = nrow(L))
      X <- as.data.frame(t(Z) %*% t(L))
    }
    else {
      if(length(alpha) == 1 && is.na(alpha)) alpha <- rt(ncol(L), df)
      else if(length(alpha) != ncol(L)) stop(paste("length of alpha must be", ncol(L)))
      Sigma <- tcrossprod(L)
      result <- find_Omega(Sigma, alpha, control = list(maxit = 1000))
      X <- as.data.frame(sn::rmsn(N, Omega = result$Omega, alpha = alpha))
    }
    if(df < Inf) X <- X / sqrt(rchisq(N, df) / df)
    if(!has_nominal) colnames(X) <- c(if(n_full)    paste("x", 1:n_full,    sep = "_"),
                                      if(n_partial) paste("y", 1:n_partial, sep = "_"),
                                      if(n_partial) paste("u", 1:n_partial, sep = "_") )
    else {
      if(length(n_cat) > 23) stop("number of nominal variables must be <= 23")
      cn <- as.character(NULL)
      for(i in seq_along(n_cat)) cn <- c(cn, paste(letters[i], 1:n_cat[i], sep = "_"))
      colnames(X) <- c(if(n_full) paste("x", 1:n_full, sep = "_"), if(n_partial > length(n_cat)) 
        paste("y", 1:(n_partial - length(n_cat)), sep = "_") else NULL,
                       cn,
                       paste("u", 1:n_partial, sep = "_") )
    }
    
    if(experiment) {
      row_mark <- X[,1] == 1
      col_mark <- c(FALSE, is.na(treatment_cor))
      col_mark[grepl("^u_", colnames(X))] <- FALSE
      if(any(col_mark)) X[row_mark,col_mark] <- X[row_mark,col_mark] + 1 # ATT
    }
    X_obs <- X
    correlations <- rep(NA_real_, if(!has_nominal) n_partial else n_partial - length(n_cat) + sum(n_cat))
    end <- n_partial - length(n_cat) * has_nominal
    if(end > 0) for(i in 1:end) {
      y_var <- paste("y", i, sep = "_")
      u_var <- paste("u", i, sep = "_")
      X_obs[X[,u_var] < quantile(X[,u_var], probs = pr_miss[i]), y_var] <- NA_real_
      X_obs[[u_var]] <- NULL
      if(!estimate_CPCs) next  
      f_miss <- colnames(X)
      if(n_full > 0) f_miss <- f_miss[1:(n_full + i - 1)]
      else f_miss <- "1"
      f_miss <- paste(f_miss, collapse = " + ")
      f_miss <- as.formula(paste(u_var, "~", f_miss))
      ols_u <- lm(f_miss, data = X)
      
      f_true <- colnames(X)
      if(n_full > 0) f_true <- f_true[1:(n_full + i - 1)]
      else f_true <- "1"
      f_true <- paste(f_true, collapse = " + ")
      f_true <- as.formula(paste(y_var, "~", f_true))
      ols_y <- lm(f_true, data = X)
      
      correlations[i] <- cor(residuals(ols_u), residuals(ols_y)) # this differs only randomly from 0 under MAR due to finite N
    }
    letter_mark <- 1
    if(has_nominal) for(i in (end + 1):n_partial) {
      y_var <- paste("y", i, sep = "_")
      u_var <- paste("u", i, sep = "_")
      mark <- grepl(paste("^", letters[letter_mark], "_", sep = ""), colnames(X))
      lev <- as.character(NULL)
      for(j in 1:ceiling(n_cat[letter_mark] / 26)) lev <- c(lev, rep(letters, each = j))
      lev <- lev[1:n_cat[letter_mark]]
      X_obs[[y_var]] <- X[[y_var]] <- factor(max.col(X[,mark]), labels = lev)
      X_obs[X[,u_var] < quantile(X[,u_var], probs = pr_miss[i]), y_var] <- NA
      if(!estimate_CPCs) {
        letter_mark <- letter_mark + 1
        next
      }
      f_miss <- colnames(X)
      if(letter_mark == 1) f_miss <- f_miss[1:(n_full + n_partial - length(n_cat))]
      else f_miss <- f_miss[1:(n_full + n_partial - length(n_cat) + sum(n_cat[1:(letter_mark - 1)]))]
      f_miss <- paste(f_miss, collapse = " + ")
      f_miss <- as.formula(paste(u_var, "~", f_miss))
      ols_u <- lm(f_miss, data = X)
      
      for(j in 1:n_cat[letter_mark]) {
        f_true <- colnames(X)
        if(letter_mark == 1) f_true <- f_true[1:(n_full + n_partial - length(n_cat))]
        else f_true <- f_true[1:(n_full + n_partial - length(n_cat) + sum(n_cat[1:(letter_mark - 1)]))]
        f_true <- paste(f_true, collapse = " + ")
        n_var <- paste(letters[letter_mark], j, sep = "_")
        f_true <- as.formula(paste(n_var, "~", f_true))
        ols_n <- lm(f_true, data = X)
        correlations[which(is.na(correlations))[1]] <- cor(residuals(ols_u), residuals(ols_n)) # this differs only randomly from 0 under MAR
      }
      letter_mark <- letter_mark + 1
    }
    if(!has_nominal) names(correlations) <- if(n_partial) paste("e", 1:n_partial, sep = "_") else NULL
    else {
      cn <- if(n_partial > length(n_cat)) paste("e", 1:(n_partial - length(n_cat)), sep = "_") else as.character(NULL)
      for(i in seq_along(n_cat)) cn <- c(cn, paste("e:", letters[i], "_", 1:n_cat[i], sep = ""))
      names(correlations) <- cn
    }
    
    X_obs <- X_obs[,grepl("^[xy]_", colnames(X_obs))]
    mark_ord <- 1
    for(i in seq_along(types)) {
      mark <- is.na(X_obs[,i])
      if(types[i] %in% c("binary", "treatment")) {
        if(i == 1 && experiment) {
          X_obs[,i] <- X[,i] <- as.factor(X[,i] > 0)
          colnames(X_obs)[1] <- colnames(X)[1] <- "treatment"
        }
        else {
          X[[toupper(colnames(X)[i])]] <- X[,i]
          X_obs[,i] <- X[,i] <- cut(X[,i], breaks = 2, labels = c("FALSE", "TRUE"))
        }
      }
      else if(types[i] == "ordinal") {
        X[[toupper(colnames(X)[i])]] <- X[,i]
        breaks <- 3 
        if(length(n_cat) == 1) breaks <- n_cat 
        else if(length(n_cat) > 1) {
          breaks <- n_cat[mark_ord]
          mark_ord <- mark_ord + 1
        }
        qs <- quantile(X[,i], prob = seq(from = 0, to = 1, length.out = breaks + 1))
        qs[1] <- -Inf
        qs[length(qs)] <- Inf
        X_obs[,i] <- X[,i] <- cut(X[,i], breaks = qs, ordered_result = TRUE, labels = LETTERS[1:breaks])
      }
      else if(types[i] == "count") { # this is not quite consistent with the DGP
        X[[toupper(colnames(X)[i])]] <- X[,i]
        X_obs[,i] <- X[,i] <- as.integer(qpois(pt(X[,i], df = df), lambda = 5))
      }
      else if(types[i] == "proportion") { # this is not quite consistent with the DGP
        X[[toupper(colnames(X)[i])]] <- X[,i]
        X_obs[,i] <- X[,i] <- pt(X[,i], df = df)
      }
      else if(types[i] == "positive") {
        X[[toupper(colnames(X)[i])]] <- X[,i]
        X_obs[,i] <- X[,i] <- exp(X[,i])
      }
      X_obs[mark,i] <- NA
    }
    ord <- c(colnames(X_obs), grep("^u_", colnames(X), value = TRUE))
    extras <- colnames(X)
    extras <- extras[!(extras %in% ord)]
    ord <- c(ord, extras)
    X <- X[,ord]
    cn <- colnames(X)
    cn <- cn[sapply(1:ncol(X), FUN = function(i) {
      !is.factor(X[,i]) && !(toupper(cn[i]) %in% cn[-i])
    })]
    resort <- function(s) {
      ord <- order(as.integer(gsub("^[a-z,A-Z]_", "", s)))
      return(s[ord])
    }
    cn <- c(if(experiment) "treatment_propensity",
            resort(grep("^x", cn, ignore.case = TRUE, value = TRUE)),
            resort(grep("^y", cn, ignore.case = TRUE, value = TRUE)),
            grep("^[a-t]_", cn, ignore.case = FALSE, value = TRUE),
            grep("^u", cn, ignore.case = FALSE, value = TRUE))
    rownames(L) <- colnames(L) <- cn
    out <- list(true = X, obs = X_obs, empirical_CPCs = correlations, L = L)
    if(!is.null(alpha)) out <- c(out, list(alpha = alpha, skewness = result$sn_skewness, kurtosis = result$sn_kurtosis))
    return(out)
  }

## this function makes a positive definite correlation matrix given choose(n,2) unbounded parameters
make_O.cor <-
  function(theta) {
    n <- (1 + sqrt(1 + 8 * length(theta))) / 2
    CPCs <- exp(2 * theta)
    CPCs <- (CPCs - 1) / (CPCs + 1)
    
    L <- matrix(0, n, n)
    L[1,1] <- 1
    
    start <- 1
    end <- n - 1
    L[-1,1] <- partials <- CPCs[start:end]
    W <- log(1 - partials^2)
    for(i in 2:(n-1)) {
      start <- end + 1
      end <- start + n - i - 1
      gap <- (i+1):n
      gap1 <- i:(n-1)
      partials <- CPCs[start:end]
      L[i,i] <- exp(0.5 * W[i-1])
      L[gap,i] <- partials * exp(0.5 * W[gap1])
      W[gap1] <- W[gap1] + log(1 - partials^2)
    }
    L[n,n] <- exp(0.5 * W[n-1])
    return(tcrossprod(L))
  }

## this objective function is the Frobenius norm of the difference between Sigma and Sigma_proposed
fmin <-
  function(theta, Sigma, alpha, final = FALSE, ...) {
    n <- nrow(Sigma)
    omega <- exp(theta[1:n]) # standard deviations of the implicit Omega matrix
    O.cor <- make_O.cor(theta[-(1:n)])
    alphaTO.cor <- alpha %*% O.cor
    Sigma_proposed <- ( O.cor - 2 / (pi * c(1 + alphaTO.cor %*% alpha)) * crossprod(alphaTO.cor) ) * tcrossprod(omega)
    if(final) return(Sigma_proposed)
    return(crossprod( c(Sigma - Sigma_proposed) )[1])
  }

## this function makes a 3-factor Cholesky factorization of a PSD A matrix
LDLt <-
  function(A) {
    n <- nrow(A)
    L <- diag(n)
    D <- matrix(0, n, n)
    for(j in 1:n) {
      s <- 0
      if(j > 1) for(k in 1:(j-1)) s <- s + L[j,k]^2 * D[k,k]
      D[j,j] <- A[j,j] - s
      if(D[j,j] < 1e-15) {
        D[j,j] <- 0
        break
      }
      if(j < n) for(i in (j+1):n) {
        s <- 0
        if(j > 1) for(k in 1:(j-1)) s <- s + L[i,k] * L[j,k] * D[k,k]
        L[i,j] <- (A[i,j] - s) / D[j,j]
      }
    }
    return(list(L = L, D = D))
  }

## this function makes plausible starting values (basically treating alpha is if it were a zero vector)
make_start <-
  function(Sigma) {
    log_omega <- log(sqrt(diag(Sigma)))
    Sigma <- cov2cor(Sigma)
    n <- nrow(Sigma)
    U <- t(LDLt(Sigma)$L)
    holder <- matrix(NA_real_, n, n)
    holder[1,-1] <- U[1,-1]
    W <- c(NA_real_, 1 - holder[1,-1]^2)
    for(i in 2:(n-1)) {
      denominator <- W[i]
      gap <- (i+1):n
      temp <- U[i,gap] / sqrt(W[gap] / denominator)
      invalid <- is.na(temp)
      temp[invalid] <- sign(U[i,gap][invalid])
      invalid <- abs(temp) > 1
      temp[invalid] <- sign(temp[invalid])
      holder[i,gap] <- temp
      W[gap] <- W[gap] * (1 - holder[i,gap]^2)
    }
    holder <- t(holder)
    CPCs <- holder[lower.tri(holder)]
    return(c(log_omega, atanh(CPCs)))
  }

## this function finds Omega via optim() and returns it as part of a list with 
find_Omega <-
  function(Sigma, alpha, method = "BFGS", start = make_start(Sigma), ...) {
    stopifnot(isSymmetric(Sigma)) # Sigma is the intended covariance matrix of the multivariate skew-normal variable
    stopifnot(all(eigen(Sigma, TRUE, TRUE)$values > 0))
    n <- nrow(Sigma)
    alpha <- c(alpha)
    stopifnot(length(alpha) == n) # alpha is a shape parameter for the multivariate skew-normal variable
    
    opt <- optim(start, fmin, method = method, Sigma = Sigma, alpha = alpha, ...)
    if(opt$convergence != 0) {
      gradients <- opt$counts["gradient"]
      warning(paste("Convergence problem. Pass something like 'control = list(maxit = ", 
                    5 * gradients, ")' if alpha is far from a zero vector", sep = ""))
    }
    
    theta <- opt$par
    omega <- exp(theta[1:n])
    O.cor <- make_O.cor(theta[-(1:n)])
    opt$Omega <- O.cor * tcrossprod(omega)
    alphaTO.cor <- c(alpha %*% O.cor)
    delta <- c( (O.cor %*% alpha) / sqrt(1 + alphaTO.cor %*% alpha)[1] )
    mu_z <- sqrt(2/pi) * delta
    num <- c( mu_z %*% chol2inv(chol(O.cor)) %*% mu_z )
    opt$delta <- delta
    opt$sn_skewness <- ( (4 - pi) / 4 )^2 * ( num / (1 - num) )^3
    opt$sn_kurtosis <- 2 * (pi - 3)       * ( num / (1 - num) )^2
    return(opt)
  }

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## Some S3 methods for convenience

as.double.missing_variable <-
  function(x, ...) {
    stop("you must write an 'as.double' method for the", class(x), "class")
  }

as.double.categorical <-
  function(x, ...) {
    x@data
  }

as.double.continuous <-
  function(x, transformed = TRUE, ...) {
    if(transformed) x@data
    else x@inverse_transformation(x@data)
  }

as.double.count <-
  function(x, ...) {
    x@data
  }

as.double.irrelevant <-
  function(x, ...) {
    as.double(x@raw_data)
  }

as.double.missing_data.frame <-
  function(x, transformed = TRUE, ...) {
    sapply(x@variables, as.double, transformed = transformed)
  }

as.data.frame.missing_data.frame <- 
  function(x, row.names = NULL, optional = FALSE, ...) {
    as.data.frame(lapply(x@variables, FUN = function(y) y@raw_data),
                  row.names = if(is.null(row.names)) rownames(x) else row.names)
  }

dim.missing_data.frame <-
  function(x) {
    x@DIM
  }

dimnames.missing_data.frame <-
  function(x) {
    x@DIMNAMES
  }

names.missing_data.frame <-
  function(x) {
    x@DIMNAMES[[2]]
  }

dim.mi <-
  function(x) {
    if(isS4(x)) x@data[[1]]@DIM else {
    	class(x) <- "list"
    	return(dim(x))
    }
  }

dimnames.mi <-
  function(x) {
    if(isS4(x)) x@data[[1]]@DIMNAMES else {
    	class(x) <- "list"
    	return(dimnames(x))
    }
  }

names.mi <-
  function(x) {
    if(isS4(x)) x@data[[1]]@DIMNAMES[[2]] else {
    	class(x) <- "list"
    	return(names(x))
    }
  }

is.na.missing_variable <-
  function(x) {
    out <- rep(FALSE, x@n_total)
    out[x@which_miss] <- TRUE
    return(out)
  }

is.na.missing_data.frame <-
  function(x) {
    sapply(x@variables, is.na)
  }

is.na.mi <-
  function(x) {
    if(isS4(x)) is.na(x@data[[1]]) else {
    	class(x) <- "list"
    	return(is.na(x))
    }
  }

length.missing_variable <-
  function(x) {
    x@n_total
  }

length.missing_data.frame <-
  function(x) {
    ncol(x)
  }

length.mi <-
  function(x) {
    if(isS4(x)) length(x@data) else {
    	class(x) <- "list"
    	return(length(x))
    }
  }

print.mdf_list <-
  function(x ,...) {
    show(x)
  }

print.mi_list <-
  function(x, ...) {
    show(x)
  }

"[.missing_data.frame" <-
  function(x, i, j, drop = if (missing(i)) TRUE else length(j) == 1) {
    if(!missing(i)) {
      cdf <- complete(x, m = 0L)
      if(!missing(j)) return(cdf[i,j,drop = drop])
      else return(cdf[i,,drop = drop])
    }
    else if(length(j) > 1) return(new(class(x), variables = x@variables[j]))
    else if(is.numeric(j) && j < 0) return(new(class(x), variables = x@variables[j]))
    else return(x@variables[[j]])
  }

"[<-.missing_data.frame" <-
  function (x, i, j, value) {
    if(!missing(i)) {
      if(!missing(j)) x@variables[[j]]@raw_data[i,] <- value
      else stop("a variable (column) must be specified when replacing")
    }
    else if(is.null(value)) x@variables[j] <- value
    else if(is(value, "missing_variable")) x@variables[[j]] <- value
    else stop("replacement must be 'NULL' or a 'missing_variable'")
    return(new(class(x), variables = x@variables))
  }

"[[.missing_data.frame" <-
  function(x, ..., exact = TRUE) {
    return(x[,...])
  }

"[[<-.missing_data.frame" <-
  function (x, i, j, value) {
    if(missing(j)) x[,i] <- value
    else x[i,j] <- value
    return(x)
  }

"$.missing_data.frame" <-
  function(x, name) {
    return(x[,name])
  }


"$<-.missing_data.frame" <-
  function(x, name, value) { # this never gets dispatched for some reason
    x[,name] <- value
    return(x)
  }

stopifnot(require(mi))

rdf <- rdata.frame(N = 100, n_partial = 2, n_full = 2)
mdf <- missing_data.frame(rdf$obs)

rdf <- rdata.frame(N = 100, n_partial = 6, n_full = 1, 
                   types = c("ordinal", "cont", "count", "binary",
                             "proportion", "positive", "nominal"))
mdf <- missing_data.frame(rdf$obs)
mdf <- missing_data.frame(rdf$obs, favor_positive = TRUE)

rdf <- rdata.frame(N = 100, n_partial = 5, n_full = 1, experiment = TRUE,
                   types = c("treatment", "cont", "count", "binary",
                             "proportion", "positive"))
mdf <- missing_data.frame(rdf$obs, subclass = "experiment", concept = 
                          as.factor(c("treatment", rep("covariate", 4), "outcome")))

rdf <- rdata.frame(N = 100, n_partial = 5, n_full = 0, types = "ordinal")
mdf <- missing_data.frame(rdf$obs, subclass = "allcategorical")

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## these coerce a missing_variable to a different type of missing_variable

setMethod("change_type", signature(data = "missing", y = "missing_variable", to = "character"), def = 
  function(y, to, ...) {
    to <- match.arg(to, names(getClass("missing_variable")@subclasses))
    if(to %in% c("ordered-categorical", "binary")) raw <- as.ordered(y@raw_data)
    else if(to == "unordered-categorical")         raw <- factor(y@raw_data, ordered = FALSE)
    else raw <- as.numeric(y@raw_data)
    
    vals <- unique(raw)
    vals <- vals[!is.na(vals)]
    if(length(vals) <= 1) {
      warning(paste(y@variable_name, ": cannot change type because only one unique value"))
      return(y)
    }
    else return(new(to, variable_name = y@variable_name, raw_data = raw, imputation_method = y@imputation_method, ...))
  })

setMethod("change_type", signature(data = "missing", y = "missing_variable", to = "missing"), def = 
  function(y, to) {
    classes <- .possible_missing_variable(y@raw_data)
    classes <- names(classes[classes])
    cat("Likely choices include:", classes, sep = "\n")
    return(invisible(NULL))
  })

setMethod("change_type", signature(data = "missing_data.frame", y = "character", to = "missing"), def =
  function(data, y, to) {
    if(all(y %in% names(getClass("missing_variable")@subclasses))) {
      classes <- sapply(data@variables, class)
      y <- c(sapply(y, FUN = function(x) {
        names(classes[which(classes == x)])
      }))
      if(is.list(y)) stop(paste("no variables of class", names(y)[1]))
      else y <- y[1]
    }
    y <- match.arg(y, data@DIMNAMES[[2]], several.ok = TRUE)
    for(i in 1:length(y)) change_type(y = data@variables[[y[i]]])
    return(data)
  })

setMethod("change_type", signature(data = "missing_data.frame", y = "character", to = "character"), def =
  function(data, y, to, ...) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    if(all(y %in% names(getClass("missing_variable")@subclasses))) {
      classes <- sapply(data@variables, class)
      y <- c(sapply(y, FUN = function(x) {
        names(classes[which(classes == x)])
      }))
      if(is.list(y)) stop(paste("no variables of class", names(y)[1]))
      to <- rep(to[1], length(y))
    }
    y <- match.arg(y, data@DIMNAMES[[2]], several.ok = TRUE)
    for(i in 1:length(y)) {
      data@variables[[y[i]]] <- change_type(y = data@variables[[y[i]]], to = to[i], ...)
      data@variables[[y[i]]]@variable_name = y[i]
    }
    return(new(class(data), variables = data@variables))
  })

setMethod("change_type", signature(data = "missing_data.frame", y = "numeric", to = "character"), def =
  function(data, y, to, ...) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    for(i in 1:length(y)) {
      data@variables[[y]] <- change_type(y = data@variables[[y]], to = to[[i]], ...)
    }
    return(new(class(data), variables = data@variables))
  })

setMethod("change_type", signature(data = "missing_data.frame", y = "logical", to = "character"), def =
  function(data, y, to, ...) {
    if(length(y) != data@DIM[2]) {
      stop("the length of 'y' must equal the number of variables in 'data'")
    }
    return(change_type(data, which(y), to, ...))
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## these change the imputation method

setMethod("change_imputation_method", signature(data = "missing", y = "missing_variable", to = "character"), def = 
  function(y, to) {
    to <- match.arg(tolower(to) , getClass(class(y))@prototype@imputation_method)
    y@imputation_method <- to
    validObject(y, complete = TRUE)
    return(y)
  })

setMethod("change_imputation_method", signature(data = "missing", y = "missing_variable", to = "missing"), def = 
  function(y, to) {
    cat("Possible methods include:", getClass(class(y))@prototype@imputation_method, sep = "\n")
    return(invisible(NULL))
  })

setMethod("change_imputation_method", signature(data = "missing_data.frame", y = "character", to = "character"), def =
  function(data, y, to) {
    if(all(y %in% c("missing_variable", names(getClass("missing_variable")@subclasses)))) {
      mark <- sapply(colnames(data), FUN = function(x) {
        if(data@variables[[x]]@all_obs) return(FALSE)
        is(data@variables[[x]], y)
      })
      if(!any(mark)) stop(paste("no variables with missingness have class", y))
      else y <- names(mark)[mark]
    }
    y <- match.arg(y, colnames(data), several.ok = TRUE)
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    for(i in 1:length(y)) {
      data@variables[[y[i]]] <- change_imputation_method(y = data@variables[[y[i]]], to = to[i])
    }
    return(data)
  })

setMethod("change_imputation_method", signature(data = "missing_data.frame", y = "numeric", to = "character"), def =
  function(data, y, to) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    for(i in 1:length(y)) {
      data@variables[[y]] <- change_imputation_method(y = data@variables[[y]], to = to[i])
    }
    return(data)
  })

setMethod("change_imputation_method", signature(data = "missing_data.frame", y = "logical", to = "character"), def =
  function(data, y, to) {
    if(length(y) != data@DIM[2]) {
      stop("the length of 'y' must equal the number of variables in 'data'")
    }
    return(change_imputation_method(data, which(y), to))
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# Copyright (C) 1995-2012 The R Core Team
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## NOTE: If you change something here, also update the UML graph thingie

setClassUnion("MatrixTypeThing", c("matrix"))
setOldClass("family")
suppressWarnings(setClassUnion("WeAreFamily", c("family", "character"))) # arm + lme4 = warnings
setOldClass("mi_list")
setOldClass("mdf_list")

.known_imputation_methods <- c("ppd", "pmm", "mean", "median", "expectation", "mode", "mcar", NA_character_)
.known_families <- c("binomial", "gaussian", "Gamma", "inverse.gaussian", "poisson",
                     "quasibinomial", "quasipoisson") # "quasi" is not supported at the moment (FIXME)
.known_links <- c("logit", "probit", "cauchit", "log", "cloglog", # for binomial()
                  "identity", "inverse",                          # for gaussian() plus "log", 
#                 "inverse", "identity", "log",                   # for Gamma()
                  "sqrt",                                         # for poisson() plus "log", and "identity",
                  "1/mu^2")                                       # for inverse.gaussian() plus "inverse", "identity" and "log"


# An important class in library(mi) is the missing_variable class, which is a virtual class
# for a variable that may (or may not) have missingness. The usual types of variables that
# we are interested in imputing all inherit (perhaps indirectly) from the missing_variable
# superclass, e.g. continuous, binary, etc. In principle, these class definitions should 
# provide ALL the necessary information for that variable, like the extent of its missingness
# and how the missing values will be (or have been) imputed. Thus, in principle, it should
# be possible to tweak the behavior of library(mi) simply by 1) creating a new class that
# inherits from the relevant existing class, 2) writing methods for the mi() and fit_model()
# generics and 3) perhaps a few other things that you will have to discover on your own.

## missing_variable is a virtual class for a variable that may (or may not) have missingness
setClass("missing_variable", 
         representation(
           variable_name = "character", # name of the variable but do not rely on for anything important 
           raw_data = "ANY",           ## DO NOT EVER CHANGE THE VALUES OF THIS SLOT
           data = "ANY",               ## Copy the raw_data into data and modify data as necessary
           n_total = "integer",         # total number of potential datapoints, i.e. length of raw_data
           all_obs = "logical",         # are ALL datapoints actually observed, i.e. not missing?
           n_obs = "integer",           # number of observed datapoints 
           which_obs = "integer",       # which datapoints are observed
           all_miss = "logical",        # are ALL datapoints missing, only true for latent variables
           n_miss = "integer",          # number of missing datapoints in the data slot (originally)
           which_miss = "integer",      # which datapoints are missing in the data slot
           n_extra = "integer",         # number of extra datapoints added (as missing)
           which_extra = "integer",     # which datapoints are extras
           n_unpossible = "integer",    # number of datapoints for which the variable could not be observed
           which_unpossible = "integer",# which datapoints could not be observed
           n_drawn = "integer",         # number of datapoints to impute
           which_drawn = "integer",     # which datapoints are imputed
           imputation_method = "character", # how to impute them
           family = "WeAreFamily",      # see help(family)
           known_families = "character",# families listed on help(family) plus multinomial()
           known_links = "character",   # see help(family)
           imputations = "MatrixTypeThing",      # iterations x n_drawn matrix of imputation history
           done = "logical",            # are we finished imputing?
           parameters = "MatrixTypeThing",       # history of estimated parameters in modeling this variable
           model = "ANY",               # last model fit
           fitted = "ANY",              # last fitted values
           "VIRTUAL"),
         prototype(
           variable_name = NA_character_,
           imputations = matrix(NA_real_, 0, 0),
           parameters = matrix(NA_real_, 0, 0),
           imputation_method = .known_imputation_methods,
           family = NA_character_,
           known_families = .known_families,
           known_links = .known_links
           ),
         validity = function(object) {
           out <- TRUE
           l <- length(object@raw_data)
           if(l == 0) return(out)
           
           if(sum(-object@n_total, object@n_obs, object@n_miss, object@n_extra, object@n_unpossible, na.rm = TRUE)) {
             out <- paste(object@variable_name, 
                          ": slots 'n_obs', 'n_miss', 'n_extra', and 'n_unpossible' must sum to 'n_total'")
           }
           else if(!(length(object@which_obs) %in%  c(0:1, object@n_obs))) {
             out <- paste(object@variable_name, ": 'n_obs' must equal the length of 'which_obs'")
           }
           else if(!(length(object@which_miss) %in% c(0:1, object@n_miss))) {
             out <- paste(object@variable_name, ": 'n_miss' must equal the length of 'which_miss'")
           }
           else if(!(length(object@which_extra) %in% c(0:1, object@n_extra))) {
             out <- paste(object@variable_name, ": 'n_extra' must equal the length of 'which_extra'")
           }
           else if(!(length(object@which_extra) %in% c(0:1, object@n_unpossible))) {
             out <- paste(object@variable_name, ": 'n_unpossible' must equal the length of 'which_unpossible'")
           }
           else if(sum(object@n_obs)) {
             temp <- sort(c(object@which_obs, object@which_miss, object@which_extra, object@which_unpossible))
             names(temp) <- NULL
             if(!identical(1:object@n_total, temp)) {
               out <- paste(object@variable_name, ": ''which_*' slots must be mutually exclusive and exhaustive")
             }
           }
           for(i in slotNames(object)) {
             if(i %in% c("raw_data", "data", "which_obs", "which_miss", "which_extra", 
                         "which_unpossible", "which_drawn", "imputation_method", "known_transformations",
                         "family", "known_families", "known_links", "levels", "cutpoints")) next
             if((l <- length(slot(object, i))) > 1) {
               out <- paste(object@variable_name, ": length of", i, "must be 0 or 1 but is", l)
               break
             }
           }
           return(out)
         }
         )

## this initialize() method gets called for everything that inherits from missing_variable
## but can be modified by a subsequently-called initialize() method
setMethod("initialize", "missing_variable", def = 
  function(.Object, NA.strings = c("", ".", "Na", "N/a", "N / a", "NaN",
                                   "Not Applicable", "Not applicable",
                                   "Not Available", "Not available", 
                                   "Not Ascertained", "Not ascertained",
                                   "Unavailable", "Unknown", "Missing",
                                   "Dk", "Don't Know", "Don't know", "Do Not Know", "Do not know"), ...) {
    .Object <- callNextMethod()
    if(length(.Object@raw_data) == 0) return(.Object)
    if(length(.Object@data) == 0) { # copy raw_data into data
      .Object@data <- .Object@raw_data
      names(.Object@data) <- .Object@variable_name
    }
    # bookkeeping
    infinites <- is.infinite(.Object@raw_data)
    if(any(infinites)) {
      warning(paste(.Object@variable_name, ": some observations are infinite, changing to NA"))
      .Object@data[infinites] <- NA
    }
    nans <- is.nan(.Object@raw_data)
    if(any(nans)) {
      warning(paste(.Object@variable_name, ": some observations are NaN, changing to NA"))
      .Object@data[nans] <- NA
    }
    NA.strings <- unique(c(NA.strings, toupper(NA.strings), tolower(NA.strings)))
    if(!is.numeric(.Object@raw_data)) for(i in seq_along(NA.strings)) {
      mark <- .Object@raw_data == NA.strings[i]
      if(any(mark, na.rm = TRUE)) {
        warning(paste(.Object@variable_name, ": some observations", NA.strings[i], "changing to NA"))
        .Object@data[mark] <- NA
      }
    }
    NAs <- which(is.na(.Object@data))
    if(length(NAs)) .Object@imputation_method <- "ppd" else .Object@imputation_method <- NA_character_
    .Object@n_miss <- length(NAs)
    .Object@which_miss <- NAs
    notNAs <- which(!is.na(.Object@data))
    .Object@n_obs <- length(notNAs)
    .Object@which_obs <- notNAs
    .Object@n_total <- length(NAs) + length(notNAs)
    .Object@all_miss <- length(notNAs) == 0
    .Object@all_obs  <- length(NAs) == 0
    if(!length(.Object@n_extra)) .Object@n_extra <- 0L
    if(!length(.Object@n_unpossible)) .Object@n_unpossible <- 0L
    .Object@n_drawn <- .Object@n_miss + .Object@n_extra
    .Object@which_drawn <- c(.Object@which_miss, .Object@which_extra)
    .Object@done <- FALSE
    return(.Object)
  })

setClass("irrelevant", representation("missing_variable"), 
         #   prototype(
         #     imputation_method = NA_character_,
         #     family = NA_character_)
         )

## a constant variable that has no missingness (and very few methods)
setClass("fixed", representation("irrelevant"), 
         validity = function(object) {
           out <- TRUE
           vals <- unique(object@raw_data)
           vals <- vals[!is.na(vals)]
           if(sum(object@n_miss)) {
             out <- paste(object@variable_name, ": fixed variables cannot have missingness")
           }
           else if(length(vals) > 1) {
             out <- paste(object@variable_name, ": purportedly 'fixed' variables cannot have multiple unique values")
           }
           return(out)
         }
         )

setClass("group", representation("irrelevant"))

## virtual class for categorical variables, which may be unordered, ordered, binary, or interval
setClass("categorical", 
         representation(
           "missing_variable", 
           levels = "character",
           "VIRTUAL"),
         prototype(
           known_families = c("multinomial", "binomial", "gaussian")
           )
         )

setMethod("initialize", "categorical", def = 
  function(.Object, ...) {
    .Object <- callNextMethod()
    l <- length(.Object@raw_data)
    if(l == 0) return(.Object)
    ## FIXME: check on the unused levels thing
    #   .Object@raw_data <- factor(.Object@raw_data)
    lev <- levels(factor(.Object@raw_data))
#     dummies <- t(sapply(.Object@raw_data, FUN = function(x) as.integer(x == lev)))[,-1, drop = FALSE]
#     if(ncol(dummies) == 1) colnames(dummies) <- .Object@variable_name
#     else colnames(dummies) <- lev[-1]
#     mark <- !apply(dummies, 2, FUN = function(x) all(x == 0, na.rm = TRUE))
#     dummies <- dummies[,mark, drop = FALSE]
#     lev <- c(lev[1], lev[-1][mark])
    .Object@levels <- lev
    .Object@data <- as.integer(factor(.Object@raw_data))
    return(.Object)
  })

## this is a hacked version of binomial()
multinomial <- function (link = "logit") 
{
  linktemp <- substitute(link)
  if (!is.character(linktemp)) {
    linktemp <- deparse(linktemp)
    if (linktemp == "link") {
      warning("use of multinomial(link=link) is deprecated\n", 
              domain = NA)
      linktemp <- eval(link)
      if (!is.character(linktemp) || length(linktemp) != 
        1L) 
        stop("'link' is invalid", domain = NA)
    }
  }
  okLinks <- c("logit", "probit", "cloglog", "cauchit", "log")
  if (linktemp %in% okLinks) 
    stats <- make.link(linktemp)
  else if (is.character(link)) {
    stats <- make.link(link)
    linktemp <- link
  }
  else {
    if (inherits(link, "link-glm")) {
      stats <- link
      if (!is.null(stats$name)) 
        linktemp <- stats$name
    }
    else {
      stop(gettextf("link \"%s\" not available for multinomial family; available links are %s", 
                    linktemp, paste(sQuote(okLinks), collapse = ", ")), 
           domain = NA)
    }
  }
  variance <- function(mu) mu * (1 - mu)
  validmu <- function(mu) all(mu > 0) && all(mu < 1)
  dev.resids <- binomial()$dev.resids
  aic <- function(y, n, mu, wt, dev) {
    m <- if (any(n > 1)) 
      n
    else wt
    -2 * sum(ifelse(m > 0, (wt/m), 0) * dbinom(round(m * 
      y), round(m), mu, log = TRUE))
  }
  initialize <- expression({
    if (NCOL(y) == 1) {
      if (is.factor(y)) y <- y != levels(y)[1L]
      n <- rep.int(1, nobs)
      y[weights == 0] <- 0
      if (any(y < 0 | y > 1)) stop("y values must be 0 <= y <= 1")
      mustart <- (weights * y + 0.5)/(weights + 1)
      m <- weights * y
      if (any(abs(m - round(m)) > 0.001)) warning("non-integer #successes in a multinomial glm!")
    } else if (NCOL(y) == 2) {
      if (any(abs(y - round(y)) > 0.001)) warning("non-integer counts in a multinomial glm!")
      n <- y[, 1] + y[, 2]
      y <- ifelse(n == 0, 0, y[, 1]/n)
      weights <- weights * n
      mustart <- (n * y + 0.5)/(n + 1)
    } else stop("for the multinomial family, y must be a vector of 0 and 1's\n", 
                "or a 2 column matrix where col 1 is no. successes and col 2 is no. failures")
  })
  simfun <- function(object, nsim) {
    ftd <- fitted(object)
    n <- length(ftd)
    ntot <- n * nsim
    wts <- object$prior.weights
    if (any(wts%%1 != 0)) 
      stop("cannot simulate from non-integer prior.weights")
    if (!is.null(m <- object$model)) {
      y <- model.response(m)
      if (is.factor(y)) {
        yy <- factor(1 + rbinom(ntot, size = 1, prob = ftd), 
                     labels = levels(y))
        split(yy, rep(seq_len(nsim), each = n))
      }
      else if (is.matrix(y) && ncol(y) == 2) {
        yy <- vector("list", nsim)
        for (i in seq_len(nsim)) {
          Y <- rbinom(n, size = wts, prob = ftd)
          YY <- cbind(Y, wts - Y)
          colnames(YY) <- colnames(y)
          yy[[i]] <- YY
        }
        yy
      }
      else rbinom(ntot, size = wts, prob = ftd)/wts
    }
    else rbinom(ntot, size = wts, prob = ftd)/wts
  }
  structure(list(family = "multinomial", link = linktemp, linkfun = stats$linkfun, 
                 linkinv = stats$linkinv, variance = variance, dev.resids = dev.resids, 
                 aic = aic, mu.eta = stats$mu.eta, initialize = initialize, 
                 validmu = validmu, valideta = stats$valideta, simulate = simfun), 
            class = "family")
}

## unordered categorical, which corresponds to an unordered factor with more than 2 levels
setClass("unordered-categorical", representation("categorical", 
                                                 estimator = "character", 
                                                 use_NA = "logical",
                                                 rank = "integer"),
         prototype(
           estimator = "MNL",
           imputation_method = c("ppd", "pmm", "mode", "mcar", NA_character_),
           family = multinomial(link = "logit"),
           known_families = c("multinomial", "binomial"),
           known_links = c("logit", "probit", "cauchit", "log", "cloglog"),
           use_NA = FALSE,
           rank = NA_integer_
           ),
         validity = function(object) {
           out <- TRUE
           values <- unique(object@raw_data)
           values <- values[!is.na(values)]
           
           im <- getClass(class(object))@prototype@imputation_method
           if(length(values) > 0 && length(values) <= 2) {
             out <- paste(object@variable_name, "unordered-categoricals must have more than 2 levels; otherwise use binary")
           }
           else if(!all(object@imputation_method %in% im)) {
             out <- paste(object@variable_name, ": 'imputation_method' must be one of:\n", paste(im, collapse = ", "))
           }
#            else if(object@family$family != "multinomial") {
#              out <- "the 'family' slot of 'unordered-categorial' class must be 'multinomial(link = 'logit')'"
#            }
#            else if(object@family$link != "logit") {
#              out <- "the 'family' slot of 'unordered-categorial' class must be 'multinomial(link = 'logit')'"
#            }
           else if(!(object@estimator %in% c("MNL", "RNL"))) {
             out <- paste(object@variable_name, ": estimator not recognized")
           }
           else if(!(object@use_NA %in% c(TRUE, FALSE))) {
             out <- paste(object@variable_name, ": use_NA must be TRUE or FALSE")
           }
           return(out)
         }
)

## ordered categorical, which corresponds to an ordered factor
setClass("ordered-categorical", representation("categorical", 
                                               cutpoints = "numeric"),
         prototype(
           imputation_method = c("ppd", "pmm", "mode", "mcar", NA_character_),
           family = multinomial(link = "logit"),
           known_families = c("multinomial", "gaussian", "binomial", "quasibinomial"),
           known_links = "logit"
           ),
         validity = function(object) {
           out <- TRUE
           im <- getClass(class(object))@prototype@imputation_method
           if(!(object@family$family %in% getClass(class(object))@prototype@known_families)) { # interval and binary are validated separately
             out <- "the 'family' slot of 'ordered-categorial' class must be 'multinomial()'"
           }
           else if(object@family$family == "multinomial" && object@family$link != "logit") {
             out <- "the 'family' slot of 'ordered-categorial' class must be 'multinomial(link = 'logit')'"
           }
           else if(!all(object@imputation_method %in% im)) {
             out <- paste(object@variable_name, ": 'imputation_method' must be one of:\n", paste(im, collapse = ", "))
           }
           return(out)
         }
         )

## ordered categorical with known cutpoints that discretize a continuous variable (like income)
setClass("interval", representation("ordered-categorical"),
         prototype(
           imputation_method = c("ppd", NA_character_),
           family = gaussian(),
           known_families = "gaussian",
           known_links = c("identity", "inverse", "log")
           ),
         validity = function(object) {
           out <- TRUE
           if(!(object@imputation_method[1] == "ppd")) {
             out <- paste(object@variable_name, ": 'imputation_method' must be 'ppd'")
           }
           else if(object@family$family != "gaussian") {
             out <- "the 'family' slot of 'interval' class must be 'gaussian()'"
           }
           return(out)
         }
         )

## binary variable
# binary inherits from ordered-categorical because it often makes sense to think of
# those who are coded as 1 as having "more" of something than those who are coded as
# zero. Also, binary logit, probit, etc. are special cases of ordinal logit, probit,
# etc. with one cutpoint fixed at zero.
setClass("binary", representation("ordered-categorical"),
         prototype(
           family = binomial(link = "logit"),
           known_families = c("binomial", "quasibinomial"),
           known_links = c("logit", "probit", "cauchit", "log", "cloglog"),
           cutpoints = 0.0),
         validity = function(object) {
           out <- TRUE
           if(length(object@raw_data) == 0) return(out)
           vals <- unique(object@raw_data)
           vals <- vals[!is.na(vals)]
           kf <- getClass(class(object))@prototype@known_families
           kl <- getClass(class(object))@prototype@known_links
           if(length(vals) != 2) {
             out <- paste(object@variable_name, ": binary variables must have exactly two response categories")
           }
           else if(!identical(object@cutpoints, 0.0)) {
             out <- paste(object@variable_name, ": 'cutpoints' must be 0.0 for a binary variable")
           }
           else if(!(object@family$family %in% kf)) {
             out <- paste(object@variable_name, ": the 'family' slot of a object of class 'binary' must be one of", paste(kf, collapse = ", "))
           }
           else if(!(object@family$link %in% kl)) {
             out <- paste(object@variable_name, ": the 'link' slot of the 'family' slot of a object of class 'binary' must be one of", paste(kl, collapse = ", "))
           }
           return(out)
         }
         )

setMethod("initialize", "binary", def = 
  function(.Object, ...) {
    .Object <- callNextMethod()
    l <- length(.Object@raw_data)
    if(l == 0) return(.Object)
    .Object@data <- as.integer(.Object@data == max(.Object@data, na.rm = TRUE)) + 1L
    return(.Object)
  })

setClass("grouped-binary", representation("binary", strata = "character"),
         prototype(
           imputation_method = "pmm"
           ),
         validity = function(object) {
           out <- TRUE
           if(length(object@raw_data) == 0) return(out)
           if(!requireNamespace("survival")) {
             out <- "the 'survival' package must be installed to use 'grouped-binary' variables"
           }
           else if(length(object@strata) == 0) {
             warning(paste("you must specify the 'strata' slot for", object@variable_name, 
                           "see help('grouped-binary-class')"))
           }
           return(out)
         }
         )

setMethod("initialize", "grouped-binary", def = 
  function(.Object, ...) {
    .Object <- callNextMethod()
    l <- length(.Object@raw_data)
    if(l == 0) return(.Object)
    .Object@imputation_method <- "pmm"
    return(.Object)
  })

## count variables, which must be nonnegative integers
setClass("count", representation("missing_variable"),
         prototype(
           imputation_method = c("ppd", "pmm", "mean", "median", "expectation", "mcar", NA_character_),
           family = quasipoisson(),
           known_families = c("quasipoisson", "poisson"),
           known_links = c("log", "identity", "sqrt")
           ),
         validity = function(object) {
           out <- TRUE
           l <- length(object@raw_data)
           if(l == 0) return(out)
           im <- getClass(class(object))@prototype@imputation_method
           if(any(object@raw_data < 0, na.rm = TRUE)) {
             out <- paste(object@variable_name, ": counts must be nonnegative")
           }
           else if(any(object@raw_data != as.integer(object@raw_data), na.rm = TRUE)) {
             out <- paste(object@variable_name, ": must contain all nonnegative integers to use the 'count' class")
           }
           else if(!all(object@imputation_method %in% im)) {
             out <- paste(object@variable_name, ": 'imputation_method' must be one of:\n", paste(im, collapse = ", "))
           }
           else if(sum(object@n_unpossible)) {
             out <- paste(object@variable_name, ": unpossible observations not supported for count variables yet")
           }
           return(out)
         }
         )

.identity_transform <- function(y, ...) return(y)
.standardize_transform <- function(y, mean = stop("must supply mean"), sd = stop("must supply sd"), inverse = FALSE) {
  if(inverse) return(y * 2 * sd + mean)
  else return( (y - mean) / (2 * sd) )
}

## continuous variables, which may have inequality restrictions or transformation functions
setClass("continuous", 
         representation(
           "missing_variable",
           transformation = "function",
           inverse_transformation = "function",
           transformed = "logical", # TRUE -> in transformed state
           known_transformations = "character"
           ),
         prototype(
           imputation_method = c("ppd", "pmm", "mean", "median", "expectation", "mcar", NA_character_),
           transformed = TRUE,
           transformation = .standardize_transform,
           inverse_transformation = .standardize_transform,
           family = gaussian(), 
           known_families = c("gaussian", "Gamma", "inverse.gaussian", "binomial"), # binomial() is only for (SC_)proportions
           known_links = .known_links[.known_links != "sqrt"],
           known_transformations = c("standardize", "identity", "log", "logshift", "squeeze", "sqrt", "cuberoot", "qnorm")
           ),
         validity = function(object) {
           out <- TRUE
           im <- getClass(class(object))@prototype@imputation_method
           kf <- getClass(class(object))@prototype@known_families
           kl <- getClass(class(object))@prototype@known_links
           if(!all(object@imputation_method %in% im)) {
             out <- paste(object@variable_name, ": 'imputation_method' must be one of:\n", paste(im, collapse = ", "))
           }
           else if(sum(object@n_unpossible)) {
             out <- paste(object@variable_name, ": unpossible observations not supported for continuous variables yet")
           }
           else if(!(object@family$family %in% kf)) {
             out <- paste(object@variable_name, ": the 'family' slot of a object of class 'binary' must be one of", paste(kf, collapse = ", "))
           }
           else if(!(object@family$link %in% kl)) {
             out <- paste(object@variable_name, ": the 'link' slot of the 'family' slot of a object of class 'binary' must be one of", paste(kl, collapse = ", "))
           }
           return(out)
         }
         )

setMethod("initialize", "continuous", def = 
  function(.Object, ...) {
    .Object <- callNextMethod()
    l <- length(.Object@raw_data)
    if(l == 0) return(.Object)
    if(identical(.Object@transformation, .standardize_transform)) {
      mean <- mean(.Object@raw_data, na.rm = TRUE)
      sd <- sd(.Object@raw_data, na.rm = TRUE)
      formals(.Object@transformation)$mean <- formals(.Object@inverse_transformation)$mean <- mean
      formals(.Object@transformation)$sd <- formals(.Object@inverse_transformation)$sd <- sd
      formals(.Object@inverse_transformation)$inverse <- TRUE
    }
    else if(identical(.Object@transformation, .logshift)) {
      y <- .Object@raw_data
      if(any(y < 0, na.rm = TRUE)) a <- - min(y, na.rm = TRUE)
      else a <- 0
      a <- (a + min(y[y > 0], na.rm = TRUE)) / 2
      formals(.Object@transformation)$a <- formals(.Object@inverse_transformation)$a <- a
      formals(.Object@inverse_transformation)$inverse <- TRUE
    }
    .Object@data <- .Object@transformation(.Object@raw_data)
    .Object@data[.Object@which_miss] <- NA_real_
    return(.Object)
  })

setClass("bounded-continuous", representation("continuous", lower = "numeric", upper = "numeric"),
         prototype(
           imputation_method = "ppd",
           transformation = .identity_transform,
           inverse_transformation = .identity_transform
           ), 
         validity = function(object) {
           out <- TRUE
           #     if(any(object@raw_data <= object@lower, na.rm = TRUE)) {
           #       out <- paste(object@variable_name, ": all observed data must be strictly greater than 'lower'")
           #     }
           #     else if(any(object@raw_data >= object@upper, na.rm = TRUE)) {
           #       out <- paste(object@variable_name, ": all observed data must be strictly less than 'upper'")
           #     }
           if(any(object@lower > object@upper)) {
             out <- paste(object@variable_name, ": lower bounds must be less than or equal to upper bounds")
           }
           else if(object@imputation_method != "ppd") {
             out <- paste(object@variable_name, ": 'imputation_method' must be 'ppd' for 'bounded-continuous' variables ")
           }
           else if(!requireNamespace("truncnorm")) {
             out <- paste(object@variable_name, ": the 'truncnorm' package must be installed to use the 'bounded-continuous' class")
           }
           return(out)
         }
         )

setMethod("initialize", "bounded-continuous", def = 
  function(.Object, lower = -Inf, upper = Inf, ...) {
    .Object <- callNextMethod()
    l <- length(.Object@raw_data)
    if(l == 0) return(.Object)
    .Object@lower <- lower
    .Object@upper <- upper
    return(.Object)
  })

setClass("positive-continuous", representation("continuous"),
         prototype(
           transformation = log,
           inverse_transformation = exp,
           known_transformations = c("log", "sqrt", "squeeze", "qnorm") 
           ),
         validity = function(object) {
           out <- TRUE
           if(any(object@raw_data <= 0, na.rm = TRUE)) {
             out <- paste(object@variable_name, ": positive variables must be positive")
           }
           return(out)
         }
         )

## must be on the (0,1) interval
setClass("proportion", representation("positive-continuous", link.phi = "WeAreFamily"),
         prototype(
           transformed = FALSE,
           transformation = .identity_transform,
           inverse_transformation = .identity_transform,
           known_transformations = c("squeeze", "qnorm"),
           family = binomial(),
           known_families = c("binomial", "gaussian"),
           known_links = .known_links[.known_links != "sqrt"],
           link.phi = "log"),
         validity = function(object) {
           out <- TRUE
           kf <- getClass(class(object))@prototype@known_families
           kl <- getClass(class(object))@prototype@known_links
           if(any(object@raw_data > 1, na.rm = TRUE)) {
             out <- paste(object@variable_name, ": proportions must be on the unit interval")
           }
           else if(any(object@raw_data == 1, na.rm = TRUE)) {
             out <- paste(object@variable_name, ": some proportions are equal to 1.0 so use the SC_proportion class")
           }
           else if(!(object@family$family %in% kf)) {
             out <- paste(object@variable_name, ": the 'family' slot of a object of class 'proportion' must be one of", paste(kf, collapse = ", "))
           }
           else if(!(object@family$link %in% kl)) {
             out <- paste(object@variable_name, ": the 'link' slot of the 'family' slot of a object of class 'proportion' must be one of", paste(kl, collapse = ", "))
           }
           else if(object@family$family == "binomial" && !requireNamespace("betareg")) {
             out <- paste(object@variable_name, ": you must install the 'betareg' package to model proportions as proportions")
           }
           return(out)
         }
         )

# setClass("truncated-continuous", 
#   representation("continuous",
#                  lower = "ANY", 
#                  upper = "ANY", 
#                  n_lower = "integer",
#                  which_lower = "integer",
#                  n_upper = "integer",
#                  which_upper = "integer",
#                  n_both = "integer",
#                  which_both = "integer",
#                  n_truncated = "integer",
#                  which_truncated = "integer",
#                  "VIRTUAL")
# )
# 
# setClass("NN_truncated-continuous", representation("truncated-continuous", lower = "numeric", upper = "numeric"))
# 
# setMethod("initialize", "NN_truncated-continuous", def = 
# function(.Object, ...) {
#   .Object <- callNextMethod()
#   l <- length(.Object@raw_data)
#   if(l == 0) return(.Object)
#   if(identical(.Object@transformation, .standardize_transform)) {
#     mean <- mean(.Object@raw_data, na.rm = TRUE)
#     sd <- sd(.Object@raw_data, na.rm = TRUE)
#     formals(.Object@transformation)$mean <- formals(.Object@inverse_transformation)$mean <- mean
#     formals(.Object@transformation)$sd <- formals(.Object@inverse_transformation)$sd <- sd
#     formals(.Object@inverse_transformation)$inverse <- TRUE
#   }
#   .Object@data <- .Object@transformation(.Object@raw_data)
# 
#   if(length(.Object@lower) == 0 & length(.Object@upper) == 0) {
#     stop("at least one of 'lower' and 'upper' must be specified")
#   }
#   ## FIXME: Deal with interval censoring or force it to the interval class
#   .Object@n_both <- 0L
#   lowers <- .Object@raw_data <= .Object@lower
#   .Object@n_lower <- sum(lowers)
#   .Object@which_lower <- which(lowers)
#   uppers <- .Object@raw_data >= .Object@upper
#   .Object@n_uppers <- sum(uppers)
#   .Object@which_uppers <- which(uppers)
#   .Object@n_truncated <- .Object@n_lower + .Object@n_upper
#   .Object@which_truncated <- c(.Object@which_lower, .Object@which_upper)
#   return(.Object)
# })
# 
# setClass("FN_truncated-continuous", representation("truncated-continuous", lower = "function", upper = "numeric"))
# setClass("NF_truncated-continuous", representation("truncated-continuous", lower = "numeric", upper = "function"))
# setClass("FF_truncated-continuous", representation("truncated-continuous", lower = "function", upper = "function"))
# 
# setClass("censored-continuous", 
#   representation("continuous", 
#                  lower = "ANY", 
#                  upper = "ANY", 
#                  n_lower = "integer",
#                  which_lower = "integer",
#                  n_upper = "integer",
#                  which_upper = "integer",
#                  n_both = "integer",
#                  which_both = "integer",
#                  n_censored = "integer",
#                  which_censored = "integer",
#                  lower_indicator = "binary",
#                  upper_indicator = "binary",
#                  "VIRTUAL")
# )
# setClass("NN_censored-continuous", representation("censored-continuous", lower = "numeric", upper = "numeric"))
# setMethod("initialize", "NN_censored-continuous", def = 
# function(.Object, ...) {
#   .Object <- callNextMethod()
#   l <- length(.Object@raw_data)
#   if(l == 0) return(.Object)
#   if(identical(.Object@transformation, .standardize_transform)) {
#     mean <- mean(.Object@raw_data, na.rm = TRUE)
#     sd <- sd(.Object@raw_data, na.rm = TRUE)
#     formals(.Object@transformation)$mean <- formals(.Object@inverse_transformation)$mean <- mean
#     formals(.Object@transformation)$sd <- formals(.Object@inverse_transformation)$sd <- sd
#     formals(.Object@inverse_transformation)$inverse <- TRUE
#   }
#   .Object@data <- .Object@transformation(.Object@raw_data)
# 
#   if(length(.Object@lower) == 0 & length(.Object@upper) == 0) {
#     stop("at least one of 'lower' and 'upper' must be specified")
#   }
#   ## FIXME: Deal with interval censoring or force it to the interval class
#   .Object@n_both <- 0L
#   lowers <- .Object@raw_data <= .Object@lower
#   .Object@n_lower <- sum(lowers, na.rm = TRUE)
#   .Object@which_lower <- which(lowers)
#   if(.Object@n_lower > 0) {
#     .Object@lower_indicator <- missing_variable(as.ordered(lowers), type = "binary", 
#                                  variable_name = paste(.Object@variable_name, "lower", sep = ""))
#   }
#   uppers <- .Object@raw_data >= .Object@upper
#   .Object@n_upper <- sum(uppers, na.rm = TRUE)
#   .Object@which_upper <- which(uppers)
#   if(.Object@n_upper > 0) {
#     .Object@lower_indicator <- missing_variable(as.ordered(uppers), type = "binary", 
#                                  variable_name = paste(.Object@variable_name, "upper", sep = ""))
#   }
#   .Object@n_censored <- .Object@n_lower + .Object@n_upper
#   .Object@which_censored <- c(.Object@which_lower, .Object@which_upper)
#   return(.Object)
# })
# 
# setClass("FN_censored-continuous", representation("censored-continuous", lower = "function", upper = "numeric"))
# setClass("NF_censored-continuous", representation("censored-continuous", lower = "numeric", upper = "function"))
# setClass("FF_censored-continuous", representation("censored-continuous", lower = "function", upper = "function"))

setClass("semi-continuous", representation("continuous", indicator = "ordered-categorical"),
         prototype(
           transformation = .identity_transform,
           inverse_transformation = .identity_transform)
         )

.logshift <- function(y, a, inverse = FALSE) {
  if(inverse) exp(y) - a
  else log(y + a)
}

setClass("nonnegative-continuous", representation("semi-continuous"),
         prototype(transformation = .logshift,
                   inverse_transformation = .logshift,
                   known_transformations = c("logshift", "squeeze", "identity")),
         validity = function(object) {
           out <- TRUE
           if(any(object@raw_data < 0, na.rm = TRUE)) {
             out <- paste(object@variable_name, ": nonnegative variables must be nonnegative")
           }
           return(out)
         }
         )

setMethod("initialize", "nonnegative-continuous", def = 
  function(.Object, ...) {
    .Object <- callNextMethod()
    l <- length(.Object@raw_data)
    if(l == 0) return(.Object)
    is_zero <- as.integer(.Object@raw_data == 0)
    if(any(is_zero, na.rm = TRUE)) {
      .Object@indicator <- missing_variable(is_zero, type = "binary", 
                                            variable_name = paste(.Object@variable_name, ":is_zero", sep = ""))
    }
    .Object@data <- .Object@transformation(.Object@raw_data)
    if(!all(is.finite(.Object@data[!is.na(.Object)]))) {
      stop(paste(.Object@variable_name, ": some transformed values are infinite or undefined"))
    }
    return(.Object)
  })

.squeeze_transform <- function(y, inverse = FALSE) {
  n <- length(y)
  if(inverse) (y * n - .5) / (n - 1)
  else (y * (n - 1) + .5) / n
}

## some values are zero and / or one
setClass("SC_proportion", representation("nonnegative-continuous", link.phi = "WeAreFamily"),
         prototype(
           transformation = .squeeze_transform,
           inverse_transformation = .squeeze_transform,
           known_transformations = c("squeeze", "qnorm"),
           family = binomial(),
           known_families = "binomial",
           known_links = getClass("binary")@prototype@known_links,
           link.phi = "log"
           ),
         validity = function(object) {
           out <- TRUE
           if(any(object@data > 1, na.rm = TRUE)) {
             out <- paste(object@variable_name, ": proportions must be less than or equal to 1")
           }
           else if(object@family$family != "binomial") {
             out <- paste(object@variable_name, ": 'family' must be 'binomial'")
           }
           else if(!identical(body(object@transformation), body(.squeeze_transform))) {
             out <- paste(object@variable_name, ": 'transformation' must be 'squeeze'")
           }
           else if(!requireNamespace("betareg")) {
             out <- paste(object@variable_name, ": you must install the 'betareg' package to model proportions")
           }
           return(out)
         }
         )

setMethod("initialize", "SC_proportion", def = 
  function(.Object, ...) {
    .Object <- callNextMethod()
    l <- length(.Object@raw_data)
    if(l == 0) return(.Object)
    
    if(any(.Object@raw_data == 0, na.rm = TRUE)) {
      if(any(.Object@raw_data == 1, na.rm = TRUE)) {
        is_bound <- ifelse(.Object@raw_data == 0, -1, ifelse(.Object@raw_data == 1, 1, 0))
        .Object@indicator <- missing_variable(is_bound, type = "ordered-categorical", 
                                              variable_name = paste(.Object@variable_name, ":is_bound", sep = ""))
      }
      else {
        is_zero <- as.integer(.Object@raw_data == 0)
        .Object@indicator <- missing_variable(is_zero, type = "binary", 
                                              variable_name = paste(.Object@variable_name, ":is_zero", sep = ""))
      }
    }
    else {
      is_one <- as.integer(.Object@raw_data == 1)
      .Object@indicator <- missing_variable(is_one, type = "binary", 
                                            variable_name = paste(.Object@variable_name, ":is_one", sep = ""))
    }
    return(.Object)
  })

# A missing_data.frame is a another important S4 class that is not unlike a data.frame, except
# that its "columns" (actually list elements) are objects that inherit from the missing_variable
# class. The missing_data.frame class should, in principle, contain ALL the necessary information
# regarding how the missing_variables relate to each other. Together, the missing_variable class(es)
# and the missing_data.frame class supplant the mi.info S4 class in previous versions of library(mi).

.get_slot <- 
  function(object, name, simplify = TRUE) {
    if(isS4(object)) return(slot(object, name))
    else if(is.list(object)) sapply(object, FUN = slot, name = name, simplify = simplify)
    else stop("'object' not supported")
  }

setOldClass("data.frame")
setClass("missing_data.frame", 
         representation(       
           variables = "list",      # of missing_variables
           no_missing = "logical",  # basically a collection of the all_obs slots of the missing_variables
           patterns = "factor",     # indicates which missingness_pattern an observation belongs to
           DIM = "integer",         # observations x variables
           DIMNAMES = "list",       # list of rownames and colnames
           postprocess = "function",# makes additional variables from existing variables (interactions, etc.)
           index = "list",          # this indicate which variables to exclude when modeling a given variable
           X = "MatrixTypeThing",   # ALL variables (categorical variables are in dummy-variable form)
           weights = "list",        # this gets passed to bayesglm() and similar modeling functions
           priors = "list",         # the elements of this get passed to bayesglm() and other modeling functions in arm
           correlations = "matrix", # has SMCs and Spearman correlations
           done = "logical",        # are we done?
           workpath = "character"),
         contains = "data.frame", 
         prototype(postprocess = function() stop("postprocess does not work yet"), X = matrix(NA_real_, 0, 0), done = FALSE),
         validity = function(object) {
           out <- TRUE
           l <- length(object@variables)
           if(l == 0) return(out)
           
           if(!all(sapply(object@variables, FUN = is, class2 = "missing_variable"))) {
             out <- "all of the list elements in 'variables' must inherit from the 'missing_variable' class"
           }
           else if(length(unique(.get_slot(object@variables, "n_total"))) > 1) {
             out <- "all missing_variables must have the same 'n_total'"
           }
           else if(!is.numeric(object@X)) {
             out <- "'X' must be a numeric matrix"
           }
           missingness <- .get_slot(object@variables, "which_miss", simplify = FALSE)
           varnames <- .get_slot(object@variables, "variable_name")
           names(missingness) <- varnames
           missingness <- missingness[sapply(missingness, length) > 0]
           if(length(missingness) > 1) { ## FIXME: Very slow
             combos <- combn(length(missingness), 2) 
             dupes <- apply(combos, 2, FUN = function(x) {
               mx1 <- missingness[[x[1]]]
               mx2 <- missingness[[x[2]]]
               if(length(mx1) == length(mx2)) {
                 if(identical(mx1, mx2)) return(1L)
               }
               else if(length(mx1) > length(mx2)) {
                 if(all(mx2 %in% mx1)) return(2L)
               }
               else if(all(mx1 %in% mx2)) return(3L)
               return(0L)
             })
             if(any(dupes == 1L)) {
               temp <- matrix(names(missingness)[combos[,which(dupes == 1L)]], ncol = 2, byrow = TRUE)
               cat("NOTE: The following pairs of variables appear to have the same missingness pattern.\n",
                   "Please verify whether they are in fact logically distinct variables.\n")
               print(temp)
               # 	warning("Potentially duplicated variables detected by duplicated variable detector")
             }
             else if(any(dupes == 2L)) {
               temp <- matrix(names(missingness)[combos[,which(dupes == 2L)]], ncol = 2, byrow = TRUE)
               cat("NOTE: In the following pairs of variables, the missingness pattern of the second is a subset of the first.\n",
                   "Please verify whether they are in fact logically distinct variables.\n")
               print(temp)
             }
             else if(any(dupes == 3L)) {
               temp <- matrix(names(missingness)[combos[,which(dupes == 3L)]], ncol = 2, byrow = TRUE)
               cat("NOTE: In the following pairs of variables, the missingness pattern of the first is a subset of the second.\n",
                   "Please verify whether they are in fact logically distinct variables.\n")
               print(temp)
             }
           }
           return(out)
         }
         )

.set_priors <- function(variables, mu = 0) { ## FIXME: maybe add an option to draw from such a t distribution?
  foo <- function(y) {
    out <- list(prior.mean = mu, prior.scale  = 2.5, prior.df = 1, 
                prior.mean.for.intercept = mu, prior.scale.for.intercept = 10, prior.df.for.intercept = 1)
    if(is(y, "irrelevant") | y@all_obs) return(NULL)
    else if(is(y, "binary")) {
      if(y@family$link == "probit") {
        out[[2]] <- out[[2]] * dnorm(0) / dlogis(0)
        out[[4]] <- out[[4]] * dnorm(0) / dlogis(0)
      }
    }
    else if(is(y, "categorical")) {
      out <- list(prior.mean = mu, prior.scale = 2.5, prior.df = 1, prior.counts.for.bins = 1/(1 + length(y@levels)))
    }
    return(out)
  }
  
  out <- lapply(variables, FUN = function(y) foo(y))
  for(i in seq_along(variables)) if(is(y <- variables[[i]], "semi-continuous")) out[[y@indicator@variable_name]] <- foo(y@indicator)
  return(out)
}

setMethod("initialize", "missing_data.frame", def =
  function(.Object, include_missingness = TRUE, skip_correlation_check = FALSE, ...) {
    .Object <- callNextMethod()
    l <- length(.Object@variables)
    if(l == 0) return(.Object)
    varnames <- names(.Object@variables)
    if(is.null(varnames)) {
      if(is.null(.Object@DIMNAMES[[2]])) names(.Object@variables) <- sapply(.Object@variables, FUN = .get_slot, name = "variable_name")
      else names(.Object@variables) <- .Object@DIMNAMES[[2]]
    }
    else for(i in 1:l) .Object@variables[[i]]@variable_name <- varnames[i]
    .Object@DIM <- c(.Object@variables[[1]]@n_total, l)
    .Object@no_missing <- sapply(.Object@variables, FUN = .get_slot, name = "all_obs")
    if(length(.Object@DIMNAMES) == 0) .Object@DIMNAMES <- list(NULL, names(.Object@variables))

    Z <- lapply(.Object@variables, FUN = function(y) {
      if(is(y, "irrelevant")) return(NULL) else return(is.na(y))
    })
    Z <- as.matrix(as.data.frame(Z[!sapply(Z, is.null)]))
    if(any(apply(Z, 1, all))) {
      warning("Some observations are missing on all included variables.\n",
              "Often, this indicates a more complicated model is needed for this missingness mechanism")
    }
    uZ <- unique(Z)
    if(nrow(uZ) == 1) {
      if(all(uZ[1,] == 0)) patterns <- factor(rep("nothing", nrow(Z)))
      else patterns <- factor(colnames(uZ)[which(uZ[1,] == 1)], nrow(Z))
    }
    else {
      uZ <- uZ[order(rowSums(uZ)),,drop = FALSE]
      patterns <- apply(Z, 1, FUN = function(x) which(apply(uZ, 1, FUN = function(u) all(u == x))))
      pattern_labels <- apply(uZ, 1, FUN = function(x) paste(names(x)[x], collapse = ", "))
      if(length(pattern_labels)) {
        if(pattern_labels[1] == "") pattern_labels[1] <- "nothing"
        pattern_lables <- paste("missing:", pattern_labels)
        patterns <- factor(patterns, labels = pattern_labels, ordered = FALSE)
      }
      else patterns <- factor(patterns)
    }
    .Object@patterns <- patterns

    if(!length(.Object@workpath)) {
      .Object@workpath <- file.path(tempdir(), paste("mi", as.integer(Sys.time()), sep = ""))
    }
    dir.create(.Object@workpath, showWarnings = FALSE)

    if(is(.Object, "allcategorical_missing_data.frame")) return(.Object)
    
    Z <- Z[,!duplicated(t(Z)), drop = FALSE]
    Z <- Z[,apply(Z, 2, FUN = function(x) length(unique(x))) > 1, drop = FALSE]
    ## FIXME: What to do if two columns of Z are collinear?
    if(ncol(Z) > 0) colnames(Z) <- paste("missing", colnames(Z), sep = "_")
    else include_missingness <- FALSE

    X <- lapply(.Object@variables, FUN = function(x) {
      if(is(x, "irrelevant"))  return(NULL)
      else if(is(x, "categorical")) return(.cat2dummies(x))
      else if(is(x, "semi-continuous")) {  
        out <- cbind(x@data, .cat2dummies(x@indicator))
        colnames(out) <- c(x@variable_name, 
                           paste(x@variable_name, 2:ncol(out) - 1, sep = "_"))
        return(out)
      }
      else if(is(x, "censored-continuous")) {
        temp <- x@data
        if(x@n_lower) temp <- cbind(temp, lower = x@lower_indicator@data)
        if(x@n_upper) temp <- cbind(temp, upper = x@upper_indicator@data)
        if(x@n_both)  stop("FIXME: censoring on both sides not supported yet")
        return(temp)
      }
      else if(is(x, "truncated-continuous")) {
        temp <- x@data
        n <- length(temp)
        if(x@n_lower) temp <- cbind(lower = x@lower_indicator@data, temp)
        if(x@n_upper) temp <- cbind(upper = x@upper_indicator@data, temp)
        if(x@n_both)  stop("FIXME: censoring on both sides not supported yet")
        return(temp)
      }
      else return(x@data)
    }) ## NOTE: Might need to make this more complicated in the future
    X <- X[!sapply(X, is.null)]
    
    index <- vector("list", length = length(X))
    names(index) <- names(X)
    start <- 2L
    end <- 0L
    for(i in seq_along(index)) {
      end <- start + NCOL(X[[i]]) - 1L
      index[[i]] <- start:end
      start <- end + 1L
    }
    if(include_missingness) for(i in seq_along(index)) {
      nas <- is.na(.Object@variables[[i]])
      check <- apply(Z, 2, FUN = function(x) all(x == nas))
      index[[i]] <- c(index[[i]], which(check) + start - 1)
    }
    else for(i in seq_along(index)) index[[i]] <- c(index[[i]], start:(start + ncol(Z) - 1))
    
    grouped <- names(which(sapply(.Object@variables, is, class2 = "grouped-binary")))
    for(i in grouped) index[[i]] <- c(index[[i]], index[[.Object@variables[[i]]@strata]], 1)
    
    .Object@index <- index
    .Object@X <- cbind("(Intercept)" = 1, as.matrix(as.data.frame(X)), Z)
    
    correlations <- matrix(NA_real_, l,l)
    if(!skip_correlation_check) for(i in 1:(l - 1)) { ## FIXME: Put SMCs in the lower triangle
      if(is(.Object@variables[[i]], "irrelevant")) next
      x <- try(rank(xtfrm(.Object@variables[[i]]@raw_data)), silent = TRUE)
      if(!is.numeric(x)) next
      for(j in (i + 1):l) {
        if(is(.Object@variables[[j]], "irrelevant")) next
        y <- try(rank(xtfrm(.Object@variables[[j]]@raw_data)))
        if(!is.numeric(y)) next
        rho <- cor(x, y, use = "pair", method = "pearson") # on ranks
        if(is.finite(rho) && abs(rho) == 1) {
          warning(paste(names(.Object@variables)[i], "and", names(.Object@variables)[j], 
                        "have the same rank ordering.\n",
                        "Please verify whether they are in fact distinct variables.\n"))
        }
        if(is.finite(rho)) correlations[i,j] <- rho
      }
    }
    
    .Object@correlations <- correlations
    .Object@priors <- .set_priors(.Object@variables)
    .Object
  })

setClass("allcategorical_missing_data.frame", 
         representation("missing_data.frame", "Hstar" = "integer",
                        "parameters" = "list","latents" = "unordered-categorical"),
         
         prototype = prototype(Hstar = 20L),
         validity = function(object) {
           out <- TRUE
           types <- sapply(object@variables, 
                           FUN = function(y) is(y, "irrelevant") | is(y, "categorical"))
           if(!all(types)) {
             out <- "all variable classes must be 'irrelevant' or 'categorical'"
           }
           else if(length(object@Hstar) && object@Hstar < 1) {
             out <- "'Hstar' must be >= 1"
           }
           return(out)
         })

setMethod("initialize", "allcategorical_missing_data.frame", def =
  function(.Object, include_missingness = TRUE, ...) {
    .Object <- callNextMethod()
    l <- length(.Object@variables)
    n <- nrow(.Object)
    uc <- factor(rep(NA_integer_, n))
    .Object@latents <- new("unordered-categorical", raw_data = rep(NA_integer_, n))
    .Object@priors <- list(a = rep(1, ncol(.Object)), a_alpha = 1, b_alpha = 1)
    names(.Object@priors$a) <- colnames(.Object)
    return(.Object)
  })

setClass("experiment_missing_data.frame", representation("missing_data.frame", 
                                                         concept = "factor",
                                                         case = "character"),
         validity = function(object) {
           out <- TRUE
           l <- length(object@concept)
           if(l != length(object@variables)) {
             out <- "length of 'concept' must equal the number of variables"
           }
           else if(!all(levels(object@concept) %in% c("outcome", "covariate", "treatment"))) {
             out <- "all elements of 'concept' must be exactly one of 'outcome', 'covariate', or 'treatment'"
           }
           else if(sum(object@concept == "treatment") != 1) {
             out <- "there must be exactly one variable designated 'treatment'"
           }
           else if(!is(object@variables[[which(object@concept == "treatment")]], "binary")) {
             out <- "the 'treatment' variable must be of class 'binary'"
           }
           else if(object@variables[[which(object@concept == "treatment")]]@n_miss) {
             out <- "'treatment' variable cannot have any missingness"
           }
           else if(length(object@case) > 1) {
             out <- "'case' must be exactly one of 'outcomes', 'covariates', or 'both'"
           }
           else if(length(object@case) && !(object@case %in% c("outcomes", "covariates", "both"))) {
             out <- "'case' must be exactly one of 'outcomes', 'covariates', or 'both'"
           }
           return(out)
         })

setMethod("initialize", "experiment_missing_data.frame", def =
  function(.Object, include_missingness = TRUE, ...) {
    .Object <- callNextMethod()
    l <- 1  ## FIXME
    if(l == 0) return(.Object)
    names(.Object@concept) <- .Object@DIMNAMES[[2]]
    outcomes   <- any(!.Object@no_missing[.Object@concept == "outcomes"])
    covariates <- any(!.Object@no_missing[.Object@concept == "covariates"])
    
    .Object@case <- if(outcomes & covariates) "both" else if(outcomes) "covariates" else "outcomes"
    return(.Object)
  })

.empty_mdf_list <- list()
class(.empty_mdf_list) <- "mdf_list"
setClass("multilevel_missing_data.frame", representation("missing_data.frame", 
                                                         groups = "character", 
                                                         mdf_list = "mdf_list"),
         prototype(
           mdf_list = .empty_mdf_list
           ),
         
         validity = function(object) {
           out <- TRUE
           return(out)
         }
         )

setMethod("initialize", "multilevel_missing_data.frame", def =
  function(.Object, include_missingness = TRUE, ...) {
    .Object <- callNextMethod()
    classes <- sapply(.Object@variables, class)
    for(i in .Object@groups) classes[names(classes) == i] <- "fixed"
    df <- complete(.Object, m = 0L)
    mdf_list <- missing_data.frame(df, by = .Object@groups, types = classes)
    .Object@mdf_list <- mdf_list
    return(.Object)
  })

## an object of class mi merely holds the results of a call to mi(), primary the list of missing_data.frames
setClass("mi",
         representation(
           call       = "call",
           data = "list",             # of missing_data.frames
           total_iters  = "integer"), # how many iterations were conducted (can be a vector)
         )

## an object of class pooled has regression results using the Rubin rules
setClass("pooled", 
         representation(
           formula = "formula",
           fit = "character",
           models = "list",
           coefficients = "numeric",
           ses = "numeric",
           pooled_summary  = "ANY",
           call = "language"),
         )

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## these change the parametric family used in the imputation process

setMethod("change_family", signature(data = "missing", y = "missing_variable", to = "family"), def = 
  function(y, to) {
    y@family <- to
    validObject(y, complete = TRUE)
    return(y)
  })

setMethod("change_family", signature(data = "missing", y = "missing_variable", to = "missing"), def = 
  function(y, to) {
    cat("Likely families include:", y@known_families, sep = "\n")
    return(invisible(NULL))
  })

setMethod("change_family", signature(data = "missing", y = "proportion", to = "family"), def = 
  function(y, to) {
    y@family <- to
    if(y@family$family == "gaussian") {
      y@transformation <- qnorm
      y@inverse_transformation <- pnorm
    }
    else if(y@family$family == "binomial") {
      y@transformation <- .identity_transform
      y@transformation <- .identity_transform
    }
    y@data <- y@transformation(y@data)
    validObject(y, complete = TRUE)
    return(y)
  })

setMethod("change_family", signature(data = "missing", y = "unordered-categorical", to = "family"), def = 
  function(y, to) {
    y@family <- to
    y@estimator <- if(y@family$family == "binomial") "RNL" else "MNL"
    return(y)
  }
)

setMethod("change_family", signature(data = "missing_data.frame", y = "character", to = "character"), def =
  function(data, y, to) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    if(all(y %in% names(getClass("missing_variable")@subclasses))) {
      classes <- sapply(data@variables, class)
      y <- c(sapply(y, FUN = function(x) {
        names(classes[which(classes == x)])
      }))
      if(is.list(y)) stop(paste("no variables of class", names(y)[1]))
#       else y <- y[1]
    }
    y <- match.arg(y, data@DIMNAMES[[2]], several.ok = TRUE)
    for(i in 1:length(y)) {
      #     link <- data@variables[[y[i]]]@family$link
      if(i <= length(to)) fam <- do.call(to[i], args = list())
      data@variables[[y[i]]] <- change_family(y = data@variables[[y[i]]], to = fam)
    }
    return(invisible(data))
  })

setMethod("change_family", signature(data = "missing_data.frame", y = "character", to = "list"), def =
  function(data, y, to) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    if(all(y %in% names(getClass("missing_variable")@subclasses))) {
      classes <- sapply(data@variables, class)
      y <- c(sapply(y, FUN = function(x) {
        names(classes[which(classes == x)])
      }))
      if(is.list(y)) stop(paste("no variables of class", names(y)[1]))
#       else y <- y[1]
    }
    y <- match.arg(y, data@DIMNAMES[[2]], several.ok = TRUE)
    for(i in 1:length(y)) {
      data@variables[[y[i]]] <- change_family(y = data@variables[[y[i]]], to = to[[i]])
    }
    return(invisible(data))
  })

setMethod("change_family", signature(data = "missing_data.frame", y = "character", to = "family"), def =
  function(data, y, to) {
    if(all(y %in% names(getClass("missing_variable")@subclasses))) {
      classes <- sapply(data@variables, class)
      y <- c(sapply(y, FUN = function(x) {
        names(classes[which(classes == x)])
      }))
    }
    y <- match.arg(y, data@DIMNAMES[[2]], several.ok = TRUE)
    for(i in 1:length(y)) {
      data@variables[[y[i]]] <- change_family(y = data@variables[[y[i]]], to = to)
    }
    return(invisible(data))
  })

setMethod("change_family", signature(data = "missing_data.frame", y = "numeric", to = "character"), def =
  function(data, y, to, ...) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    for(i in 1:length(y)) {
      data@variables[[y]] <- change_family(y = data@variables[[y]], to = do.call(to[i], args = list(...)))
    }
    return(invisible(data))
  })

setMethod("change_family", signature(data = "missing_data.frame", y = "numeric", to = "list"), def =
  function(data, y, to) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    for(i in 1:length(y)) {
      data@variables[[y]] <- change_family(y = data@variables[[y]], to = to[[i]])
    }
    return(invisible(data))
  })

setMethod("change_family", signature(data = "missing_data.frame", y = "numeric", to = "family"), def =
  function(data, y, to) {
    for(i in 1:length(y)) {
      data@variables[[y]] <- change_family(y = data@variables[[y]], to = to)
    }
    return(invisible(data))
  })

setMethod("change_family", signature(data = "missing_data.frame", y = "logical", to = "character"), def =
  function(data, y, to) {
    if(length(y) != data@DIM[2]) {
      stop("the length of 'y' must equal the number of variables in 'data'")
    }
    return(change_family(data, which(y), to))
  })

setMethod("change_family", signature(data = "missing_data.frame", y = "logical", to = "family"), def =
  function(data, y, to) {
    if(length(y) != data@DIM[2]) {
      stop("the length of 'y' must equal the number of variables in 'data'")
    }
    return(change_family(data, which(y), to))
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

## these change the transformation and inverse_transformation slots of a continuous variable

setMethod("change_transformation", signature(data = "missing", y = "missing_variable", to = "function"), def = 
  function(y, to, inverse = FALSE) {
    if(!is(y, "continuous")) stop(paste(y@variable_name, "is not a continuous variable and hence has no transformation"))
    else if(is(y, "SC_proportion")) stop(paste(y@variable_name, "is a SC_proportion and cannot change its transformation (yet)"))
    if(inverse) {
      if(identical(to, .standardize_transform)) {
        formals(to)$mean <- mean(y@raw_data, na.rm = TRUE)
        formals(to)$sd <- sd(y@raw_data, na.rm = TRUE)
      }
      else if(identical(to, .logshift)) {
        yy <- y@raw_data
        if(any(yy < 0, na.rm = TRUE)) a <- - min(yy, na.rm = TRUE)
        else a <- 0
        a <- (a + min(yy[yy > 0], na.rm = TRUE)) / 2
        formals(to)$a <- a
      }
      if("inverse" %in% names(formals(to))) formals(to)$inverse <- TRUE
      y@inverse_transformation <- to
    }
    else {
      if(identical(to, .standardize_transform)) {
        formals(to)$mean <- mean(y@raw_data, na.rm = TRUE)
        formals(to)$sd <- sd(y@raw_data, na.rm = TRUE)
      }
      else if(identical(to, .logshift)) {
        yy <- y@raw_data
        if(any(yy < 0, na.rm = TRUE)) a <- - min(yy, na.rm = TRUE)
        else a <- 0
        a <- (a + min(yy[yy > 0], na.rm = TRUE)) / 2
        formals(to)$a <- a
      }
      y@transformation <- to
      y@data <- y@transformation(y@raw_data)
    }
    return(y)
  })

setMethod("change_transformation", signature(data = "missing", y = "missing_variable", to = "missing"), def =
  function(y) {
    if(is(y, "continuous")) cat("Likely choices include:", y@known_transformations, sep = "\n")
    else cat("No transformation possible for non-continuous variables\n")
    return(invisible(NULL))
  })

setMethod("change_transformation", signature(data = "missing_data.frame", y = "character", to = "missing"), def =
  function(data, y) {
    if(all(y %in% c("continuous", names(getClass("continuous")@subclasses)))) {
      classes <- sapply(data@variables, class)
      y <- c(sapply(y, FUN = function(x) {
        names(classes[which(classes == x)])
      }))
      if(is.list(y)) stop(paste("no variables of class", names(y)[1]))
      else y <- y[1]
    }
    y <- match.arg(y, data@DIMNAMES[[2]], several.ok = TRUE)
    for(i in 1:length(y)) change_transformation(y = data@variables[[y[i]]])
    return(data)
  })

setMethod("change_transformation", signature(data = "missing_data.frame", y = "character", to = "character"), def =
  function(data, y, to) {
    if(length(to) == 1) to <- rep(to, length(y))
    else if(length(to) != length(y)) stop("'y' and 'to' must have the same length")
    if(all(y %in% c("continuous", names(getClass("continuous")@subclasses)))) {
      classes <- sapply(data@variables, class)
      y <- c(sapply(y, FUN = function(x) {
        names(classes[which(classes == x)])
      }))
      to <- rep(to[1], length(y))
    }
    y <- match.arg(y, data@DIMNAMES[[2]], several.ok = TRUE)
    trans <- lapply(to, FUN = function(x) {
      switch(x, 
             "identity" = .identity_transform,
             "standardize" = .standardize_transform,
             "squeeze" = .squeeze_transform,
             "logshift" = .logshift,
             "log" = log,
             "sqrt" = sqrt,
             "cuberoot" = .cuberoot,
             function(...) stop(paste("must replace the transformation slot for", x)))
    })
    inverse <- lapply(to, FUN = function(x) {
      switch(x, 
             "identity" = .identity_transform,
             "standardize" = .standardize_transform,
             "squeeze" = .squeeze_transform,
             "logshift" = .logshift,
             "log" = exp,
             "sqrt" = function(y, ...) y^2,
             "cuberoot" = .cuberoot,
             function(...) stop(paste("must replace the inverse_transformation slot for", x)))
    })
    for(i in 1:length(y)) {
      data@variables[[y[i]]] <- change_transformation(y = data@variables[[y[i]]], to = trans[[i]])
      data@variables[[y[i]]] <- change_transformation(y = data@variables[[y[i]]], to = inverse[[i]], inverse = TRUE)
      
      mark <- data@index[[y[i]]][1]
      data@X[,mark] <- data@variables[[y[i]]]@data
    }
    #   initialize(data)
    return(data)
  })

setMethod("change_transformation", signature(data = "missing_data.frame", y = "numeric", to = "character"), def =
  function(data, y, to) {
    return(change_transformation(data = data, y = colnames(data)[y], to = to))
  })

setMethod("change_transformation", signature(data = "missing_data.frame", y = "logical", to = "character"), def =
  function(data, y, to) {
    if(length(y) != data@DIM[2]) {
      stop("the length of 'y' must equal the number of variables in 'data'")
    }
    return(change_transformation(data = data, y = names(data@variables)[y], to = to))
  })

setMethod("change_transformation", signature(data = "missing_data.frame", y = "character", to = "function"), def =
  function(data, y, to, inverse = stop("you must specify 'inverse = FALSE' or 'inverse = TRUE'")) {
    if(all(y %in% c("continuous", names(getClass("continuous")@subclasses)))) {
      classes <- sapply(data@variables, class)
      y <- c(sapply(y, FUN = function(x) {
        names(classes[which(classes == x)])
      }))
    }
    y <- match.arg(y, data@DIMNAMES[[2]], several.ok = TRUE)
    for(i in 1:length(y)) {
      if(inverse) data@variables[[y[i]]] <- change_transformation(y = data@variables[[y[i]]], to = to, inverse = TRUE)
      else data@variables[[y[i]]] <- change_transformation(y = data@variables[[y[i]]], to = to, inverse = FALSE)
      
      mark <- data@index[[y[i]]][1]
      data@X[,mark] <- data@variables[[y[i]]]@data
    }
    return(data)
  })

setMethod("change_transformation", signature(data = "missing_data.frame", y = "numeric", to = "function"), def =
  function(data, y, to, inverse) {
    y <- names(data@variables)[y]
    return(change_transformation(data = data, y = y, to = to, inverse))
  })

setMethod("change_transformation", signature(data = "missing_data.frame", y = "logical", to = "function"), def =
  function(data, y, to, inverse) {
    if(length(y) != data@DIM[2]) {
      stop("the length of 'y' must equal the number of variables in 'data'")
    }
    return(change_transformation(data = data, y = names(data@variables)[y], to = to, inverse = inverse))
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# Copyright (C) 2011 Douglas Bates and Martin Maechler
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

setMethod("image", "dgTMatrix", # slight hack of a method in the library(Matrix)
          function(x,
                   xlim = .5 + c(0, di[2]),
                   ylim = .5 + c(di[1], 0),
                   aspect = "iso", ## was default "fill"
                   sub = sprintf("Dimensions: %d x %d", di[1], di[2]),
                   xlab = "Column", ylab = "Row", cuts = 15,
                   useAbs = NULL, colorkey = !useAbs, col.regions = NULL,
                   lwd = NULL, ...)
          {
            di <- x@Dim
            xx <- x@x
            if(missing(useAbs)) ## use abs() when all values are non-neg
              useAbs <- min(xx, na.rm=TRUE) >= 0
            else if(useAbs)
              xx <- abs(xx)
            rx <- range(xx, finite=TRUE)
            if(is.null(col.regions))
              col.regions <-
                if(useAbs) {
                  grey(seq(from = 0.7, to = 0, length = 100))
                } else { ## no abs(.), rx[1] < 0
                  nn <- 100
                  n0 <- min(nn, max(0, round((0 - rx[1])/(rx[2]-rx[1]) * nn)))
                  col.regions <-
                    c(colorRampPalette(c("blue3", "gray80"))(n0),
                      colorRampPalette(c("gray75","red3"))(nn - n0))
                }
            if(!is.null(lwd) && !(is.numeric(lwd) && all(lwd >= 0))) # allow lwd=0
              stop("'lwd' must be NULL or non-negative numeric")
            lattice::levelplot(x@x ~ (x@j + 1L) * (x@i + 1L),
                      sub = sub, xlab = xlab, ylab = ylab,
                      xlim = xlim, ylim = ylim, aspect = aspect,
                      colorkey = colorkey, col.regions = col.regions, cuts = cuts,
                      # 		    par.settings = list(background = list(col = "transparent")),
                      panel = function(x, y, z, subscripts, at, ..., col.regions)
                      {
                        x <- as.numeric(x[subscripts])
                        y <- as.numeric(y[subscripts])
                        numcol <- length(at) - 1
                        num.r <- length(col.regions)
                        col.regions <-
                          if (num.r <= numcol)
                            rep(col.regions, length = numcol)
                        else col.regions[1+ ((1:numcol-1)*(num.r-1)) %/% (numcol-1)]
                        zcol <- rep.int(NA_integer_, length(z))
                        for (i in seq_along(col.regions))
                          zcol[!is.na(x) & !is.na(y) & !is.na(z) &
                            at[i] <= z & z < at[i+1]] <- i
                        zcol <- zcol[subscripts]
                        if (any(subscripts)) {
                          if(is.null(lwd)) {
                            wh <- grid::current.viewport()[c("width", "height")]
                            ## wh : current viewport dimension in pixel
                            wh <- c(grid::convertWidth(wh$width, "inches",
                                                       valueOnly=TRUE),
                                    grid::convertHeight(wh$height, "inches",
                                                        valueOnly=TRUE)) *
                                                          par("cra") / par("cin")
                            pSize <- wh/di ## size of one matrix-entry in pixels
                            pA <- prod(pSize) # the "area"
                            p1 <- min(pSize)
                            lwd <- ## crude for now
                              if(p1 < 2 || pA < 6) 0.01 # effectively 0
                            else if(p1 >= 4) 1
                            else if(p1 > 3) 0.5 else 0.2
                          } else stopifnot(is.numeric(lwd), all(lwd >= 0)) # allow 0
                          
                          grid::grid.rect(x = x, y = y, width = 1, height = 1,
                                           default.units = "native",
                                           gp = grid::gpar(fill = ifelse(is.na(zcol), "black", col.regions[zcol]),
                                                            lwd = lwd, col = if(lwd < .01) NA else NA))
                        }
                      }, ...)
          })

setMethod("image", signature(x = "missing_data.frame"), def =
  function (x, y.order = FALSE, x.order = FALSE,  clustered = TRUE, grayscale = FALSE, ...) {
    data <- lapply(x@variables, FUN = function(z) if(is(z, "irrelevant")) NULL else is.na(z) * 1)
    data <- as.matrix(as.data.frame(data[!sapply(data, is.null)]))
    index <- seq(nrow(data))
    x.at <- 1:nrow( data )
    x.lab <- index
    if( x.order ) {
      orderIndex <- order(colSums(data), decreasing = TRUE)
      sub <- "Ordered by number of missing items per variable" 
    }
    if( y.order ) {
      orderIndex <- order(rowSums(data), decreasing = FALSE)
      index <- row.names( data )
      sub <- "Ordered by number of missing items per observation" 
      x.at <- NULL
      x.lab <- FALSE
    }
    if(clustered){
      orderIndex <-  order.dendrogram(as.dendrogram(hclust(dist(data, method = "binary"), method="mcquitty")))
      sub <- "Clustered by missingness"
    }
    if(!grayscale) {
      data <- lapply(x@variables, FUN = function(z) {
        y <- z@data
        if(is(z, "irrelevant")) return(NULL)
        else if(is(z, "continuous"))  return( (y - mean(y, na.rm = TRUE)) / (2 * sd(y, na.rm = TRUE)) )
        else if(is(z, "count"))  return( (y - mean(y, na.rm = TRUE)) / (2 * sd(y, na.rm = TRUE)) )
        else if(is(z, "categorical")) {
          y <- z@data
          if(is(z, "binary")) {
            y <- y == max(y, na.rm = TRUE)
            return( (y - 0.5) * 2 )
          }
          else {
            the_range <- seq(from = -.99, to = 1, length.out = length(unique(na.omit(y))))
            return(the_range[as.integer(as.factor(y))])
          }
        }
        else return( (y - mean(y, na.rm = TRUE)) / (2 * sd(y, na.rm = TRUE)) ) 
      })
      data <- as.matrix(as.data.frame(data[!sapply(data, is.null)]))      
    }
    
    if(y.order) X <- Matrix(data[,orderIndex])
    else        X <- Matrix(data[orderIndex,])
    
    if(grayscale) {
      plot(image(X, aspect = "fill", xlab = "Standardized Variable", ylab = "Observation Number", sub = sub,
                 scales = list(x = list(at = 1:ncol(data), labels = colnames(data), rot = 90, abbreviate = TRUE, minlength = 8)), 
                 main = "Dark represents missing data", colorkey = FALSE, alpha.regions = 1, ...))
      return(invisible(NULL))
    }
    nn <- 100
    rx <- range(X, finite = TRUE)
    n0 <- min(nn, max(0, round((0 - rx[1])/(rx[2]-rx[1]) * nn)))
    col.regions <- heat.colors(17)
    breaks <- seq(from = rx[1] - 1e-8, to = rx[2] + 1e-8, length.out = 16)
    plot(image(X, aspect = "fill", xlab = "Standardized Variable", ylab = "Observation Number", sub = sub, at = breaks,
               scales = list(x = list(at = 1:ncol(data), labels = colnames(data), rot = 90, abbreviate = TRUE, minlength = 8)), 
               main = "Dark represents missing data", colorkey = TRUE, col.regions = col.regions, alpha.regions = 1, ...))
    
    return(invisible(NULL))
  })

setMethod("image", signature(x = "mdf_list"), def =
  function (x, y.order = FALSE, x.order = FALSE,  clustered = TRUE, grayscale = FALSE, ask = TRUE, ...) {
    if (.Device != "null device") {
      oldask <- grDevices::devAskNewPage(ask = ask)
      if (!oldask) on.exit(grDevices::devAskNewPage(oldask), add = TRUE)
      op <- options(device.ask.default = ask)
      on.exit(options(op), add = TRUE)
    }
    sapply(x, FUN = image, y.order = y.order, x.order = x.order, clustered = clustered, grayscale = grayscale, ...)
    return(invisible(NULL))
  })

setMethod("image", signature(x = "mi"), def =
  function (x, y.order = FALSE, x.order = FALSE,  clustered = TRUE, ...) {
    data <- lapply(x@data[[1]]@variables, FUN = function(z) if(is(z, "irrelevant")) NULL else is.na(z) * 1)
    data <- as.matrix(as.data.frame(data[!sapply(data, is.null)]))
    if( x.order ) {
      orderIndex <- order(colSums(data), decreasing = TRUE)
      sub <- "Ordered by number of missing items per variable" 
    }
    if( y.order ) {
      orderIndex <- order(rowSums(data), decreasing = FALSE)
      index <- row.names( data )
      sub <- "Ordered by number of missing items per observation" 
    }
    if(clustered){
      orderIndex <-  order.dendrogram(as.dendrogram(hclust(dist(data, method = "binary"), method="mcquitty")))
      sub <- "Clustered by missingness"
    }
    foo <- function(z, raw = FALSE) {
      y <- if(raw) z@raw_data else z@data
      # 	    y <- z@data
      if(is(z, "irrelevant")) return(NULL)
      else if(is(z, "continuous"))  return( (y - mean(y, na.rm = TRUE)) / (2 * sd(y, na.rm = TRUE)) )
      else if(is(z, "count"))  return( (y - mean(y, na.rm = TRUE)) / (2 * sd(y, na.rm = TRUE)) )
      else if(is(z, "categorical")) {
        y <- if(raw) as.numeric(z@raw_data) else z@data
        if(is(z, "binary")) {
          y <- y == max(y, na.rm = TRUE)
          return( (y - 0.5) * 2 )
        }
        else {
          the_range <- seq(from = -.99, to = 1, length.out = length(unique(na.omit(y))))
          return(the_range[as.integer(as.factor(y))])
        }
      }
      else return( (y - mean(y, na.rm = TRUE)) / (2 * sd(y, na.rm = TRUE)) ) 
    }
    temp <- lapply(x@data[[1]]@variables, FUN = foo)
    temp <- as.matrix(as.data.frame(temp[!sapply(temp, is.null)]))
    temp[data == 1] <- NA_real_
    data <- temp
    if(y.order) data <- data[,orderIndex]
    else        data <- data[orderIndex,]
    X0 <- Matrix(data)
    data <- 0
    chains <- min(3, length(x@data))
    for(i in seq_along(x@data)) {
      temp <- lapply(x@data[[i]]@variables, FUN = foo, raw = FALSE)
      temp <- as.matrix(as.data.frame(temp[!sapply(temp, is.null)]))
      data <- data +  temp / chains 
    }
    if(y.order) data <- data[,orderIndex]
    else        data <- data[orderIndex,]
    X1 <- Matrix(data)
    X <- rbind2(X0, X1)
    breaks <- seq(from = min(X, na.rm = TRUE), to = max(X, na.rm = TRUE), length.out = 15)
    plot(image(X0, aspect = "fill", xlab = "", ylab = "Observation Number", sub = "", at = breaks,
               scales = list(x = list(at = 1:ncol(data), labels = colnames(data), rot = 90, abbreviate = TRUE, minlength = 5)), 
               main = "Original data", colorkey = TRUE, col.regions = heat.colors(17), ...), 
         split = c(1,1,1,2))
    
    plot(image(X1, aspect = "fill", xlab = "", ylab = "Observation Number", sub = "", at = breaks,
               scales = list(x = list(at = 1:ncol(data), labels = colnames(data), rot = 90, abbreviate = TRUE, minlength = 5)), 
               main = "Average completed data", colorkey = TRUE, col.regions = heat.colors(17), ...), 
         newpage = FALSE, split = c(1,2,1,2))
    return(invisible(NULL))
  })

setMethod("image", signature(x = "mi_list"), def =
  function (x, y.order = FALSE, x.order = FALSE,  clustered = TRUE, grayscale = FALSE, ask = TRUE, ...) {
    if (.Device != "null device") {
      oldask <- grDevices::devAskNewPage(ask = ask)
      if (!oldask) on.exit(grDevices::devAskNewPage(oldask), add = TRUE)
      op <- options(device.ask.default = ask)
      on.exit(options(op), add = TRUE)
    }
    sapply(x, FUN = image, y.order = y.order, x.order = x.order, clustered = clustered, grayscale = grayscale, ...)
    return(invisible(NULL))
  })

.binnedplot <-
  function (x, y, nclass = NULL, xlab = "Expected Values", ylab = "Average residual", 
            main = "", cex.pts = 0.8, col.pts = "blue", col.int = "gray") 
  {
    n <- length(x)
    if (is.null(nclass)) {
      if (n >= 100) {
        nclass = floor(sqrt(length(x)))
      }
      if (n > 10 & n < 100) {
        nclass = 10
      }
      if (n <= 10) {
        nclass = floor(n/2)
      }
    }
    aa <- data.frame(arm::binned.resids(x, y, nclass)$binned)
    #     aa <- aa[!is.na(aa$X2se),] ## FIXME: remove once Yu-Sung fixes arm::binned.resids
    plot(range(aa$xbar), range(aa$ybar, aa$X2se, -aa$X2se), xlab = xlab, 
         ylab = ylab, type = "n", main = main, mgp = c(2, 1, 0), tcl = .05)
    abline(0, 0, lty = 2)
    lines(aa$xbar, aa$X2se, col = col.int)
    lines(aa$xbar, -aa$X2se, col = col.int)
    points(aa$xbar, aa$ybar, pch = 19, cex = cex.pts, col = col.pts)
  }

.binnedpoints <-
  function (x, y, nclass = NULL, cex.pts = 0.8, col.pts = "red") 
  {
    n <- length(x)
    if (is.null(nclass)) {
      if (n >= 100) {
        nclass = floor(sqrt(length(x)))
      }
      if (n > 10 & n < 100) {
        nclass = 10
      }
      if (n <= 10) {
        nclass = floor(n/2)
      }
    }
    if(n > 5) {
      aa <- data.frame(arm::binned.resids(x, y, nclass)$binned)
      points(aa$xbar, aa$ybar, pch = 19, cex = cex.pts, col = col.pts)
    }
    return(invisible(NULL))
  }

setMethod("plot", signature(x = "missing_data.frame", y = "missing_variable"), def =
  function(x, y, ...) {
    NAs <- is.na(y@raw_data)
    z <- y@data 
    hist(y)
    yhat <- y@fitted
    the_range <- range(c(yhat, z))
    plot(the_range, the_range, type = "n", xlab = "Expected Values", ylab = "Completed",
         mgp = c(2, 1, 0), tcl = .05)
    abline(0, 1, lty = 2, col = "lightgray")
    points(yhat, z, col = ifelse(NAs, "red", "blue"), pch = ".", cex = 2)
    lines(lowess(x = yhat[!NAs], y = z[!NAs]), col = "blue")
    .binnedplot(yhat[!NAs], (y@data - yhat)[!NAs])
    .binnedpoints(yhat[NAs], (y@data - yhat)[NAs])
    return(invisible(NULL))
  })

setMethod("plot", signature(x = "missing_data.frame", y = "categorical"), def =
  function(x, y, ...) {
    NAs <- is.na(y@raw_data)
    z <- y@data
    hist(y)
    s <- nrow(y@parameters) + 1
    yhat <- y@fitted
    if(length(yhat) == 0) { # embedded
      varname <- y@variable_name
      varname <- strsplit(varname, ":")[[1]][1]
      to_drop <- x@index[[varname]]
      X <- x@X[,-to_drop]
      s <- nrow(y@parameters) + 1
      model <- fit_model(y, data = x, s = s, warn = TRUE, X = X)
      yhat <- fitted(model)
    }
    if(is.matrix(yhat)) yhat <- yhat %*% (1:ncol(yhat))
    the_range <- range(c(yhat, z)) #+ c(-.1, .1)
    plot(the_range, the_range, type = "n", xlab = "Expected Values", ylab = "Completed (jittered)",
         mgp = c(2, 1, 0), tcl = .05)
    abline(0, 1, lty = 2, col = "lightgray")
    points(yhat, jitter(z), col = ifelse(NAs, "red", "blue"), pch = ".", cex = 2)
    .binnedplot(yhat[!NAs], (y@data - yhat)[!NAs])
    .binnedpoints(yhat[NAs], (y@data - yhat)[NAs])
    return(invisible(NULL))
  })

setMethod("plot", signature(x = "missing_data.frame", y = "binary"), def =
  function(x, y, ...) {
    
    NAs <- is.na(y@raw_data)
    z <- y@data - 1L
    hist(y)
    s <- nrow(y@parameters) + 1
    yhat <- y@fitted
    if(length(yhat) == 0) { # embedded
      varname <- y@variable_name
      varname <- strsplit(varname, ":")[[1]][1]
      to_drop <- x@index[[varname]]
      X <- x@X[,-to_drop]
      model <- fit_model(y = y, data = x, s = s, warn = TRUE, X = X)
      yhat <- fitted(model)
    }
    the_range <- range(c(yhat, z)) #+ c(-.1, .1)
    plot(the_range, the_range, type = "n", xlab = "Expected Values", ylab = "Completed (jittered)",
         mgp = c(2, 1, 0), tcl = .05)
    abline(0, 1, lty = 2, col = "lightgray")
    points(yhat, jitter(z), col = ifelse(NAs, "red", "blue"), pch = ".", cex = 2)
    .binnedplot(yhat[!NAs], (y@data - 1 - yhat)[!NAs])
    .binnedpoints(yhat[NAs], (y@data - 1 - yhat)[NAs])
    return(invisible(NULL))
  })

setMethod("plot", signature(x = "allcategorical_missing_data.frame", 
                            y = "categorical"), def =
            function(x, y, ...) {
  NAs <- is.na(y@raw_data)
  z <- y@raw_data
  hist(y)
  latents <- x@latents@data
  yhat <- t(sapply(latents[!NAs], FUN = function(l) y@fitted[l,]))
  tab_obs <- table(z[!NAs])
  tab_model <- table(apply(yhat, 1, FUN = function(p) which(rmultinom(1, 1, p) == 1)))  
  the_range <- c(0, max(c(tab_obs, tab_model)))
  barplot(tab_obs, beside = TRUE, xlab = "Observed Values", ylim = the_range)
  names(tab_model) <- levels(z)
  barplot(tab_model, beside = TRUE, xlab = "Expected Values", ylim = the_range)
  return(invisible(NULL))
})

setMethod("plot", signature(x = "allcategorical_missing_data.frame", 
                            y = "binary"), def =
            function(x, y, ...) {
  NAs <- is.na(y@raw_data)
  z <- y@raw_data
  hist(y)
  latents <- x@latents@data
  yhat <- t(sapply(latents[!NAs], FUN = function(l) y@fitted[l,]))
  tab_obs <- table(z[!NAs])
  tab_model <- table(apply(yhat, 1, FUN = function(p) which(rmultinom(1, 1, p) == 1)))  
  the_range <- c(0, max(c(tab_obs, tab_model)))
  barplot(tab_obs, beside = TRUE, xlab = "Observed Values", ylim = the_range)
  names(tab_model) <- levels(z)
  barplot(tab_model, beside = TRUE, xlab = "Expected Values", ylim = the_range)
  return(invisible(NULL))
})

setMethod("plot", signature(x = "missing_data.frame", y = "semi-continuous"), def =
  function(x, y, ...) {
    NAs <- is.na(y@raw_data)
    z <- y@data 
    hist(y)
    s <- nrow(y@parameters) + 1
    yhat <- z
    yhat[complete(y@indicator, m = 0L, to_factor = TRUE) == 0] <- y@fitted #fitted(model)
    the_range <- range(c(yhat, z))
    plot(the_range, the_range, type = "n", xlab = "Expected Values", ylab = "Completed",
         mgp = c(2, 1, 0), tcl = .05)
    abline(0, 1, lty = 2, col = "lightgray")
    points(yhat, z, col = ifelse(NAs, "red", "blue"), pch = ".", cex = 2)
    lines(lowess(x = yhat[!NAs], y = z[!NAs]), col = "blue")
    .binnedplot(yhat[!NAs], (y@data - yhat)[!NAs])
    .binnedpoints(yhat[NAs], (y@data - yhat)[NAs])
    return(invisible(NULL))
  })

setMethod("plot", signature(x = "mi", y = "ANY"), def = 
  function(x, y, ask = TRUE, header = character(0), ...) {
    if(missing(y))           select <- 1:ncol(x@data[[1]])
    else if(is.logical(y))   select <- which(y)
    else if(is.character(y)) select <- which(colnames(x@data[[1]]) %in% y)
    else if(is.numeric(y))   select <- which(1:nrow(x@data[[1]]) %in% y)
    
    for(i in seq_along(x@data[[1]]@variables)) {
      if(x@data[[1]]@no_missing[i]) next
      else if(is(x@data[[1]]@variables[[i]], "irrelevant")) next
      else if(x@data[[1]]@variables[[i]]@imputation_method == "mcar") {
        warning(x@data[[1]]@variables[[i]]@variable_name, " not plotted because it assumes MCAR")
        next
      }
      if(!(i %in% select)) next
      l <- min(3, length(x@data))
      
      if (.Device != "null device") {
        oldask <- grDevices::devAskNewPage(ask = ask)
        if (!oldask) on.exit(grDevices::devAskNewPage(oldask), add = TRUE)
        op <- options(device.ask.default = ask)
        on.exit(options(op), add = TRUE)
      }
      
      par(mfrow = c(l,3), mar = c(5,4,1,1) + .1)
      if(is(x@data[[1]]@variables[[i]], "semi-continuous")) {
        for(j in 1:l) plot(x@data[[j]], x@data[[j]]@variables[[i]]@indicator, ...)
        title(main = paste("\n", header, x@data[[1]]@variables[[i]]@indicator@variable_name, sep = ""), outer = TRUE)
      }
      for(j in 1:l) plot(x@data[[j]], x@data[[j]]@variables[[i]], ...)      
      new_header <- paste(header, x@data[[1]]@variables[[i]]@variable_name)
      if(is(x@data[[1]]@variables[[i]], "continuous")) {
        trans <- .show_helper(x@data[[1]]@variables[[i]])$transformation[1]
        new_header <- paste("\n", new_header, " (", trans, ")", sep = "")
      }
      else new_header <- paste("\n", new_header, sep = "")
      title(main = new_header, outer = TRUE)
    }
    return(invisible(NULL))
  })

setMethod("plot", signature(x = "mi_list", y = "ANY"), def = 
  function(x, y, ask = TRUE, ...) {
    if(missing(y)) for(i in seq_along(x)) plot(x[[i]], ask = ask, header = paste(names(x)[i], ": ", sep = ""), ...)
    else for(i in seq_along(x)) plot(x[[i]], y = y, ask = ask, header = paste(names(x)[i], ": ", sep = ""), ...)
    return(invisible(NULL))
  })

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# This is superceded by the tobin5() function below
tobin5 <-
  function(mdf, y, f = NULL) {
    if(!is(mdf, "missing_data.frame")) stop("'mdf' must be a 'missing_data.frame'")
    if(!is.character(y)) stop("'y' must be a character string")
    if(length(y) != 1) stop("'y' must have length one")
    if(!(y %in% colnames(mdf))) stop("'y' must be a variable in 'mdf'")
    
    y <- mdf@variables[[y]]
    NAs <- is.na(y)
    to_drop <- mdf@index[[y@variable_name]]
    X <- mdf@X[,-to_drop]
    
    probit <- bayesglm.fit(X, y = NAs, family = binomial(link = "probit"))
    class(probit) <- c("bayesglm", "glm", "lm")
    gamma <- coef(probit)
    IMR_0 <- dnorm(-fitted(probit)) / pnorm(-fitted(probit))
    IMR_1 <- dnorm( fitted(probit)) / pnorm( fitted(probit))
    
    if(is.null(f)) {
      mark <- colnames(mdf@X)[!grepl("^missing_", colnames(mdf@X))][-1]
      mark <- mark[mark != y@variable_name]
      f <- paste(mark, collapse = " + ")
      f <- paste(y@variable_name, " ~ ", f, " + IMR", sep = "")
      f <- as.formula(f)
    }
    else if(!is(f, "formula")) stop("'f' must be 'NULL' or a formula")
    
    df <- as.data.frame(cbind(mdf@X, IMR = IMR_0))
    if(is(y, "continuous")) {
      model_0 <- bayesglm(f, family = gaussian, data = df, subset = !NAs)
    }
    else stop("only continuous dependent variables are supported at the moment")
    
    df <- as.data.frame(cbind(mdf@X, IMR = IMR_1)) 
    if(is(y, "continuous")) {
      model_1 <- bayesglm(f, family = gaussian, data = df, subset = NAs)
    }
    
    se_0 <- model_0$dispersion
    se_1 <- model_1$dispersion
    
    delta_0 <- IMR_0^2 - fitted(probit) * IMR_0
    delta_1 <- IMR_1^2 + fitted(probit) * IMR_1
    
    betaL_0 <- coef(model_0)
    betaL_0 <- betaL_0[length(betaL_0)]
    betaL_1 <- coef(model_1)
    betaL_1 <- betaL_1[length(betaL_1)]
    
    sigma_0 <- sqrt(se_0^2 + (betaL_0 * delta_0)^2)
    sigma_1 <- sqrt(se_1^2 + (betaL_1 * delta_1)^2)
    
    rho_0 <- -betaL_0 / sigma_0
    rho_1 <-  betaL_1 / sigma_1
    
    ## FIXME: correct vcov(model_0) and vcov(model_1) now
    
    return(list(probit = probit, model_0 = model_0, model_1 = model_1,
                rho_0 = rho_0, rho_1 = rho_1))
  }

tobin5 <-
  function(imputations, y, f = NULL) {
    if(!is(imputations, "mi")) stop("'imputations' must be a 'mi' object")
    if(!is.character(y)) stop("'y' must be a character string")
    if(length(y) != 1) stop("'y' must have length one")
    if(!(y %in% colnames(imputations))) stop("'y' must be a variable in 'imputations'")
    
    dfs <- complete(imputations)
    mdf <- imputations@data[[1]]
    to_drop <- mdf@index[[y@variable_name]]
    cn <- colnames(mdf@X[,-to_drop])[-1]
    f1 <- paste(cn, collapse = " + ")
    NAs <- is.na(mdf@variables[[y]])
    if(paste("missing", y, sep = "_") %in% colnames(mdf@X)) {
      f1 <- paste(paste("missing", y, sep = "_"), "~", f1)
    }
    else for(i in seq_along(dfs)) {
      dfs[[i]] <- cbind(dfs[[i]], NAs)
      colnames(dfs[[i]]) <- c(colnames(dfs[[i]]), paste("missing", y, sep = "_"))
    }
    f1 <- as.formula(f1)
    probit <- pool(f1, data = dfs, family = binomial(link = "probit"))
    
    gamma <- sapply(probit@models, coef)
    Pr <- sapply(probit@models, fitted)
    IMR_0 <- apply(Pr, 2, FUN = function(p) dnorm(-p) / pnorm(-p))
    IMR_1 <- apply(Pr, 2, FUN = function(p) dnorm( p) / pnorm( p))
    
    if(is.null(f)) {
      mark <- colnames(mdf@X)[!grepl("^missing_", colnames(mdf@X))][-1]
      mark <- mark[mark != y]
      f <- paste(mark, collapse = " + ")
      f <- paste(y@variable_name, " ~ ", f, " + IMR", sep = "")
      f <- as.formula(f)
    }
    else if(!is(f, "formula")) stop("'f' must be 'NULL' or a formula")
    
    if(!is(mdf@variables[[y]], "continuous")) {
      stop("only continuous dependent variables are supported at the moment")
    }
    
    for(i in seq_along(dfs)) dfs[[i]]$IMR <- IMR_0[,i]
    model_0 <- pool(f, data = dfs, family = gaussian, subset = !NAs)
    
    for(i in seq_along(dfs)) dfs[[i]]$IMR <- IMR_1[,i]
    model_1 <- pool(f, data = dfs, family = gaussian, subset = NAs)
    
    se_0 <- sapply(model_0@models, FUN = function(m) m$dispersion)
    se_1 <- sapply(model_1@models, FUN = function(m) m$dispersion)
    
    delta_0 <- IMR_0^2 - Pr * IMR_0
    delta_1 <- IMR_1^2 + Pr * IMR_1
    
    betaL_0 <- sapply(model_0@models, coef)
    betaL_0 <- betaL_0[nrow(betaL_0)]
    betaL_1 <- sapply(model_1@models, coef)
    betaL_1 <- betaL_1[nrow(betaL_1)]
    
    sigma_0 <- sqrt(se_0^2 + sweep(delta_0, 2, betaL_0, FUN = "*")^2)
    sigma_1 <- sqrt(se_1^2 + sweep(delta_1, 2, betaL_1, FUN = "*")^2)
    
    rho_0 <- -betaL_0 / sigma_0
    rho_1 <-  betaL_1 / sigma_1
    
    ## FIXME: correct vcov(model_0) and vcov(model_1) now
    
    return(list(probit = probit, model_0 = model_0, model_1 = model_1,
                rho_0 = rho_0, rho_1 = rho_1))
  }

# Part of the mi package for multiple imputation of missing data
# Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Trustees of Columbia University
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


## these fit a regression and return the model

# note, helper functions are good because they are checked more rigorously by R CMD check

setMethod("fit_model", signature(y = "missing_variable", data = "missing_data.frame"), def =
  function(y, data, s, warn, ...) {
    stop("This method should not have been called. You need to define the relevant fit_model() S4 method")
  })

setMethod("fit_model", signature(y = "irrelevant", data = "missing_data.frame"), def = 
  function(y, data, ...) {
    stop("'fit_model' should not have been called on an 'irrelevant' variable")
  })

setMethod("fit_model", signature(y = "binary", data = "missing_data.frame"), def =
  function(y, data, s, warn, X = NULL, ...) {
    if(is.null(X)) {
      to_drop <- data@index[[y@variable_name]]
      if(length(to_drop)) X <- data@X[,-to_drop]
      else X <- data@X[,]
      if(is(data, "experiment_missing_data.frame")) {
        treatment <- names(which(data@concept == "treatment"))
        if(data@concept[y@variable_name] == "outcome") {
          X <- cbind(X, interaction = X * data@variables[[treatment]]@data)
        }
      }
    }
    
    if(s > 1)       start <- y@parameters[s-1,]
    else if(s < -1) start <- y@parameters[1,]
    else start <- NULL
    start <- NULL
    weights <- if(length(data@weights) == 1) data@weights[[1]] else data@weights[[y@variable_name]]
    CONTROL <- list(epsilon = max(1e-8, exp(-abs(s))), maxit = 25, trace = FALSE)
    priors <- data@priors[[y@variable_name]]
    out <- bayesglm.fit(X, y@data - 1L, weights = weights,
                                   prior.mean = priors[[1]], prior.scale = priors[[2]], prior.df = priors[[3]],
                                   prior.mean.for.intercept = priors[[4]], prior.scale.for.intercept = priors[[5]],
                                   prior.df.for.intercept = priors[[6]],
                                   start = start, family = y@family, Warning = FALSE, control = CONTROL)
    if(warn && !out$converged) {
      warning(paste("bayesglm() did not converge for variable", y@variable_name, "on iteration", abs(s)))
    }
    if(any(abs(coef(out)) > 100)) {
      warning(paste(y@variable_name, ": separation on iteration", abs(s)))
    }
    out$x <- X
    class(out) <- c("bayesglm", "glm", "lm")
    return(out)
  })

.fit_MNL <-
  function(y, X, weights) {
    model<-nnet::multinom(y@data ~ X -1, weights = weights, Hess = y@imputation_method == "ppd", 
                           model = TRUE, trace = FALSE, MaxNWts = 10000)
    return(model)
  }

.fit_RNL <-
  function(y, X, weights, CONTROL) {
    if (y@use_NA==TRUE) values <- c(-.Machine$integer.max, 1:length(y@levels))  
    else values <- 1:length(y@levels)
    out <- sapply(values, simplify = FALSE, FUN = function(l) {
      model <- bayesglm.fit(X, y@data == l, weights = weights, 
                            family = y@family, control=CONTROL)
      model$x <- X # bayesglm.fit() by default does not retain the model matrix it uses 
      class(model) <- c("bayesglm", "glm", "lm")  
      return(model)
    })
    class(out) <- "RNL"
    return(out)
  }

setMethod("fit_model", signature(y = "unordered-categorical", data = "missing_data.frame"), def =
  function(y, data, warn, s, ...) {
    to_drop <- data@index[[y@variable_name]]
    if (y@use_NA) {
      y@data[y@which_drawn] <- -.Machine$integer.max # make NAs the smallest possible signed integer 
    }
	if(length(to_drop)) X <- data@X[,-to_drop]
    else X <- data@X[,]
    if(is(data, "experiment_missing_data.frame")) {
      treatment <- names(which(data@concept == "treatment"))
      if(data@concept[y@variable_name] == "outcome") {
        X <- cbind(X, interaction = X * data@variables[[treatment]]@data)
      }
    }
    weights <- if(length(data@weights) == 1) data@weights[[1]] else data@weights[[y@variable_name]]
    if(y@estimator == "MNL") {
      out <- .fit_MNL(y, X, weights)
      data@X
    }
  	else if(y@estimator == "RNL"){
      CONTROL <- list(epsilon = max(1e-8, exp(-abs(s))), maxit = 25, trace = FALSE)
      out <- .fit_RNL(y, X, weights, CONTROL)
      data@X
    }
    else stop("estimator not recognized")
    return(out)
  })

.clogit <- # similar to the survival::clogit function
  function(formula, data, n, method, weights, subset, 
           x = TRUE, na.action = "na.exclude") {
    coxcall <- match.call()
    coxcall[[1]] <- as.name("coxph")
    newformula <- formula
    newformula[[2]] <- substitute(survival::Surv(rep(1, nn), case), 
                                  list(case = formula[[2]], nn = n))
    environment(newformula) <- environment(formula)
    coxcall$formula <- newformula
    coxcall$n <- NULL
    coxcall <- eval(coxcall, sys.frame(sys.parent()))
    coxcall$userCall <- sys.call()
    class(coxcall) <- c("clogit", "coxph")
    coxcall
  }

setMethod("fit_model", signature(y = "grouped-binary", data = "missing_data.frame"), def =
  function(y, data, s, warn) {
    # see http://www.stata.com/support/faqs/stat/clogitcl.html for a good explanation of this model
    to_drop <- data@index[[y@variable_name]]
    X <- data@X[,-to_drop]
    weights <- if(length(data@weights) == 1) data@weights[[1]] else data@weights[[y@variable_name]]
    groups <- sapply(y@strata, FUN = function(x) complete(data@variables[[x]], m = 0L), simplify = FALSE)  
    out <- .clogit(y@data ~ X + strata(groups), method = "breslow", weights = weights, n = nrow(X)) 
    out$x <- X
    return(out)
  })

setMethod("fit_model", signature(y = "ordered-categorical", data = "missing_data.frame"), def =
  function(y, data, s, warn, X = NULL, ...) {
    if(is.null(X)) {
      to_drop <- data@index[[y@variable_name]]
      if(length(to_drop)) X <- data@X[,-to_drop]
      else X <- data@X[,]
      if(is(data, "experiment_missing_data.frame")) {
        treatment <- names(which(data@concept == "treatment"))
        if(data@concept[y@variable_name] == "outcome") {
          X <- cbind(X, interaction = X * data@variables[[treatment]]@data)
        }
      }
      X <- X[,-1]
    }
    
    method <- if(y@family$link == "logit") "logistic" else y@family$link
    start <- NULL
    start <- c(rep(0, ncol(X)), qlogis(cumsum(table(y@data)) / nrow(X)))
    start <- start[-length(start)]
    weights <- if(length(data@weights) == 1) data@weights[[1]] else data@weights[[y@variable_name]]
    CONTROL <- list(reltol = max(1e-8, exp(-abs(s))))
    priors <- data@priors[[y@variable_name]]
    out <- bayespolr(as.ordered(y@data) ~ X, weights = weights, method = method,
                                prior.mean = priors[[1]], prior.scale = priors[[2]], prior.df = priors[[3]],
                                prior.counts.for.bins = priors[[4]], control = list(reltol = max(1e-8, exp(-abs(s)))), ...)
    
    if(warn && out$convergence != 0) {
      warning(paste("bayespolr() did not converge for variable", y@variable_name, "on iteration", abs(s)))
    }
    out$x <- X
    return(out)
  })

setMethod("fit_model", signature(y = "interval", data = "missing_data.frame"), def =
  function(y, data, s, warn, ...) {
    stop("FIXME: write this method")
  })

## helper function
.fit_continuous <-
  function(y, data, s, warn, X, subset = 1:nrow(X)) {
    weights <- if(length(data@weights) == 1) data@weights[[1]] else data@weights[[y@variable_name]]
    if(!is.null(weights)) weights <- weights[subset]
    
    if(s > 1)       start <- y@parameters[s-1,]
    else if(s < -1) start <- y@parameters[1,]
    else start <- NULL
    start <- NULL
    mark <- c(TRUE, apply(X[subset,-1, drop = FALSE], 2, FUN = function(x) length(unique(x)) > 1))
    if(!all(mark)) {
      if(abs(s) == 1) {
        stop(paste(y@variable_name, ": imputed values on iteration 0 randomly inadmissible; try mi() again with different seed"))
      }
      X <- X[,mark]
      if(!is.null(start)) start <- start[mark]
    }
    CONTROL <- list(epsilon = max(1e-8, exp(-abs(s))), maxit = 25, trace = FALSE)
    priors <- data@priors[[y@variable_name]]
    out <- bayesglm.fit(X[subset,], y@data[subset], weights = weights, start = start, family = y@family, 
                                   prior.mean = priors[[1]], prior.scale = priors[[2]], prior.df = priors[[3]],
                                   prior.mean.for.intercept = priors[[4]], prior.scale.for.intercept = priors[[5]],
                                   prior.df.for.intercept = priors[[6]], Warning = FALSE, control = CONTROL)
    
    if(warn && !out$converged) {
      warning(paste("bayesglm() did not converge for variable", y@variable_name, "on iteration", abs(s)))
    }
    out$x <- X
    class(out) <- c("bayesglm", "glm", "lm")
    return(out)
  }

setMethod("fit_model", signature(y = "continuous", data = "missing_data.frame"), def =
  function(y, data, s, warn, ...) {
    to_drop <- data@index[[y@variable_name]]
    if(length(to_drop)) X <- data@X[,-to_drop]
    else X <- data@X[,]
    if(is(data, "experiment_missing_data.frame")) {
      treatment <- names(which(data@concept == "treatment"))
      if(data@concept[y@variable_name] == "outcome") {
        X <- cbind(X, interaction = X * data@variables[[treatment]]@data)
      }
    }
    return(.fit_continuous(y, data, s, warn, X))
  })

# setMethod("fit_model", signature(y = "truncated-continuous", data = "missing_data.frame"), def =
# function(y, data, s, warn, ...) {
#   stop("FIXME: write this method using library(survival)")
# })
# 
# setMethod("fit_model", signature(y = "censored-continuous", data = "missing_data.frame"), def =
# function(y, data, s, warn, ...) {
#   stop("FIXME: mi does not do censored-continuous variables yet")
#   to_drop <- data@index[[y@variable_name]]
#   X <- cbind(y@raw_data, data@X[,-to_drop])
#   if(is(data, "experiment_missing_data.frame")) {
#     treatment <- names(which(data@concept == "treatment"))
#     if(data@concept[y@variable_name] == "outcome") {
#       X <- cbind(X, interaction = X * data@variables[[treatment]]@data)
#     }
#   }  
# })

setMethod("fit_model", signature(y = "semi-continuous", data = "missing_data.frame"), def =
  function(y, data, s, warn, ...) {
    stop("the semi-continuous fit_model() method should not have been called")
  })

setMethod("fit_model", signature(y = "nonnegative-continuous", data = "missing_data.frame"), def =
  function(y, data, s, warn, ...) {  
    to_drop <- data@index[[y@variable_name]]
    if(length(to_drop)) X <- data@X[,-to_drop]
    else X <- data@X[,]
    model <- fit_model(y@indicator, data, s, warn, X)
    if(abs(s) > 1) subset <- complete(y@indicator, m = 0L, to_factor = TRUE) == 0
    else subset <- 1:nrow(X)
    return(.fit_continuous(y = y, data = data, s = s, warn = warn, X = X, subset = subset))
  })

setMethod("fit_model", signature(y = "SC_proportion", data = "missing_data.frame"), def =
  function(y, data, s, warn, ...) {
    to_drop <- data@index[[y@variable_name]]
    if(length(to_drop)) X <- data@X[,-to_drop]
    else X <- data@X[,]
    model <- fit_model(y@indicator, data, s, warn, X)
    if(abs(s) > 1) subset <- complete(y@indicator, m = 0L, to_factor = TRUE) == 0
    else subset <- 1:nrow(X)
    return(.fit_proportion(y = y, data = data, s = s, warn = warn, X = X, subset = subset))
  })

## helper function
.fit_proportion <-
  function(y, data, s, warn, X, subset = 1:nrow(X)) {
    weights <- if(length(data@weights) == 1) data@weights[[1]] else data@weights[[y@variable_name]]
    if(!is.null(weights)) weights <- weights[subset]
    
    if(s > 1)       start <- y@parameters[s-1,]
    else if(s < -1) start <- y@parameters[1,]
    else start <- NULL
    start <- NULL
    mark <- c(TRUE, apply(X[subset,-1, drop = FALSE], 2, FUN = function(x) length(unique(x)) > 1))
    if(!all(mark)) {
      if(abs(s) == 1) {
        stop(paste(y@variable_name, ": imputed values on iteration 0 randomly inadmissible; try mi() again with a different seed"))
      }
      X <- X[,mark]
      if(!is.null(start)) start <- start[c(mark, TRUE)]
    }
    out <- betareg::betareg.fit(X[subset,], y@data[subset], weights = if(!is.null(weights)) weights[subset],
                                      link = y@family$link, link.phi = y@link.phi, 
                                      control = betareg::betareg.control(reltol = 1e-8, start = start, fsmaxit = 0))
    if(warn && !out$converged) {
      warning(paste("betareg() did not converge for variable", y@variable_name, "on iteration", abs(s)))
    }
    out$x <- X
    class(out) <- c("betareg")
    return(out)
  }

setMethod("fit_model", signature(y = "proportion", data = "missing_data.frame"), def =
  function(y, data, s, warn, ...) {
    to_drop <- data@index[[y@variable_name]]
    if(length(to_drop)) X <- data@X[,-to_drop]
    else X <- data@X[,]
    if(y@family$family == "gaussian") out <- .fit_continuous(y, data, s, warn, X)
    else out <- .fit_proportion(y, data, s, warn, X)
    return(out)
  })

setMethod("fit_model", signature(y = "count", data = "missing_data.frame"), def =
  function(y, data, s, warn, ...) {
    to_drop <- data@index[[y@variable_name]]
    if(length(to_drop)) X <- data@X[,-to_drop]
    else X <- data@X[,]
    return(.fit_continuous(y, data, s, warn, X)) # even though counts are not continuous
  })

## experiments
setMethod("fit_model", signature(y = "missing_variable", data = "experiment_missing_data.frame"), def =
  function(y, data, ...) {
    stop("you need to write a specific fit_model() method for the", class(y), "class")
  })

setMethod("fit_model", signature(y = "continuous", data = "experiment_missing_data.frame"), def =
  function(y, data, s, warn, ...) {
    to_drop <- data@index[[y@variable_name]]
    ## For each case, make an X matrix based on the giant matrix in data@X
    if(data@case == "outcomes") { # missingness on outcome(s) only
      if(length(to_drop)) X <- data@X[,-to_drop]
      else X <- data@X[,]
      treatment_name <- names(data@concept[data@concept == "treatment"])
      X <- cbind(X, interaction = X[,!(colnames(X) %in% c("(Intercept)", treatment_name))] * X[,treatment_name])
    }
    else if(data@case == "covariates") { # missingness on covariate(s) only
      to_drop <- c(to_drop, which(data@concept == "treatment"))
      if(length(to_drop)) X <- data@X[,-to_drop]
      else X <- data@X[,]
    }
    else { # missing on both outcome(s) and covariate(s)
      if(data@concept[y@variable_name] == "covariate") {
        to_drop <- c(to_drop, which(data@concept == "treatment"))
      }
      if(length(to_drop)) X <- data@X[,-to_drop]
      else X <- data@X[,]
    }
    return(mi::.fit_continuous(y, data, s, warn, X))
  })

## here y indicates which variable to model
setMethod("fit_model", signature(y = "character", data = "mi"), def = 
  function(y, data, m = length(data@data), ...) {
    s <- sum(data@total_iters) + 1
    if(length(m) == 1) {
      models <- vector("list", m)
      for(i in 1:m) {
        model <- data@data[[i]]@variables[[y]]@model
        if(is.null(model)) {
          model <- fit_model(y = data@data[[i]]@variables[[y]], data = data@data[[i]], s = s, warn = TRUE, ...)
          if(!isS4(model)) model$x <- model$X <- model$y <- NULL
        }
        models[[i]] <- model
      }
    }
    else {
      models <- vector("list", length(m))
      models <- for(i in 1:length(m)) {
        if(is.null(data@data[[i]]@variables[[y]]@model)) {
          models[[m[i]]] <- fit_model(y = data@data[[m[i]]]@variables[[y]], data = data@data[[m[i]]], s = s, warn = TRUE, ...)
        }
        else models[[i]] <- data@data[[i]]@variables[[y]]@model
      }
    }
    return(models)
  })

## fit all variables with missingness
setMethod("fit_model", signature(y = "missing", data = "mi"), def = 
  function(data, m = length(data@data)) {
    varnames <- names(data@data[[1]]@variables)
    exclude <- data@data[[1]]@no_missing | 
               sapply(data@data[[1]], FUN = function(y) is(y, "irrelevant"))
    models <- sapply(varnames, simplify = FALSE, FUN = function(v) {
                if(v %in% exclude) paste(v, "not modeled") ## maybe just skip these? 
                else fit_model(y = v, data = data, m = m)
                })
    return(models)
  })

## fit all elements of a mdf_list
setMethod("fit_model", signature(y = "missing", data = "mdf_list"), def =
  function(data, s = -1, verbose = FALSE, warn = FALSE, ...) {
    out <- lapply(data, fit_model, s = s, verbose = verbose, warn = warn, ...)
    class(out) <- "mdf_list"
    return(out)
  })

.fit_model_y <-
  function(y, data, s, verbose, warn, ...) {
    if(s != 0 && y@imputation_method != "mcar") {
      if(is(y, "semi-continuous")) {
        to_drop <- data@index[[y@variable_name]]
        if(length(to_drop)) X <- data@X[,-to_drop]
        else X <- data@X[,]
        model <- fit_model(y = y@indicator, data = data, s = s, warn = warn, X = X)
        indicator <-    mi(y = y@indicator, model = model, s = ifelse(s < 0, 1L, s))
        if(s > 1) indicator@parameters[s,] <- get_parameters(model)
        else if(abs(s) == 1) {
          parameters <- get_parameters(model)
          rows <- if(s == 1) nrow(indicator@parameters) else 1
          if(ncol(indicator@parameters) == 0) {
            temp <- matrix(NA_real_, nrow = rows, ncol = length(parameters))
          }
          temp[1,] <- parameters
          indicator@parameters <- temp
        }
        else indicator@parameters[1,] <- get_parameters(model)
        y@indicator <- indicator
      }
      model <- fit_model(y = y, data = data, s = s, warn = warn)
      y <- mi(y = y, model = model, s = ifelse(s < 0, 1L, s))
    }
    else y <- mi(y = y)
    if(y@imputation_method == "mcar") {
      # do nothing
    }
    else if(s > 1) {
      parameters <- get_parameters(model)
      if(length(parameters) != ncol(y@parameters)) parameters <- y@parameters[s-1,] # scary
      y@parameters[s,] <- parameters
    }
    else if(abs(s) == 1) {
      parameters <- get_parameters(model)
      rows <- if(s == 1) nrow(y@parameters) else 1
      if(ncol(y@parameters) == 0) {
        temp <- matrix(NA_real_, nrow = rows, ncol = length(parameters))
      }
      temp[1,] <- parameters
      y@parameters <- temp
    }
    else if(s != 0) {
      parameters <- get_parameters(model)
      if(length(parameters) == ncol(y@parameters)) y@parameters[s,] <- parameters
    }
    
    return(y)
  }

.update_X <-
  function(y, data) {
    which_drawn <- y@which_drawn
    varname <- y@variable_name
    if(is(y, "categorical")) {
      dummies <- .cat2dummies(y)[which_drawn,,drop = FALSE]
      data@X[ which_drawn, data@index[[varname]][1:NCOL(dummies)]] <- dummies
    }
    else if(is(y, "semi-continuous")) {
      mark <- data@index[[varname]]
      data@X[ which_drawn, mark[1]  ] <- y@data[which_drawn]
      dummies <- .cat2dummies(y@indicator)
      data@X[ which_drawn, mark[1 + 1:NCOL(dummies)] ] <- dummies[which_drawn,,drop = FALSE]
    }
    else if(is(y, "censored_continuous")) {
      temp <- y@data[which_drawn]
      if(y@n_lower) temp <- cbind(temp, lower = y@lower_indicator@data[y@which_drawn])
      if(y@n_upper) temp <- cbind(temp, upper = y@upper_indicator@data[y@which_drawn])
      data@X[ which_drawn, data@index[[varname]][1:NCOL(temp)]] <- temp
      data@X[ y@which_censored, data@index[[varname]][1] ] <- y@data[y@which_censored]
    }
    else if(is(y, "truncated_continuous")) {
      temp <- y@data[which_drawn]
      if(y@n_lower) temp <- cbind(temp, lower = y@lower_indicator@data[y@which_drawn])
      if(y@n_upper) temp <- cbind(temp, upper = y@upper_indicator@data[y@which_drawn])
      data@X[ which_drawn, data@index[[varname]][1:NCOL(temp)]] <- temp
      data@X[ y@which_truncated, data@index[[varname]][1] ] <- y@data[y@which_truncated]
    }
    else data@X[ which_drawn, data@index[[varname]][1] ] <- y@data[which_drawn]
    return(data)    
  }

.fit_model_mdf <-
  function(data, s, verbose, warn, ...) {
    if(verbose) {
      txt <- paste("Iteration:", abs(s))
      if(isatty(stdout()) && !(any(search() == "package:gWidgets"))) cat("\n", txt)
      else cat("<br>", txt, file = file.path(data@workpath, "mi.html"), append = TRUE)
    }
    on.exit(print("the problematic variable is"))
    on.exit(show(y), add = TRUE)
    for(jj in sample(1:ncol(data), ncol(data), replace = FALSE)) {
      y <- data@variables[[jj]]
      if(y@all_obs) next
      if(is(y, "irrelevant")) next
      y <- .fit_model_y(y, data, s, verbose, warn, ...)
      data <- .update_X(y, data)
      data@variables[[jj]] <- y
      if(verbose) {
        txt <- "."
        if(isatty(stdout()) && !(any(search() == "package:gWidgets"))) cat(txt)
        else cat(txt, file = file.path(data@workpath, "mi.html"), append = TRUE)
      }
    }
    if(verbose) {
      if(isatty(stdout()) && !(any(search() == "package:gWidgets"))) cat(" ")
      else cat(" <br/>", file = file.path(data@workpath, "mi.html"), append = TRUE)
    }
    if(.MI_DEBUG) sapply(data@variables, validObject, complete = TRUE)
    on.exit()
    return(data)
  }

## unlike the above methods, these return a (modified) missing_data.frame
setMethod("fit_model", signature(y = "missing", data = "missing_data.frame"), def =
  function(data, s = -1, verbose = FALSE, warn = FALSE, ...) {
    return(.fit_model_mdf(data = data, s = s, verbose = verbose, warn = warn, ...))
  })

setMethod("fit_model", signature(y = "missing", data = "allcategorical_missing_data.frame"), def =
  function(data, s = -1, verbose = FALSE, warn = FALSE, ...) {
    if(verbose) {
      txt <- paste("Iteration:", abs(s))
      if(isatty(stdout()) && !(any(search() == "package:gWidgets"))) cat("\n", txt)
      else cat("<br>", txt, file = file.path(data@workpath, "mi.html"), append = TRUE)
    }
    
    Hstar <- data@Hstar
    
    if(abs(s) == 0) { # starting iteration
      V_h <- c(runif(Hstar - 1), 1)
      c_prod <- cumprod(1 - V_h)
      data@parameters$pi <- V_h * c(1, c_prod[-Hstar])
      data@variables <- lapply(data@variables, FUN = function(y) {
        if(is(y, "irrelevant")) return(y)
        if(y@all_obs) return(y)
        return(mi(y)) # bootstrapping
      })
      data@X <- do.call(cbind, args = lapply(data@variables, FUN = function(y) {
        if(is(y, "irrelevant")) return(NULL)
        else return(y@data)
      }))
      phi <- lapply(1:Hstar, FUN = function(h) {
        lapply(data@variables, FUN = function(y) {
        if(is(y, "irrelevant")) return(NULL)
        return(c(tabulate(y@data, nbins = length(y@levels)) / y@n_total))
      })})
      data@parameters$phi <- phi
      data@parameters$alpha <- 1
      
      cols <- Hstar
      rows <- nrow(data@latents@imputations)
      if(ncol(data@latents@parameters) == 0) {
        temp <- matrix(NA_real_, nrow = rows, ncol = cols)
      }
      data@latents@parameters <- temp
      
      return(data)
    }
      
    # S1: Update latent class membership
    pi  <- data@parameters$pi
    phi <- data@parameters$phi
    probs <- sapply(1:Hstar, FUN = function(h) {
           phi_h <- phi[[h]]
           numerators <- rep(1, nrow(data))
           for(j in 1:ncol(data)) {
             y <- data@variables[[j]]
             if(is(y, "irrelevant")) next
             phi_hj <- phi_h[[y@variable_name]]
             numerators <- numerators * 
                           data@X[,y@variable_name] * phi_hj[data@X[,y@variable_name]]
           }
           numerators <- numerators * pi[h]
           return(numerators)
    })
    z <- apply(probs, 1, FUN = function(prob) {
                 which(rmultinom(1,1,prob) == 1) # rmultinom normalizes internally
              })
    data@latents@data[] <- z
    data@latents@imputations[s,] <- z
    data@latents@parameters[s,] <- pi
    
    # S2: Update V_h
    n_h <- c(tabulate(z, nbins = Hstar))
    V_h <- sapply( 1:(Hstar - 1), FUN = function(h) {
                   a <- 1 + n_h[h]
                   b <- data@parameters$alpha + sum(n_h[-c(1:h)])
                   if(b == 0) return(1)
                   rbeta(1, a, b) 
                  })
    V_h <- c(V_h, 1)
    c_prod <- cumprod(1 - V_h)
    data@parameters$pi <- V_h * c(1, c_prod[-Hstar])
    
    # S3: Update choice probabilities
    phi <- lapply(1:Hstar, FUN = function(h) lapply(data@variables, FUN = function(y) {
      if(is(y, "irrelevant")) return(NULL)
      mark <- z == h
      tab <- tabulate(y@data[mark], nbins = length(y@levels))
      return(.rdirichlet(1, data@priors$a[y@variable_name] + c(tab)))
    }))
    data@parameters$phi <- phi
    
    # S4: Update alpha
    alpha <- rgamma(1, data@priors$a_alpha + Hstar - 1, 
                    data@priors$b_alpha - log(pi[Hstar]))
    data@parameters$alpha <- alpha
    
    # S5: Impute
    data@variables <- lapply(data@variables, FUN = function(y) {
      if(is(y, "irrelevant")) return(y)
      if(y@all_obs) return(y)
      if(verbose) {
        txt <- "."
        if(isatty(stdout()) && !(any(search() == "package:gWidgets"))) cat(txt)
        else cat(txt, file = file.path(data@workpath, "mi.html"), append = TRUE)
      }      
      classes <- z[y@which_drawn]
      uc <- unique(classes)
      Pr <- t(sapply(uc, FUN = function(c) phi[[c]][[y@variable_name]]))
      rownames(Pr) <- uc
      y <- mi(y, Pr[as.character(classes),,drop=FALSE])
    })
    data@X <- do.call(cbind, args = lapply(data@variables, FUN = function(y) {
      if(is(y, "irrelevant")) return(NULL)
      else return(y@data)
    }))

    return(data)
})

.fit_model_Sophie <-
  function(y, data, s = -1, verbose = FALSE, warn = FALSE, ...) {
    classes <- data@latents@data
    uc <- unique(classes)
    Pr <- t(sapply(uc, FUN = function(c) data@parameters$phi[[c]][[y@variable_name]]))
    rownames(Pr) <- uc
#     Pr <- Pr[as.character(classes),,drop=FALSE]
    Pr <- Pr / rowSums(Pr)
    return(list(fitted = Pr))    
  }

setMethod("fit_model", signature(y = "unordered-categorical", 
                                 data = "allcategorical_missing_data.frame"), def =
  function(y, data, s = -1, verbose = FALSE, warn = FALSE, ...) {
    return(.fit_model_Sophie(y, data, s, verbose, warn, ...))
  })

setMethod("fit_model", signature(y = "ordered-categorical", 
                                 data = "allcategorical_missing_data.frame"), def =
  function(y, data, s = -1, verbose = FALSE, warn = FALSE, ...) {
    return(.fit_model_Sophie(y, data, s, verbose, warn, ...))
})

setMethod("fit_model", signature(y = "binary", 
                                 data = "allcategorical_missing_data.frame"), def =
  function(y, data, s = -1, verbose = FALSE, warn = FALSE, ...) {
    return(.fit_model_Sophie(y, data, s, verbose, warn, ...))
})

setMethod("fit_model", signature(y = "missing_data.frame", data = "missing_data.frame"), def =
  function(y, data, s = -1, verbose = FALSE, warn = FALSE, ...) {
    
    if(verbose) {
      txt <- paste("Iteration:", abs(s))
      if(isatty(stdout()) && !(any(search() == "package:gWidgets"))) cat("\n", txt)
      else cat("<br>", txt, file = file.path(data@workpath, "mi.html"), append = TRUE)
    }
    for(jj in sample(1:ncol(y), ncol(y), replace = FALSE)) {
      z <- y@variables[[jj]]
      if(z@all_obs) next
      if(is(z, "irrelevant")) next
      y@variables[[jj]] <- .fit_model_y(z, data, s, verbose, warn, ...)
      if(verbose) {
        txt <- "."
        if(isatty(stdout()) && !(any(search() == "package:gWidgets"))) cat(txt)
        else cat(txt, file = file.path(data@workpath, "mi.html"), append = TRUE)
      }
    }
    if(verbose) {
      if(isatty(stdout()) && !(any(search() == "package:gWidgets"))) cat(" ")
      else cat(" <br/>", file = file.path(data@workpath, "mi.html"), append = TRUE)
    }
    if(.MI_DEBUG) sapply(data@variables, validObject, complete = TRUE)
    return(y)
    
  })

setMethod("fit_model", signature(y = "missing", data = "multilevel_missing_data.frame"), def =
  function(data, s = -1, verbose = FALSE, warn = FALSE, ...) {
    data@mdf_list <- fit_model(data = data@mdf_list, s = s, verbose = verbose, warn = warn, ...)
    if(s == 0) return(data)
    
    ## FIXME: Implement 3+ levels recursively
    
    # update group means
    means <- sapply(data@mdf_list, FUN = function(x) colMeans(x@X[,-1]))
    if(is.list(means)) {
      
    }
    else means <- t(means)
    mark <- 0L ## FIXME
    data@X[,mark] <- means
    
    # impute the group level variables if necessary
    data <- .fit_model_mdf(data = data, s = s, verbose = verbose, warn = warn, ...)
    
    # model the individual level estimates
    for(i in seq_along(ncol(data))) {
      if(is(data@variables[[i]], "irrelevant")) next
      #     if(data@no_missing[i]) next
      mark <- if(s < 0) 1 else s
      fish <- sapply(data@mdf_list, FUN = function(d) d@variables[[i]]@parameters[mark,])
      if(is.list(fish)) {
        ## FIXME: may be a list
      }
      else fish <- t(fish)
      
      for(j in seq_along(ncol(fish))) {
        model <- bayesglm.fit(data@X, y = fish[,j]) # group-level regression
        class(model) <- c("bayesglm", "glm", "lm")
        params <- arm::sim(model, 1)
        beta <- params@coef
        sigma <- params@sigma
        yhats <- rnorm(nrow(X), data@X %*% beta, sd = sigma)
        # change the priors for each element of the mdf_list accordingly
        for(k in seq_along(data@mdf_list)) {
          data@mdf_list[[k]]$mean[colnames(data)[j]] <- yhats[k]
          data@mdf_list[[k]]$sd[colnames(data)[j]] <- sigma
        }
      }
    }
    return(data)
  })

#!/usr/bin/env bash
if [ -n "$1" ]
then name=$1
else name=v10000
fi;

password=Betterme8
cloudml dev create -n $name -p $password \
--priority_class preferred \
-d cr.d.xiaomi.net/cloud-ml/devsave:33103-v100 \
-cm rw \
-rs v100 \
-c 4 -M 8G \
# -g 1 -gt v100 -gm 32g \

# http://docs.api.xiaomi.net/cloud-ml/usage/request_quota_v2.html

#!/usr/bin/env bash
if [ -n "$1" ]
then name=$1
else name=cpu5
fi;

password=Betterme8
cloudml dev create -n $name -p $password \
-d cr.d.xiaomi.net/cloud-ml/devsave:33103-mydev \
--priority_class preferred \
-hka h_browser@XIAOMI.HADOOP -hkt tql -he hdfs://zjyprc-hadoop \
-c 32 -M 64G \
-cm rw

#!/usr/bin/env bash
# @Project      : mi
# @Time         : 2021/1/13 4:51 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : ${DESCRIPTION}

source scripts/config.sh
# /user/s_feeds/wangxuntao/online/DNN/mitv/model/base/checkpoint_batch
#zjyhdfs -rmr /user/s_feeds/wangxuntao/online/DNN/mitv/model/base/checkpoint_batch/*

eval_day=${1:-$YESTERDAY}

end_day=${2:-$LAST2DAY}

days_num=${3:-21}
config_path="conf/mitv_base_online_nn.yaml"

train_day="null"

tf_record_trainSample="hdfs://zjyprc-hadoop/user/s_feeds/dev/user_growth/push/dnn/mitv/tf-record"
cur_path="/user/s_feeds/dev/user_growth/push/dnn/mitv/tf-record"
for ((i=0; i<$days_num; i+=1))
do
    #echo $i
    cur_day=`date -d"$end_day $i days ago" +"%Y%m%d"`
    filepath=$cur_path"/date="$cur_day
    if $ZJYHDFS_TEST_EXIST "$filepath";
    then
        #continue
        sleep 1s
    else
        continue
    fi
    if [ $train_day == "null" ]; then
        train_day=$cur_day
    else
        train_day=$train_day","$cur_day
    fi
done

echo $eval_day $train_day

waitZjyHdfsReady ${tf_record_trainSample}/date=$eval_day 1 &&
echo 'begin training'
python scripts/daily_train.py --config_path $config_path --train_day $train_day --test_day $eval_day
echo " traning finish"

export_model_base="hdfs://zjyprc-hadoop/user/s_feeds/wangxuntao/online/DNN/mitv/export/base"
export_model_path=$export_model_base/date=${eval_day}
echo " export_model_path: "$export_model_path

#online_model="/user/s_feeds/cuizhigang/DNN/model/online"
online_model="/user/s_feeds/songyang7/DNN/model/online"
waitZjyHdfsReady ${export_model_path} 1 &&
zjyhdfs -rm ${online_model}/id_map_*
zjyhdfs -rm ${online_model}/layers
zjyhdfs -rm ${online_model}/embedding_group
zjyhdfs -rm ${online_model}/_SUCCESS
zjyhdfs -cp ${export_model_path}/id_map_* ${online_model}
zjyhdfs -cp ${export_model_path}/layers ${online_model}
zjyhdfs -cp ${export_model_path}/embedding_group ${online_model}
zjyhdfs -cp ${export_model_path}/_SUCCESS ${online_model}

echo "final success"

#!/usr/bin/env bash
# @Project      : tql-cloudml.${PACKAGE_NAME}
# @File         : ${NAME}
# @Time         : 2019-06-05 12:13
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : ${DESCRIPTION}
cloudml tensorboard create -n tql-tb -l fds://browser-algo-nanjing/TensorBoard

#!/usr/bin/env bash
if [ -n "$1" ]
then name=$1
else name=gpu2
fi;

password=xiaomi
cloudml dev create -n $name -p $password \
--priority_class preferred \
-d cr.d.xiaomi.net/cloud-ml/tensorflow-gpu:33103tql2dev  \
-cm rw \
-c 16 -M 32G \
#-g 1

#!/usr/bin/env bash
# @Project      : mi
# @Time         : 2020-03-06 13:40
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : ${DESCRIPTION}

CLASS="com.xiaomi.algo.push.biz.label.LabelTrick"

function spark_calculate(){
spark-submit --cluster 'zjyprc-hadoop-spark2.1' \
--master yarn-cluster \
--conf spark.speculation=false \
--name $CLASS \
--conf spark.yarn.job.owners=yuanjie \
--conf spark.yarn.alert.phone.number=18550288233 \
--queue 'production.miui_group.browser.miui_browser_zjy_1' \
--num-executors 256 \
--executor-cores 2 \
--executor-memory 8g \
--driver-memory 8g \
--class ${CLASS} \
hdfs://zjyprc-hadoop/user/h_browser/algo/yuanjie/jars/MIPush-1.0-SNAPSHOT.jar \
-delta=$1

# var=$(date -d $1day +%Y%m%d)
# curl -d "id=64349&pt=date=$var" http://dp.pt.xiaomi.com/add_partition
}

for i in $(seq -8 1 -2)
do
spark_calculate $i
done

# http://docs.api.xiaomi.net/cloud-ml/usage/request_quota_v2.html

#!/usr/bin/env bash
if [ -n "$1" ]
then name=$1
else name=cpu666
fi;

password=Betterme8

#cloudml ceph create -n $name -c 100G
#sleep 10

cloudml dev create -n $name -p $password \
-d cr.d.xiaomi.net/cloud-ml/devsave:33103-v100 \
--priority_class preferred \
-c 4 -M 8G \
-cm rw \
-rs $name

#!/usr/bin/env bash
# @Project      : tql-cloudml
# @Time         : 2019-06-21 00:09
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : ${DESCRIPTION}

model="$1"
if [ -n "$2" ];
then n="$2"
else n=10
fi;

cloudml jobs logs $model -n $n

#!/usr/bin/env bash
if [ -n "$1" ]
then name=$1
else name=gpu2
fi;

# 镜像列表：http://docs.api.xiaomi.net/cloud-ml/usage/image_list.html
password=xiaomi
cloudml dev create -n $name -p $password \
--priority_class guaranteed \
-d cr.d.xiaomi.net/cloud-ml/horovod:0.19.3-tf2.1.0-torch-mxnet1.6.0-xm1.0.0-py3 \
-cm rw \
-c 32 -M 32G \
-g 1 # [u\'v100-32g\', u\'t4-16g\', u\'p4-8g\', u\'v100-16g\']

#!/usr/bin/env bash
# @Project      : tql-cloudml
# @Time         : 2019-07-01 15:56
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     :p PyCharm
# @Description  : ${DESCRIPTION}

#a=$(python p.py)
#
#echo $a

dt=$(date "+%Y%m%d")
echo "time: $dt"

# http://docs.api.xiaomi.net/cloud-ml/usage/request_quota_v2.html

#!/usr/bin/env bash
if [ -n "$1" ]
then name=$1
else name=tql2dev
fi;

password=xiaomi
cloudml dev create -n $name -p $password \
-d cr.d.xiaomi.net/cloud-ml/tensorflow-gpu:33103tql2dev \
--priority_class guaranteed \
-c 32 -M 32G \
-g 1 \
-cv aaa \
-cm rw


#https://cloud.d.xiaomi.net/product/docs/cloudml/devenv/02_use_dev_cli

# http://docs.api.xiaomi.net/cloud-ml/usage/request_quota_v2.html

#!/usr/bin/env bash
if [ -n "$1" ]
then name=$1
else name=master
fi;

#cloudml ceph create -n $name -c 100G
#sleep 10

password=Betterme8
cloudml dev create -n $name -p $password \
-d cr.d.xiaomi.net/cloud-ml/devsave:33103-cpu666 \
--priority_class guaranteed \
-c 32 -M 64G \
-cm rw \
-rs $name \
-g 1 # [u\'v100-32g\', u\'t4-16g\', u\'p4-8g\', u\'v100-16g\']

#!/usr/bin/env bash
if [ -n "$1" ]
then name=$1
else name=v100
fi;

password=Betterme8
cloudml dev create -n $name -p $password \
--priority_class guaranteed \
-d cr.d.xiaomi.net/cloud-ml/tensorflow-gpu:2.2.0-xm1.0.0-py3 \
-cm rw \
-rs v100 \
-c 31 -M 32G \
-g 1 -gt v100 -gm 32g \

#!/usr/bin/env bash
if [ -n "$1" ];
then model=$1
else model="model_horovod"
fi;

if cloudml jobs delete $model | grep "Successfully";
then echo "删除 ..." && sleep 18s;
fi;
cloudml jobs submit -f ./model.yaml -n $model;
rm -rf ./build/ ./*.egg-info/ ./dist/;
sleep 8s && cloudml jobs metrics $model;

#!/usr/bin/env bash
DATA=`date +'%Y%m%d-%H-%M-%S'`

if [ -n "$2" ]
then taget_name=$2
else taget_name="tql$DATA"
fi;

project_name=$1 # 主程序命名项目名
#if cloudml jobs delete $project_name | grep "Successfully";
#then echo "删除 ..." && sleep 18s;
#fi;
#cr.d.xiaomi.net/cloud-ml/tensorflow-gpu:2.1.0-xm1.0.0-py3 \
cloudml jobs submit \
-n $taget_name \
-m $project_name \
-u ./ \
-d cr.d.xiaomi.net/cloud-ml/devsave:33103-cpu666 \
-hka h_browser@XIAOMI.HADOOP -hkt tql -he hdfs://zjyprc-hadoop \
-pc "pip install -U --no-cache-dir -i https://pypi.python.org/pypi meutils" \
--priority_class "guaranteed" \
-c 16 -M 22G \
-g 1 #-gt v100 -gm 32g


# -a "--data-dir=/fds/cifar-10-data --job-dir=/tmp/cifar10 --num-gpus=1 --train-steps=1000"

rm -rf ./build/ ./*.egg-info/ ./dist/;
mv ./*.yaml ./checkpoints

# cloudml jobs list

#!/usr/bin/env bash
#cloudml quota_v2 update -p guaranteed -c 32 -M 64G -e yuanjie@xiaomi.com \
#-g 1 \
#-gt v100

#cloudml quota_v2 update -p best-effort \
#-gc 16 -c 256 -M 512 \
#-e yuanjie@xiaomi.com

#cloudml quota_v2 update -p guaranteed \
#-g 1 -c 128 -M 128 \
#-gt v100-16g -gt p4-8g \
#-gc 1 -gc 2 \
#-e yuanjie@xiaomi.com


#cloudml dev save-v2 -f master mydev

#!/usr/bin/env bash
if [ -n "$1" ]
then name=$1
else name=master
fi;

password=Betterme8
cloudml dev create -n $name -p $password \
--priority_class guaranteed \
-d cr.d.xiaomi.net/cloud-ml/devsave:33103-dev_ \
-hka h_browser@XIAOMI.HADOOP -hkt tql -he hdfs://zjyprc-hadoop \
-cm rw \
-c 32 -M 32G \
-g 1 # [u\'v100-32g\', u\'t4-16g\', u\'p4-8g\', u\'v100-16g\']

#!/usr/bin/env bash
# @Project      : tql-cloudml
# @Time         : 2019-06-10 19:19
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : ${DESCRIPTION}

docker rmi # 删除镜像
docker image prune # 清空镜像

docker container ls -a
docker rm # 删除容器
docker container prune -f # 清空容器

docker pull cr.d.xiaomi.net/cloud-ml/tensorflow-gpu:1.13.1-xm1.0.0-zjyprc-hadoop-py3

docker run -it --rm --name container 2f08cdcac492 bash

docker commit  -a 'yuanjie' -m 'ml tools' --change='CMD /prepare_dev.py && /run_jupyter.sh' tql-ml cr.d.xiaomi.net/yuanjie/tql-ml:v1

# milvus-admin
# docker image inspect milvusdb/milvus-admin:latest | grep -i version
docker run -p 3000:80 milvusdb/milvus-em
docker container list

ContainerID=8ca24d3ff686
ImageName=milvus-admin:latest
docker commit  -a 'yuanjie' -m 'ann' $ContainerID cr.d.xiaomi.net/yuanjie/$ImageName
docker push cr.d.xiaomi.net/yuanjie/$ImageName

# app
ContainerID=eef0cba1b857
ImageName=app:latest
docker commit  -a 'yuanjie' -m 'ann' $ContainerID cr.d.xiaomi.net/yuanjie/$ImageName
docker push cr.d.xiaomi.net/yuanjie/$ImageName

cr.d.xiaomi.net/yuanjie/app:latest
# milvus
docker run -it --rm --name ann milvusdb/milvus:cpu-latest bash
docker container list

ContainerID=8313f4e7e2b8
ImageName=milvus:cpu-latest
docker commit  -a 'yuanjie' -m 'ann' $ContainerID cr.d.xiaomi.net/yuanjie/$ImageName
docker push cr.d.xiaomi.net/yuanjie/$ImageName

ContainerID=3b0fea9464f6
ImageName=neo4j:0.0.1
docker commit  -a 'yuanjie' -m 'neo4j' $ContainerID cr.d.xiaomi.net/yuanjie/$ImageName
docker push cr.d.xiaomi.net/yuanjie/$ImageName

---
title: "An Example of mi Usage"
author: "Ben Goodrich and Jonathan Kropko, for this version, based on earlier 
versions written by Yu-Sung Su, Masanao Yajima, Maria Grazia Pittau, Jennifer Hill, 
and Andrew Gelman"
date: "06/16/2014"
output: pdf_document
---
<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{An Example of mi Usage}
-->

There are several steps in an analysis of missing data. Initially, users must 
get their data into R. There are several ways to do so, including the 
`read.table`, `read.csv`, `read.fwf` functions plus several functions in the 
__foreign__ package. All of these functions will generate a `data.frame`, which
is a bit like a spreadsheet of data.
https://cran.r-project.org/doc/manuals/R-data.html for more information.

```{r step0}
options(width = 65)
suppressMessages(library(mi))
data(nlsyV, package = "mi")
```

From there, the first step is to convert the `data.frame` to a `missing_data.frame`, 
which is an enhanced version of a `data.frame` that includes metadata about the 
variables that is essential in a missing data context. 
```{r step1}
mdf <- missing_data.frame(nlsyV)
```
The `missing_data.frame` constructor function creates a `missing_data.frame` 
called `mdf`, which in turn contains seven `missing_variable`s, one for each 
column of the `nlsyV` dataset.

The most important aspect of a `missing_variable` is its class, such as
`continuous`, `binary`, and `count` among many others (see the table in the Slots 
section of the help page for `missing_variable-class`. The `missing_data.frame`
constructor function will try to guess the appropriate class for each 
`missing_variable`, but rarely will it correspond perfectly to the user's intent. 
Thus, it is very important to call the `show` method on a `missing_data.frame` to 
see the initial guesses 
```{r step1.5}
show(mdf) # momrace is guessed to be ordered
```
and to modify them, if necessary, using the `change` function, which can be used
to change many things about a`missing_variable`, so see its help page for more 
details. In the example below, we change the class of the _momrace_ (race of the 
mother) variable from the initial guess of `ordered-categorical` to a more appropriate 
`unordered-categorical` and change the income `nonnegative-continuous`.
```{r, step2}
mdf <- change(mdf, y = c("income", "momrace"), what = "type", 
                     to = c("non", "un"))
show(mdf)
```
Once all of the `missing_variable`s are set appropriately, it is useful to get a 
sense of the raw data, which can be accomplished by looking at the `summary`,
`image`, and / or `hist` of a `missing_data.frame`
```{r, step3}
summary(mdf)
image(mdf)
hist(mdf)
```
Next we use the `mi` function to do the actual imputation, which has several 
extra arguments that, for example, govern how many independent chains to utilize, 
how many iterations to conduct, and the maximum amount of time the user is willing 
to wait for all the iterations of all the chains to finish. The imputation step 
can be quite time consuming, particularly if there are many `missing_variable`s 
and if many of them are categorical. One important way in which the computation 
time can be reduced is by imputing in parallel, which is highly recommended and 
is implemented in the mi function by default on non-Windows machines. If users encounter problems 
running `mi` with parallel processing, the problems are likely due  to the machine 
exceeding available RAM.  Sequential processing can be used instead for `mi`
by using the `parallel=FALSE` option.

```{r, step4}
rm(nlsyV)       # good to remove large unnecessary objects to save RAM
options(mc.cores = 2)
imputations <- mi(mdf, n.iter = 30, n.chains = 4, max.minutes = 20)
show(imputations)
```
The next step is very important and essentially verifies whether enough iterations 
were conducted. We want the mean of each completed variable to be roughly the 
same for each of the 4 chains. 
```{r, step5A}
round(mipply(imputations, mean, to.matrix = TRUE), 3)
Rhats(imputations)
```
If so --- and when it does in the example depends on the pseudo-random number seed 
--- we can procede to diagnosing other problems. For the sake of example, we 
continue our 4 chains for another 5 iterations by calling 
```{r, step5B}
imputations <- mi(imputations, n.iter = 5)
```
to illustrate that this process can be continued until convergence is reached.

Next, the `plot` of an object produced by `mi` displays, for all 
`missing_variable`s (or some subset thereof), a histogram of the observed, imputed, 
and completed data, a comparison of the completed data to the fitted values implied 
by the model for the completed data, and a plot of the associated binned residuals. 
There will be one set of plots on a page for the first three chains, so that the 
user can get some sense of the sampling variability of the imputations. The `hist` 
function yields the same histograms as `plot`, but groups the histograms for all 
variables (within a chain) on the same plot. The `image`function gives a sense of 
the missingness patterns in the data.
```{r, step6}
plot(imputations)
plot(imputations, y = c("ppvtr.36", "momrace"))
hist(imputations)
image(imputations)
summary(imputations)
```
Finally, we pool over `m = 5` imputed datasets -- pulled from across the 
4 chains -- in order to estimate a descriptive linear regression of test scores 
(_ppvtr.36_) at 36 months on a variety of demographic variables pertaining to the 
mother of the child.
```{r, step7}
analysis <- pool(ppvtr.36 ~ first + b.marr + income + momage + momed + momrace, 
                 data = imputations, m = 5)
display(analysis)
```
The rest is optional and only necessary if you want to perform some operation that 
is not supported by the __mi__ package, perhaps outside of R. Here we create a 
list of `data.frame`s, which can be saved to the hard disk and / or 
exported in a variety of formats with the __foreign__ package.  Imputed data can 
be exported to Stata by using the `mi2stata` function instead of `complete`.
```{r, step8}
dfs <- complete(imputations, m = 2)
```

\name{00mi-package}
\alias{mi-package}
\docType{package}
\title{Iterative Multiple Imputation from Conditional Distributions
}
\description{
The mi package performs multiple imputation for data with missing values.  The algorithm iteratively draws imputed values from the conditional distribution for each variable given the observed and imputed values of the other variables in the data. The process approximates a Bayesian framework; multiple chains are run and convergence is assessed after a pre-specified number of iterations within each chain.  The package allows customization of the conditional model and the treatment of missing values for each variable.  In addition, the package provides graphics to visualize missing data patterns, to diagnose the models used to generate the imputations, and to assess convergence. Functions are included to run statistical models post-imputation with the appropriate degree of sampling uncertainty.
}
\details{
\tabular{ll}{
Package: \tab mi\cr
Type: \tab Package\cr
Version: \tab 1.0\cr
Date: \tab \Sexpr[eval=TRUE,results=rd,stage=build]{date()} \cr
License: \tab GPL (>= 2) \cr
LazyLoad: \tab yes\cr
}
See the vignette for an example of typical usage.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions 
written by Yu-Sung Su, Masanao Yajima,Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_data.frame}}, \code{\link{change}}, \code{\link{mi}}, \code{\link{Rhats}}, 
\code{\link{pool}}, \code{\link{complete}}
}
\keyword{package}
\keyword{AimedAtusers}

\name{get_parameters}
\docType{methods}
\alias{get_parameters}
\alias{get_parameters-methods}
\title{An Extractor Function for Model Parameters}
\description{
This function is not intended to be called directly by users.
During the multiple imputation process, the \code{\link{mi}}
function estimates models and stores the estimated parameters in the
\code{parameters} slot of an object that inherits from the 
\code{\link{missing_variable-class}}. The \code{get_parameter} 
function simply extracts these parameters for storeage, which are
usually the estimated coefficients but may also include ancillary
parameters.
}
\usage{
get_parameters(object, ...)
}
\arguments{
  \item{object}{Usually an estimated model, such as that produced by \code{\link{glm}}
}
  \item{\dots}{Additional arguments, currently not used
}
}
\details{
There is method for the object produced by \code{\link{polr}}, which
also returns the estimated cutpoints in a proportional odds model. However,
the default method simply calls \code{\link{coef}} and returns the result.
If users implement their own models, it may be necessary to write a short
\code{get_parameters} method.
}
\value{
A numeric vector of estimated parameters
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{fit_model}}
}
\examples{
showMethods("get_parameters")
}
\keyword{methods}

\name{fit_model}
\docType{methods}
\alias{fit_model}
\alias{fit_model-methods}
\title{Wrappers To Fit a Model}
\description{
The methods are called by the \code{\link{mi}} function to model a given
\code{\link{missing_variable}} as a function of all the other 
\code{\link{missing_variable}}s and also their missingness pattern.
By overwriting these methods, users can change the way a 
\code{\link{missing_variable}} is modeled for the purposes of imputing
its missing values. See also the table in \code{\link{missing_variable}}.
}
\usage{
fit_model(y, data, ...)
}
\arguments{
  \item{y}{An object that inherits from \code{\link{missing_variable-class}} or
     missing
}
  \item{data}{A \code{\link{missing_data.frame}}

}
  \item{\dots}{Additional arguments, not currently utilized
}
}
\details{
In \code{\link{mi}}, each \code{\link{missing_variable}} is modeled as a function of 
all the other \code{\link{missing_variable}}s plus their missingness pattern. The
\code{fit_model} methods are typically short wrappers around a statistical model fitting
function and return the estimated model. The model is then passed to one of the 
\code{\link{mi-methods}} to impute the missing values of that \code{\link{missing_variable}}.

Users can easily overwrite these methods to estimate a different model, such as wrapping
\code{\link{glm}} instead of \code{\link[arm]{bayesglm}}. See the source code for examples,
but the basic outline is to first extract the \code{X} slot of the 
\code{\link{missing_data.frame}}, then drop some of its columns using the \code{index} slot
of the \code{\link{missing_data.frame}}, next pass the result along with the \code{data} slot
of \code{y} to a statistical fitting function, and finally returned the appropriately classed
result (along with the subset of \code{X} used in the model).

Many of the optional arguments to a statistical fitting function can be specified using the
slots of \code{y} (e.g. its \code{family} slot) or the slots of \bold{data} (e.g. its 
\code{weights} slot).

The exception is the method where \code{y} is missing, which is used internally by 
\code{\link{mi}}, and should \emph{not} be overwritten unless great care is taken to understand
its role.
}
\value{
If \code{y} is missing, then the modified \code{\link{missing_data.frame}} passed to
\code{data} is returned. Otherwise, the estimated model is returned as a classed
list object.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_variable}}, \code{\link{mi}}, \code{\link{get_parameters}}
}
\examples{
getMethod("fit_model", signature(y = "binary", data = "missing_data.frame"))
setMethod("fit_model", signature(y = "binary", data = "missing_data.frame"), def =
function(y, data, ...) {
  to_drop <- data@index[[y@variable_name]]
  X <- data@X[, -to_drop]
  start <- NULL
  # using glm.fit() instead of bayesglm.fit()
  out <- glm.fit(X, y@data, weights = data@weights[[y@variable_name]], start = start, 
                 family = y@family, Warning = FALSE, ...)
  out$x <- X
  class(out) <- c("glm", "lm") # not "bayesglm" class anymore
  return(out)
})
\dontrun{
if(!exists("imputations", env = .GlobalEnv)) {
  imputations <- mi:::imputations # cached from example("mi-package")
}
imputations <- mi(imputations) # will use new fit_model() method for binary variables
}
}
\keyword{regression}
\keyword{DirectedTowardDevelopeRs}

\name{irrelevant}
\Rdversion{1.1}
\docType{class}
\alias{irrelevant}
\alias{irrelevant-class}
\alias{fixed-class}
\alias{group-class}
\title{Class "irrelevant" and Inherited Classes}
\description{
The irrelevant class inherits from the \code{\link{missing_variable-class}} and is used to 
designate variables that are excluded from the models used to impute the missing values of 
\dQuote{relevant} variables. For example, if a survey has an \dQuote{id} variable that 
simply distinguishes observations, the user should designate it as irrelevant, although it 
will automatically be classified so if its name is either \dQuote{id} or starts with punctuation 
(including underscores). The fixed class inherits from the irrelevant class and is used 
for variables that are constant (within a sample). A variable that is instantiated from the
fixed class cannot have any missing values. The group class inherits from the fixed 
class and is used like a \code{\link{factor}} to spit samples in multilevel modeling; see 
\code{\link{multilevel_missing_data.frame-class}}. None of these classes have an additional
slots. Aside from these facts, the rest of the documentation here is primarily directed toward developeRs.
}
\section{Objects from the Classes}{The \code{\link{missing_variable}} generic function can be used to 
  instantiate an object that inherits from the irrelevant class by specifying 
  \code{type = "irrelevant"}, \code{type = "fixed"}, or \code{type = "group"}.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_variable-class}}
}
\examples{
# STEP 0: GET DATA
data(nlsyV, package = "mi")

# STEP 0.5 CREATE A missing_variable (you never need to actually do this)
first <- missing_variable(as.factor(nlsyV$first), type = "group")
show(first)
}
\keyword{classes}
\keyword{DirectedTowardDevelopeRs}

\name{nlsyV}
\alias{nlsyV}
\docType{data}
\title{
National Longitudinal Survey of Youth Extract
}
\description{
This dataset pertains to children and their families in the United States and 
is intended to illustrate missing data issues. Note that although the original
data are longitudinal, this extract is not.
}
\usage{data(nlsyV)}
\format{
  A data frame with 400 randomly subsampled observations on the following 7 variables.
  \describe{
    \item{\code{ppvtr.36}}{a numeric vector with data on the Peabody Picture Vocabulary Test (Revised) 
      administered at 36 months}
    \item{\code{first}}{indicator for whether child was first-born}
    \item{\code{b.marr}}{indicator for whether mother was married when child was born}
    \item{\code{income}}{a numeric vector with data on family income in year after the child was born}
    \item{\code{momage}}{a numeric vector with data on the age of the mother when the child was born}
    \item{\code{momed}}{educational status of mother when child was born (1 = less than high school, 
      2 = high school graduate, 3 = some college, 4 = college graduate)}
    \item{\code{momrace}}{race of mother (1 = black, 2 = Hispanic, 3 = white)}
  }
  Note that \bold{momed} would typically be an ordered \code{\link{factor}} while \bold{momrace}
  would typically be an unorderd \code{\link{factor}} but both are \code{\link{numeric}} in this
  \code{\link{data.frame}} in order to illustrate the mechanism to \code{\link{change}} the
  type of a \code{\link{missing_variable}}
}
\source{
National Longitudinal Survey of Youth, 1997, \url{https://www.nlsinfo.org/content/cohorts/nlsy97}
}
\examples{
data(nlsyV)
summary(nlsyV)
}
\keyword{datasets}

\name{hist}
\Rdversion{1.1}
\docType{methods}
\alias{hist}
\alias{hist-methods}
\title{Histograms of Multiply Imputed Data
}
\description{
This function creates a histogram from an object of
\code{\link{missing_data.frame-class}} or 
\code{\link{mi-class}}
}
\usage{
hist(x, ...)
}
\arguments{
  \item{x}{an object of \code{\link{missing_data.frame-class}} or 
   \code{\link{mi-class}}
}
  \item{\dots}{further arguments passed to \code{\link{plot.histogram}}
}
}
\details{
When called on an object of \code{\link{missing_data.frame-class}}, the
histograms of the observed data are generated, one for each \code{\link{missing_variable}}
but grouped on a single page. When called on an object of \code{\link{mi-class}}, 
the histograms of the observed, imputed, and completed data are generated, one for each
\code{\link{missing_variable}}, grouped on a single page for each chain.
}
\value{
An invisible \code{NULL} is returned with a side-effect of creating a plot
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link[graphics]{hist}}
}
\examples{
if(!exists("imputations", env = .GlobalEnv)) {
  imputations <- mi:::imputations # cached from example("mi-package")
}
hist(imputations)
}
\keyword{hplot}
\keyword{AimedAtUseRs}
\keyword{methods}

\name{rdata.frame}
\alias{rdata.frame}
\title{Generate a random data.frame with tunable characteristics}
\description{
This function generates a random \code{\link{data.frame}} with a
missingness mechanism that is used to impose a missingness pattern. The primary
purpose of this function is for use in simulations
}
\usage{
rdata.frame(N = 1000, 
            restrictions = c("none", "MARish", "triangular", "stratified", "MCAR"),
            last_CPC = NA_real_, strong = FALSE, pr_miss = .25, Sigma = NULL, 
            alpha = NULL, experiment = FALSE, 
            treatment_cor = c(rep(0, n_full - 1), rep(NA, 2 * n_partial)),
            n_full = 1, n_partial = 1, n_cat = NULL,
            eta = 1, df = Inf, types = "continuous", estimate_CPCs = TRUE)
}
\arguments{
  \item{N}{integer indicating the number of observations}
  \item{restrictions}{character string indicating what restrictions to impose on
     the the missing data mechansim, see the Details section}
  \item{last_CPC}{a numeric scalar between \eqn{-1} and \eqn{1} exclusive or 
    \code{NA_real_} (the default). If not \code{NA_real_}, then this value will
    be used to construct the correlation matrix from which the data are drawn.
    This option is useful if restrictions is \code{"triangular"} or \code{"stratified"},
    in which case the degree to which \code{last_CPC} is not zero causes a violation of 
    the Missing-At-Random assumption that is confined to the last of the partially 
    observed variables}
  \item{strong}{Integer among 0, 1, and 2 indicating how strong to
    make the instruments with multiple partially observed variables,
    in which case the missingness indicators for each partially observed variable
    can be used as instruments when predicting missingness on other partially
    observed variables. Only applies when \code{restrictions = "triangular"}}
  \item{pr_miss}{numeric scalar on the (0,1) interval or vector
    of length \code{n_partial} indicating the proportion of observations
    that are missing on partially observed variables}
  \item{Sigma}{Either \code{\link{NULL}} (the default) or a correlation matrix
    of appropriate order for the variables (including the missingness
    indicators). By default, such a matrix is generated at random.}
  \item{alpha}{Either \code{\link{NULL}}, \code{\link{NA}}, or a numeric
    vector of appropriate length that governs the skew of a multivariate
    skewed normal distribution; see \code{\link[sn]{rmsn}}. The appropriate
    length is \code{n_full - 1 + 2 * n_partial} iff none of the variable types
    is nominal. If some of the variable types are nominal, then the appropriate
    length is \code{n_full - 1 + 2 * n_partial + sum(n_cat) - length(n_cat)}.
    If \code{\link{NULL}}, \code{alpha} is taken to be zero, in which case the
    data-generating process has no skew. If \code{\link{NA}}, \code{alpha}
    is drawn from \code{\link{rt}} with \code{df} degrees of freedom}
  \item{experiment}{logical indicating whether to simulate a randomized
    experiment}
  \item{treatment_cor}{Numeric vector of appropriate length indicating the
    correlations between the treatment variable and the other variables, which
    is only relevant if \code{experiment = TRUE}. The appropriate length is
    \code{n_full - 1 + 2 * n_partial} iff none of the variable types is nominal.
    If some of the variable types are nominal, then the appropriate length is
    \code{n_full - 1 + 2 * n_partial + sum(n_cat) - length(n_cat)}. If
    treatment_cor is of length one and is zero, then it will be recylced to
    the appropriate length. The treatment variable should be uncorrelated with
    intended covariates and uncorrelated with missingness on intended
    covariates. If any elements of treatment_cor are \code{\link{NA}}, then
    those elements will be replaced with random draws. Note that the order of
    the random variables is: all fully observed variables,all partially observed
    but not nominal variables, all partially observed nominal variables, all
    missingness indicators for partially observed variables.}
  \item{n_full}{integer indicating the number of fully observed variables}
  \item{n_partial}{integer indicating the number of partially observed
    variables}
  \item{n_cat}{Either \code{\link{NULL}} or an integer vector (possibly of
    length one) indicating the number of categories in each partially observed
    nominal or ordinal variable; see the Details section}
  \item{eta}{Positive numeric scalar which serves as a hyperparameter in the
    data-generating process. The default value of 1 implies that the correlation
    matrix among the variables is jointly uniformally distributed, using
    essentially the same logic as in the \pkg{clusterGeneration} package}
  \item{df}{positive numeric scalar indicating the degress of freedom for the
    (possibly skewed) multivariate t distribution, which defaults to
    \code{\link{Inf}} implying a (possibly skewed) multivariate normal 
    distribution}
  \item{types}{a character vector (possibly of length one, in which case it
    is recycled) indicating the type for each fully observed and partially
    observed variable, which currently can be among \code{"continuous"},
    \code{"count"}, \code{"binary"}, \code{"treatment"} (which is binary),
    \code{"ordinal"}, \code{"nominal"}, \code{"proportion"}, \code{"positive"}.
    See the Details section. Unique abbreviations are acceptable.}
  \item{estimate_CPCs}{A logical indicating whether the canonical partial correlations
    between the partially observed variables and the latent missingnesses should
    be estimated. The default is \code{TRUE} but considerable wall time can be saved
    by switching it to \code{FALSE} when there are many partially observed variables.}
}
\details{
By default, the correlation matrix among the variables and missingness indicators
is intended to be close to uniform, although it is often not possible to achieve
exactly. If \code{restrictions = "none"}, the data will be Not Missing At Random
(NMAR). If \code{restrictions = "MARish"}, the departure from Missing At Random
(MAR) will be minimized via a call to \code{\link{optim}}, but generally will
not fully achieve MAR. If \code{restrictions = "triangular"}, the MAR assumption
will hold but the missingness of each partially observed variable will only 
depend on the fully observed variables and the other latent missingness indicators.
If \code{restrictions = "stratified"}, the MAR assumption will hold but the
missingness of each partially observed variable will only depend on the fully
observed variables. If \code{restrictions = "MCAR"}, the Missing Completely At
Random (MCAR) assumption holds, which is much more restrictive than MAR.

There are some rules to follow, particularly when specifying \code{types}.
First, if \code{experiment = TRUE}, there must be exactly one treatment
variable (taken to be binary) and it must come first to ensure that the
elements of \code{treatment_cor} are handled properly. Second, if there are any
partially observed nominal variables, they must come last; this is to ensure
that they are conditionally uncorrelated with each other. Third, fully observed
nominal variables are not supported, but they can be made into ordinal variables
and then converted to nominal after the fact. Fourth, including both ordinal and
nominal partially observed variables is not supported yet, Finally, if any
variable is specified as a count, it will not be exactly consistent with the
data-generating process. Essentially, a count variable is constructed from a
continuous variable by evaluating \code{\link{pt}} on it and passing that to
\code{\link{qpois}} with an intensity parameter of 5. The other non-continuous
variables are constructed via some transformation or discretization of a continuous
variable.

If some partially observed variables are either ordinal or nominal (but not both), 
then the \code{n_cat} argument governs how many categories there are. If \code{n_cat}
is \code{NULL}, then the number of categories defaults to three. If
\code{n_cat} has length one, then that number of categories will be used for all
categorical variables but must be greater than two. Otherwise, the length of 
\code{n_cat} must match the number of partially observed categorical variables and
the number of categories for the \eqn{i}th such variable will be the \eqn{i}th element
of \code{n_cat}.
}
\value{
A list with the following elements:
\item{true}{ a \code{\link{data.frame}} containing no \code{\link{NA}} values}
\item{obs}{ a \code{\link{data.frame}} derived from the previous with some
    \code{\link{NA}} values that represents a dataset that could be observed}
\item{empirical_CPCs}{ a numeric vector of empirical Canonical Partial
     Correlations, which should differ only randomly from zero iff
     \code{MAR = TRUE} and the data-generating process is multivariate normal}
\item{L}{ a Cholesky factor of the correlation matrix used to generate the
      true data}

In addition, if \code{alpha} is not \code{\link{NULL}}, then the following
elements are also included:
\item{alpha}{ the \code{alpha} vector utilized}
\item{sn_skewness}{ the skewness of the multivariate skewed normal distribution
    in the population; note that this value is only an approximation of the
    skewness when \code{df < Inf}}
\item{sn_kurtosis}{ the kurtosis of the multivariate skewed normal distribution
    in the population; note that this value is only an approximation of the
    kurtosis when \code{df < Inf}}
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung
Su, Masanao Yajima, Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{data.frame}}, \code{\link{missing_data.frame}}
}
\examples{
rdf <- rdata.frame(n_partial = 2, df = 5, alpha = rnorm(5))
print(rdf$empirical_CPCs) # not zero
rdf <- rdata.frame(n_partial = 2, restrictions = "triangular", alpha = NA)
print(rdf$empirical_CPCs) # only randomly different from zero
print(rdf$L == 0) # some are exactly zero by construction
mdf <- missing_data.frame(rdf$obs)
show(mdf)
hist(mdf)
image(mdf)
# a randomized experiment
rdf <- rdata.frame(n_full = 2, n_partial = 2, 
                   restrictions = "triangular", experiment = TRUE,
                   types = c("t", "ord", "con", "pos"),
                   treatment_cor = c(0, 0, NA, 0, NA))
Sigma <- tcrossprod(rdf$L)
rownames(Sigma) <- colnames(Sigma) <- c("treatment", "X_2", "y_1", "Y_2",
                                        "missing_y_1", "missing_Y_2")
print(round(Sigma, 3))
}
\keyword{utilities}

\name{experiment_missing_data.frame}
\Rdversion{1.1}
\docType{class}
\alias{experiment_missing_data.frame}
\alias{experiment_missing_data.frame-class}
\title{Class "experiment_missing_data.frame"}
\description{
This class inherits from the \code{\link{missing_data.frame-class}} but is customized for the situation
where the sample is a randomized experiment.
}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("experiment_missing_data.frame", ...)}.
However, its users almost always will pass a \code{\link{data.frame}} to the 
\code{\link{missing_data.frame}} function and specify the \code{subclass} and \code{concept} arguments.
}
\section{Slots}{
  The experiment_missing_data.frame class inherits from the \code{\link{missing_data.frame-class}} and
  has two additional slots
  \describe{
    \item{concept}{Object of class \code{\link{factor}} whose length is equal to the number of variables
      and whose levels are \code{"treatment"}, \code{"covariate"} and \code{"outcome"}}
    \item{case}{Object of class \code{\link{character}} of length one, indicating whether the missingness
      is in the outcomes only, in the covariates only, or in both the outcomes and covariates. This slot
      is filled automatically by the \code{\link{initialize}} method}
  }
}
\details{
  The \code{\link{fit_model-methods}} for the experiment_missing_data.frame class take into account the
  special nature of a randomized experiment. At the moment, the treatment variable must be binary and
  fully observed.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_data.frame}}
}
\examples{
rdf <- rdata.frame(n_full = 2, n_partial = 2, 
                   restrictions = "stratified", experiment = TRUE,
                   types = c("t", "ord", "con", "pos"),
                   treatment_cor = c(0, 0, NA, 0, NA))
Sigma <- tcrossprod(rdf$L)
rownames(Sigma) <- colnames(Sigma) <- c("treatment", "X_2", "y_1", "Y_2",
                                        "missing_y_1", "missing_Y_2")
print(round(Sigma, 3))

concept <- as.factor(c("treatment", "covariate", "covariate", "outcome"))
mdf <- missing_data.frame(rdf$obs, subclass = "experiment", concept = concept)
}
\keyword{classes}
\keyword{manip}
\keyword{AimedAtUseRs}

\name{multinomial}
\alias{multinomial}
\title{The multinomial family}
\description{
This function is a returns a \code{\link{family}} and is a generalization of \code{\link{binomial}}.
users would only need to call it when calling \code{\link{change}} with 
\code{what = "family", to = multinomial(link = 'logit')}
}
\usage{
multinomial(link = "logit")
}
\arguments{
  \item{link}{character string among those supported by \code{\link{binomial}}
}
}
\details{
This function is mostly cosmetic. The \code{family} slot for an object of 
\code{\link{unordered-categorical-class}} must be \code{multinomial(link = 'logit')}. For
an object of \code{\link{ordered-categorical-class}} but not its subclasses, the \code{family}
slot must be \code{multinomial()} but the link function can differ from its default (\code{"logit"})
}
\value{
A \code{\link{family}} object
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{family}}, \code{\link{binomial}}
}
\examples{
multinomial()
}
\keyword{utilities}

\name{count-class}
\Rdversion{1.1}
\docType{class}
\alias{count-class}
\title{Class "count"}
\description{
The count class inherits from the \code{\link{missing_variable-class}} and is intended for count data. 
Aside from these facts, the rest of the documentation here is primarily directed toward developers.
}
\section{Objects from the Classes}{Objects can be created that are of count class via the 
  \code{\link{missing_variable}} generic function by specifying \code{type = "count"} 
}
\section{Slots}{
  The count class inherits from the missing_variable class and its \code{raw_data} slot must consist of nonnegative
  integers. Its default family is \code{\link{quasipoisson}} and its default \code{\link{fit_model}} method is 
  a wrapper for \code{\link[arm]{bayesglm}}. The other possibility for the family is \code{\link{poisson}} but is
  not recommended due to its overly-restrictive nature.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_variable}}, \code{\link{continuous-class}}, \code{\link{positive-continuous-class}},
\code{\link{proportion-class}}
}
\examples{
# STEP 0: GET DATA
data(CHAIN, package = "mi")

# STEP 0.5 CREATE A missing_variable (you never need to actually do this)
age <- missing_variable(as.integer(CHAIN$age), type = "count")
show(age)
}
\keyword{classes}
\keyword{DirectedTowardDevelopeRs}

\name{mipply}
\alias{mipply}
\title{Apply a Function to a Object of Class mi}
\description{
This function is a wrapper around \code{\link{sapply}} that is invoked on the
\code{data} slot of an object of \code{\link{mi-class}} and / or on an object
of \code{\link{missing_data.frame-class}} after being coerced to a
\code{\link{data.frame}}
}
\usage{
mipply(X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE, 
       columnwise = TRUE, to.matrix = FALSE)
}
\arguments{
  \item{X}{Object of \code{\link{mi-class}}, \code{\link{missing_data.frame-class}}, 
    \code{\link{missing_variable-class}}, \code{\link{mi_list-class}}, or \code{\link{mdf_list-class}}
}
  \item{FUN}{Function to call}
  \item{\dots}{Further arguments passed to \code{FUN}, currently broken
}
  \item{simplify}{If \code{TRUE}, coerces result to a vector or matrix if possible
}
  \item{USE.NAMES}{ignored but included for compatibility with \code{\link{sapply}}
}
  \item{columnwise}{logical indicating whether to invoke \code{FUN} on the columns
    of a \code{\link{missing_data.frame}} after coercing it to a \code{\link{data.frame}}
    or a \code{\link{matrix}} or to invoke \code{FUN} on the \dQuote{whole} 
    \code{\link{data.frame}} or \code{\link{matrix}}
}
  \item{to.matrix}{Logical indicating whether to coerce each \code{\link{missing_data.frame}}
     to a numeric \code{\link{matrix}} or to a \code{\link{data.frame}}. The default is
     \code{FALSE}, in which case the \code{\link{data.frame}} will include \code{\link{factor}}s
      if any of the \code{\link{missing_variable}}s inherit from \code{\link{categorical-class}}
}  
}
\details{
The \code{columnwise} and \code{to.matrix} arguments are the only additions to the argument list 
in \code{\link{sapply}}, see the Examples section for an illustration of their use. Note that 
functions such as \code{\link{mean}} only accept \code{\link{numeric}} inputs, which can produce
errors or warnings when \code{to.matrix = FALSE}.
}
\value{
A list, vector, or matrix depending on the arguments
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{sapply}}
}
\examples{
if(!exists("imputations", env = .GlobalEnv)) {
  imputations <- mi:::imputations # cached from example("mi-package")
}
round(mipply(imputations, mean, to.matrix = TRUE), 3)
mipply(imputations, summary, columnwise = FALSE)
}
\keyword{utilities}

\name{CHAIN}
\docType{data}
\alias{CHAIN}
\title{
  Subset of variables from the CHAIN project
}
\description{
  The CHAIN project was a longitudinal cohort study of people living with HIV in New York City, 
  which was recruited in 1994 from a large number of medical care and social service agencies
  serving HIV in New York City. This subset of data pertain to the sixth round of interviews.
}
\usage{data(CHAIN)}
\format{
  A \code{\link{data.frame}} with 532 observations on the following 8 variables.
  \describe{
    \item{\code{log_virus}}{
      log of self reported viral load level, where zero represents an undetectable level.
    }
    \item{\code{age}}{
      age at time of the interview
    }
    \item{\code{income}}{
      annual family income in 10 intervals
    }
    \item{\code{healthy}}{
      a continuous scale of physical health with a theoretical range between 0 and 100 where
      better health is associated with higher scale values
    }
    \item{\code{mental}}{
      a binary measure of poor mental health ( 1=Yes, 0=No )
    }
    \item{\code{damage}}{
      ordered interval for the CD4 count, which is an
      indicator of how much damage HIV has caused to the immune system
    }
    \item{\code{treatment}}{
      a three-level ordered variable:
        0=Not currently taking HAART (Highly Active AntiretRoviral Therapy)
        1=taking HAART but nonadherent,
        2=taking HAART and adherent
    }
  }
}
\details{
  A missing value in the log virus load level 
  was assigned to individuals who either could not recall
  their viral load level, did not have a viral load test in the six month
  preceding the interview, or reported their viral loads as "good" or "bad".
}
\source{
  http://cchps.columbia.edu/research.cfm
}
\references{
  Messeri P, Lee G, Abramson DA, Aidala A, Chiasson MA, Jones JD. (2003).
  \dQuote{Antiretroviral therapy and declining AIDS mortality in New York City}.
  \emph{Medical Care} 41:512--521.
}
\keyword{datasets}

\name{01missing_variable}
\Rdversion{1.1}
\docType{class}
\alias{01missing_variable}
\alias{missing_variable}
\alias{missing_variable-class}
\alias{MatrixTypeThing-class}
\alias{WeAreFamily-class}
\title{Class "missing_variable" and Inherited Classes}
\description{
The missing_variable class is essentially the data comprising a variable plus all
the metadata needed to understand how its missing values will be imputed. However, no variable is
merely of missing_variable class; rather every variable is of a class that inherits from the 
missing_variable class. Even if a variable has no missing values, it needs to be coerced to a class 
that inherits from the missing_variable class before it can be used to impute values of other 
missing_variables. Understanding the properties of different subclasses of the missing_variable class 
is essential for modeling and imputing them. The \code{\link{missing_data.frame-class}} is essentially
a list of objects that inherit from the missing_variable class, plus metadata need to understand how
these missing_variables relate to each other.  Most users will never need to call \code{missing_variable} directly since it is called by \code{\link{missing_data.frame}}.
}
\section{Objects from the Classes}{The missing_variable class is virtual, so no objects 
  may be created from it. However, the missing_variable generic function can be used to 
  instantiate an object that inherits from the missing_variable class by specifying its 
  \code{type} argument. A user would call the \code{\link{missing_data.frame}}
  function on a \code{\link{data.frame}}, which in turn calls the missing_variable function
  on each column of the \code{\link{data.frame}} using various heuristics to guess the
  \code{type} argument.
}
\usage{
missing_variable(y, type, ...)
## Hidden arugments not included in the signature:
## favor_ordered = TRUE, favor_positive = FALSE, 
## variable_name = deparse(substitute(y))
}
\arguments{
  \item{y}{Can be any vector, some of whose values may be \code{\link{NA}}, which will
    comprise the \bold{raw_data} slot of a missing_variable (see the Slots section). It is 
    recommended that this vector \emph{not} have any transformations, such as a log-transformation. 
    Any continuous variable can be transformed using the function in its \bold{transformation} slot. 
    The transformations and other discretionary aspects of a missing_variable are typically changed
    by calling the \code{\link{change}} function on a \code{\link{missing_data.frame}} 
    See the Slots section for more details.
}
  \item{type}{Missing or a character string among the classes that inherit from the missing_variable
    class. If missing, the constructor will guess (sometimes incorrectly) based on the characteristics 
    of the variable. The best way to improve the guessing of categorical variables is to 
    use the \code{\link{factor}} function --- possibly with \code{ordered = TRUE} --- to create
    (possibly ordered) factors that will correctly be coerced to objects of 
    \code{\link{unordered-categorical-class}} and \code{\link{ordered-categorical-class}} respectively. 
    If you fail to do so, the hidden arguments that are not included in the signature affect the guesses. 
    If \code{favor_ordered = TRUE}, which is the default, it will tend to guess that variables with few 
    unique values are should be coerced to \code{\link{ordered-categorical-class}} and 
    \code{\link{unordered-categorical-class}} otherwise. If \code{favor_positive = FALSE}, which is the 
    default, it will tend to guess that variables with many unique values are merely continuous, whether 
    or not all the observed values are positive. If \code{favor_positive = TRUE} nonnegative or positive
    variables will get coerced to
    \code{\link{nonnegative-continuous-class}} or \code{\link{positive-continuous-class}}. See the Slots
    section and the specific help pages for more details on the subclasses.
}
  \item{\dots}{Further hidden arguments that are not in the signature. The \code{favor_ordered} and
   \code{favor_positive} arguments are documented immediately above. The \code{variable} name argument
   can be used to control what gets put in the \bold{variable_name} slot, see the Slots section below.
}
}
\section{Slots}{
  In the following table, indentation indicates inheritance from the class with less indentation, and
  italics indicates that the class is virtual so no variables can be created with that class. Inherited
  classes inherit the transformations, families, link functions, and \code{\link{fit_model-methods}}
  from their parent class, although these are often superceeded by analogues that are tailored for the
  inherited class. Also note, the default transformation for the continuous class is a standardization 
  using \emph{twice} the standard deviation of the observed values.

  The distinction between the transformation entailed by the \code{\link{family}} and the transformation
  entailed by the function in the \bold{tranformation} slot may be confusing at this point. The former pertains
  to how the linear predictor of a variable is mapped to the space of a variable when it is on the left-hand
  side of a generalized linear model. The latter pertains --- for continuous variables only --- to how the
  values in the \bold{raw_data} slot are mapped into those in the \bold{data} and thus affects how a continuous 
  variable enters into the model whether it is on the left or right-hand side. The classes are discussed in 
  much more detail below.
  \tabular{lll}{
    \bold{Class name [transformation]}   \tab \bold{Default family and link} \tab \bold{Default \code{\link{fit_model}}} \cr
    \emph{missing_variable}            \tab none                  \tab throws error \cr
    \code{  } \emph{categorical}                  \tab none        \tab throws error \cr
    \code{  } \code{  } unordered-categorical         \tab \code{binomial(link = 'logit')} \tab \code{\link[nnet]{multinom}} \cr
    \code{  } \code{  } ordered-categorical           \tab \code{binomial(link = 'logit')} \tab \code{\link[arm]{bayespolr}} \cr
    \code{  } \code{  } \code{  } binary              \tab \code{binomial(link = 'logit')} \tab \code{\link[arm]{bayesglm}} \cr
    \code{  } \code{  } \code{  } interval            \tab \code{gaussian{link = 'identity'}} \tab \code{\link[survival]{survreg}} \cr
    \code{  } continuous[standardize]                 \tab \code{gaussian{link = 'identity'}} \tab \code{\link[arm]{bayesglm}} \cr
    \code{  } \code{  } semi-continuous[identity]     \tab                                     \tab                             \cr
    \code{  } \code{  } \code{  } nonnegative-continuous[logshift] \tab                     \tab \cr
    \code{  } \code{  } \code{  } \code{  } SC_proportion[squeeze] \tab  \code{binomial(link = 'logit')} \tab \code{\link[betareg]{betareg}} \cr
    \code{  } \code{  } positive-continuous[\code{\link{log}}]           \tab                    \tab                              \cr
    \code{  } \code{  } \code{  } proportion[identity] \tab              \code{binomial(link = 'logit')} \tab \code{\link[betareg]{betareg}} \cr
    \code{  } \code{  } bounded-continuous[identity]  \tab \tab \cr
    \code{  } count                           \tab \code{quasipoisson{link = 'log'}}     \tab      \code{\link[arm]{bayesglm}}    \cr
    \code{  } irrelevant                      \tab                                       \tab      throws error                   \cr
    \code{  } \code{  } fixed                 \tab                               \tab      throws error                   \cr
  }

  The missing_variable class is virtual and has the following slots (this information is primarily directed at developeRs):
  \describe{
    \item{\code{variable_name}:}{Object of class \code{\link{character}} of length one naming the variable}
    \item{\code{raw_data}:}{Object of class \code{"ANY"} representing the observations
      on a variable, some of which may be \code{\link{NA}}. No method should ever change 
      this slot at all. Instead, methods should change the \bold{data} slot.}
    \item{\code{data}:}{Object of class \code{"ANY"}, which is initially a copy of the
      \bold{raw_data} slot --- transformed by the function in the \bold{transformation} slot 
       for continuous variables only --- and whose \code{\link{NA}} values are replaced during
       the multiple imputation process. See \code{\link{mi}}}
    \item{\code{n_total}:}{Object of class \code{"integer"} which is the \code{\link{length}}
       of the \bold{data} slot}
    \item{\code{all_obs}:}{Object of class \code{"logical"} of length one indicating whether
       all values of the \bold{data} slot are observed and thus not \code{\link{NA}} }
    \item{\code{n_obs}:}{Object of class \code{"integer"} of length one indicating the number
       of values of the \bold{data} slot that are observed and thus not \code{\link{NA}} }
    \item{\code{which_obs}:}{Object of class \code{"integer"}, which is a vector indicating 
       the positions of the observed values in the \bold{data} slot}
    \item{\code{all_miss}:}{Object of class \code{"logical"} of length one indicating whether
       all values of the \bold{data} slot are \code{\link{NA}} }
    \item{\code{n_miss}:}{Object of class \code{"integer"} of length one indicating the number
       of values of the \bold{data} slot that are \code{\link{NA}} }
    \item{\code{which_miss}:}{Object of class \code{"integer"}, which is a vector indicating
       the positions of the missing values in the \bold{data} slot }
    \item{\code{n_extra}:}{Object of class \code{"integer"} of length one indicating how many
       (missing) observations have been added to the end of the \bold{data} slot that are not
       included in the \bold{raw_data} slot. Although the extra values will be imputed, they
       are not considered to be \dQuote{missing} for the purposes of defining the previous
       three slots}
    \item{\code{which_extra}:}{Object of class \code{"integer"}, which is a vector indicating
       the positions of the extra values at the end of the \bold{data} slot }
    \item{\code{n_unpossible}:}{Object of class \code{"integer"} of length one indicating the
       number of values that are logically or structurally unobservable}
    \item{\code{which_unpossible}:}{Object of class \code{"integer"} indicating the positions
       of the unpossible values in the \bold{data} slot } 
    \item{\code{n_drawn}:}{Object of class \code{"integer"} of length one which is the sum of
       the \bold{n_miss} and \bold{n_extra} slots}
    \item{\code{which_drawn}:}{Object of class \code{"integer"} which is a vector concatinating
       the \bold{which_miss} and \bold{which_extra} slots }
    \item{\code{imputation_method}:}{Object of class \code{"character"} of length one indicating
       how the \code{\link{NA}} values are to be imputed. Possibilities include \dQuote{ppd} for
       imputation from the posterior predictive distribution, \dQuote{pmm} for imputation via
       predictive mean matching, \dQuote{mean} for mean-imputation, \dQuote{median} for 
       median-imputation, \dQuote{expectation} for conditional mean-imputation. With enough
       programming effort, other kinds of imputation can be defined and specified here.}
    \item{\code{family}:}{Object of class \code{"WeAreFamily"} that will typically be passed to 
       \code{\link{glm}} and similar functions during the multiple imputation process}
    \item{\code{known_families}:}{Object of class \code{\link{character}} indicating the families
       that are known to be supported for a class; see \code{\link{family}}}
    \item{\code{known_links}:}{Object of class \code{\link{character}} indicating what link functions
       are known to be supported by the elements of the \bold{known_families} slot; see 
       \code{\link{family}}}
    \item{\code{imputations}:}{Object of class \code{"MatrixTypeThing"} with rows equal to the number
       of iterations (initially zero) of the multiple imputation algorithm and columns equal to the 
       \bold{n_drawn} slot. The rows are appropriately extended and then filled by the 
       \code{\link{mi}} function}
    \item{\code{done}:}{Object of class \code{"logical"} of length one indicating whether the
       \code{\link{NA}} values in the \bold{data} slot have been replaced by imputed values}
    \item{\code{parameters}:}{Object of class \code{"MatrixTypeThing"} with rows equal to the number
       of iterations (initially zero) of the multiple imputation algorithm and columns equal to the number 
       of estimated parameters when modeling the \bold{data} slot. The rows are appropriately extended
       and then filled by the \code{\link{mi}} function}
    \item{\code{model}:}{Object of class \code{"ANY"} which can be filled by an object that is output
       by one of the \code{\link{fit_model-methods}}, which is done by default by \code{\link{mi}}
       when all the iterations have completed}
    \item{\code{fitted}:}{Object of class \code{"ANY"} although typically a vector or matrix that 
       contains the fitted values of the model in the slot immediately above. Note that the
       \bold{fitted} slot is filled by default by \code{\link{mi}}, although the \bold{model} slot
       is left empty by default to save RAM.}
    \item{\code{estimator}:}{Object of class \code{"character"} of length one indicating which pre-existing 
    	\code{\link{fit_model}} to use for an unordered-categorical variable.  Options are \code{"mnl"}, in which 
    	\code{\link[nnet]{multinom}} from the \pkg{nnet} package is used to fit the values of the unordered 
    	categorical variable; and \code{"rnl"}, in which each category is separately modeled as the positive 
    	binary outcome against all other categories using a \code{\link[arm]{bayesglm}} \code{fit_model} and 
    	the probabilities of each category are normalized to sum to 1 after each model is run. In general, 
    	\code{"rnl"} is slightly less accurate than \code{"mnl"}, but runs much more quickly especially when 
    	the unordered categorical variable has many unique categories.}
  }
  The WeAreFamily class is a class union of \code{\link{character}} and \code{\link{family}}, while the
  MatrixTypeThing class is a class union of \code{\link{matrix}} only at the moment.
}
\value{
The missing_variable function returns an object that inherits from the missing_variable class.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_data.frame}}, \code{\link{categorical-class}},  \code{\link{unordered-categorical-class}}, 
\code{\link{ordered-categorical-class}}, \code{\link{binary-class}}, \code{\link{interval-class}}, 
\code{\link{continuous-class}}, \code{\link{semi-continuous-class}}, \code{\link{nonnegative-continuous-class}},
\code{\link{SC_proportion-class}}, \code{\link{censored-continuous-class}}, 
\code{\link{truncated-continuous-class}}, \code{\link{bounded-continuous-class}}, 
\code{\link{positive-continuous-class}}, \code{\link{proportion-class}}, \code{\link{count-class}}
}
\examples{
# STEP 0: GET DATA
data(nlsyV, package = "mi")

# STEP 0.5 CREATE A missing_variable (you never need to actually do this)
income <- missing_variable(nlsyV$income, type = "continuous")
show(income)

# STEP 1: CONVERT IT TO A missing_data.frame
mdf <- missing_data.frame(nlsyV) # this calls missing_variable() internally
show(mdf)
}
\keyword{classes}
\keyword{AimedAtUseRs}
\keyword{DirectedTowardDevelopeRs}

\name{categorical}
\Rdversion{1.1}
\docType{class}
\alias{categorical}
\alias{categorical-class}
\alias{unordered-categorical-class}
\alias{ordered-categorical-class}
\alias{interval-class}
\alias{binary-class}
\alias{grouped-binary-class}
\title{Class "categorical" and Inherited Classes}
\description{
The categorical class is a virtual class that inherits from the \code{\link{missing_variable-class}}
and is the parent of the unordered-categorical and ordered-categorical classes. The ordered-categorical
class is the parent of both the binary and interval classes. Aside from these facts, the rest of the 
documentation here is primarily directed toward developers.
}
\section{Objects from the Classes}{The categorical class is virtual, so no objects 
  may be created from it. However, the \code{\link{missing_variable}} generic function can be used to 
  instantiate an object that inherits from the categorical class by specifying 
  \code{type = "unordered-categorical"}, \code{type = "ordered-categorical"}, 
  \code{type = "binary"}, \code{type = "grouped-binary"}, or \code{type = "interval"}.
}
\section{Slots}{
  The unordered-categorical class inherits from the categorical class and has no additional slots
  but must have more than two uniquely observed values in its \code{raw_data} slot. The default \code{\link{fit_model}}
  method is a wrapper for the \code{\link[nnet]{multinom}} function in the \pkg{nnet} package. The ordered-categorical 
  class inherits from the categorical class and has one additional slot:
  \describe{
    \item{cutpoints}{Object of class \code{"numeric"} which is a vector of thresholds (sometimes estimated) that 
      govern how an assumed latent variable is divided into observed ordered categories}
  }
  The \code{\link{fit_model}} method for an ordered-categorical variable is, by default, a wrapper for 
  \code{\link[arm]{bayespolr}}. The binary class inherits from the ordered-categorical class and has no additional slots. 
  It must have exactly two uniquely observed values in its \code{raw_data} slot and its \code{\link{fit_model}} method is,
  by default, a wrapper for \code{\link[arm]{bayespolr}}. The grouped-binary class inherits from the binary class and
  has one additional slot:
  \describe{
    \item{strata}{Object of class \code{"character"} which is a vector (possibly of length one) of variable names that
    group the observations into strata. The named external variables should also be categorical.}
  }
  The default \code{\link{fit_model}} method for a grouped-binary variable is a wrapper for the \code{\link[survival]{clogit}}
  function in the \pkg{survival} package and the variables named in the \bold{strata} slot are passed to the
  \code{\link[survival]{strata}} function.
 
  The interval class inherits from the ordered-categorical class, has no additional slots, and is intended for variables 
  whose observed values are only known up to orderable intervals. Its \code{\link{fit_model}} method is, by default, a 
  wrapper for \code{\link[survival]{survreg}} even though it may or may not be a \dQuote{survival} model in any meaningful sense.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_variable}}
}
\examples{
# STEP 0: GET DATA
data(nlsyV, package = "mi")

# STEP 0.5 CREATE A missing_variable (you never need to actually do this)
momrace <- missing_variable(as.factor(nlsyV$momrace), type = "unordered-categorical")
show(momrace)
}
\keyword{classes}
\keyword{DirectedTowardDevelopeRs}

\name{07complete}
\docType{methods}
\alias{07complete}
\alias{complete}
\alias{complete-methods}
\title{Extract the Completed Data}
\description{
This function extracts several multiply imputed \code{\link{data.frame}}s 
from an object of \code{\link{mi-class}}.
}
\usage{
complete(y, m, ...)
}
\arguments{
  \item{y}{An object of \code{\link{mi-class}} (typically) or 
    \code{\link{missing_data.frame-class}} or 
    \code{\link{missing_variable-class}}
}
  \item{m}{If \bold{y} is an object of \code{\link{mi-class}}, then
    \code{m} must be a specified integer indicating how many multiply
    imputed \code{\link{data.frame}}s to return or, if missing, the 
    number of \code{\link{data.frame}}s will be equal to the length of the 
    \bold{data} slot in \code{y}. If \code{y} is not an object of \code{\link{mi-class}}, 
    then \bold{m} must be a specified integer indicating which iteration to use in 
    the resulting \code{\link{data.frame}}, where any non-positive integer is 
    a short hand for the last iteration.
}
  \item{\dots}{Other arguments, not currently utilized}
}
\details{
Several functions within \pkg{mi} use \code{complete}, although the only reason
in principle why a user should need to call \code{complete} is to create 
\code{\link{data.frame}}s to export to another program. For analysis, it is 
better to use the \code{\link{pool}} function, although currently \code{\link{pool}}
might not offer all the necessary functionality.
}
\value{
If \bold{y} is an object of \code{\link{mi-class}} and \code{m > 1}, a \code{\link{list}} 
of \code{m} \code{\link{data.frame}}s is returned. Otherwise, a single \code{\link{data.frame}}
is returned.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{mi-class}}
}
\examples{
if(!exists("imputations", env = .GlobalEnv)) {
  imputations <- mi:::imputations # cached from example("mi-package")
}
data.frames <- complete(imputations, 3)
lapply(data.frames, summary)
}
\keyword{manip}
\keyword{AimedAtUseRs}

\name{positive-continuous-class}
\Rdversion{1.1}
\docType{class}
\alias{positive-continuous-class}
\alias{proportion-class}
\title{Class "positive-continuous" and Inherited Classes}
\description{
The positive-continuous class inherits from the \code{\link{continuous-class}} and is the parent of the proportion class.
In both cases, no observations can be zero, and in the case of the proportion class, no observations can be one. The 
\code{\link{nonnegative-continuous-class}} and the \code{\link{SC_proportion-class}} are appropriate for those situations.
Aside from these facts, the rest of the 
documentation here is primarily directed toward developeRs.
}
\section{Objects from the Classes}{Objects can be created that are of positive-continuous or proportion class via the
  \code{\link{missing_variable}} generic function by specifying \code{type = "positive-continuous"} or
  \code{type = "proportion"}
}
\section{Slots}{
  The default transformation for the positive-continuous class is the \code{\link{log}} function. The proportion class inherits 
  from the positive-continuous class and has the identity transformation and the \code{\link{binomial}} family as defaults, in
  which case the \code{\link{fit_model-methods}} call the \code{\link[betareg]{betareg}} function in the \pkg{betareg} package. 
  Alternatively, the transformation could be an inverse CDF like the \code{\link{qnorm}} function and the family could be \code{\link{gaussian}},
  in which case the \code{\link{fit_model-methods}} call the \code{\link[arm]{bayesglm}} function in the \pkg{arm} package.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_variable}}, \code{\link{continuous-class}}, \code{\link{positive-continuous-class}},
\code{\link{proportion-class}}
}
\examples{
# STEP 0: GET DATA
data(CHAIN, package = "mi")

# STEP 0.5 CREATE A missing_variable (you never need to actually do this)
healthy <- missing_variable(CHAIN$healthy / 100, type = "proportion")
show(healthy)
}
\keyword{classes}
\keyword{DirectedTowardDevelopeRs}

\name{censored-continuous-class}
\Rdversion{1.1}
\docType{class}
\alias{truncated-continuous-class}
\alias{truncated-continuous}
\alias{FF_truncated-continuous-class} 
\alias{FN_truncated-continuous-class}
\alias{NF_truncated-continuous-class} 
\alias{NN_truncated-continuous-class}
\alias{censored-continuous-class}
\alias{censored-continuous}
\alias{FF_censored-continuous-class} 
\alias{FN_censored-continuous-class}
\alias{NF_censored-continuous-class} 
\alias{NN_censored-continuous-class}
\title{The "censored-continuous" Class, the "truncated-continuous" Class and Inherited Classes}
\description{
The censored-continuous class and the truncated-continuous class are both virtual and both inherit from the \code{\link{continuous-class}}
and each is the parent of four classes that differ depending on whether the lower and upper bounds are numeric vectors or functions. A
censored observation is one whose exact value is not observed. A truncated observation is one whose exact value is not observed and which
implies that values on some \emph{other} variables are not observed for that unit of observation. An example of truncation might be that 
some taxation forms are not required when a person's income falls below a certain threshold. The methods for these classes are not
working yet. Aside from these facts, the rest of the documentation here is primarily directed toward developeRs.
}
\section{Objects from the Classes}{Both the censored-continuous class and the truncated-continuous class are virtual, so no objects can be 
  created with these classes. However, the \code{\link{missing_variable}} generic function can be used to create an object that inherits
  from one of their subclasses by specifying \code{type = "NNcensored-continuous"}, \code{type = "NFcensored-continuous"}, 
  \code{type = "FNcensored-continuous"}, \code{type = "FFcensored-continuous"}, \code{type = "NNtruncated-continuous"}, \code{type = "NFtruncated-continuous"}, 
  \code{type = "FNtruncated-continuous"}, \code{type = "FFtruncated-continuous"}. When doing so, the lower and upper slots need to be
  specified appropriately.
}
\section{Slots}{
  The censored-continuous class and the truncated-continuous class are both virtual, both inherit from
  the continuous class, both use the identity transformation by default, and both have two additional slots:
  \describe{
    \item{upper}{The upper bound for each observation}
    \item{lower}{The lower bound for each observation}
  }
  Both the censored-continuous class and the truncated-continuous class have four subclasses that differ depending 
  on whether the upper and / or lower bounds are numeric vectors or functions that output numeric 
  vectors (scalars are recycled and can be \code{Inf}). These subclasses are 
  \describe{
    \item{NN_censored-continuous}{where both the lower and upper bounds are numeric vectors}
    \item{FN_censored-continuous}{where the lower bound is a function and the upper bound is a numeric vector}
    \item{NF_censored-continuous}{where the lower bound is a numeric vector and the upper bound is a function}
    \item{FF_censored-continuous}{where both the lower and upper bounds are functions}
    \item{NN_truncated-continuous}{where both the lower and upper bounds are numeric vectors}
    \item{FN_truncated-continuous}{where the lower bound is a function and the upper bound is a numeric vector}
    \item{NF_truncated-continuous}{where the lower bound is a numeric vector and the upper bound is a function}
    \item{FF_truncated-continuous}{where both the lower and upper bounds are functions}
  }
}
\author{
Ben Goodrich, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_variable}}, \code{\link{continuous-class}}
}
\examples{
# STEP 0: GET DATA
data(CHAIN, package = "mi")

# STEP 0.5 CREATE A missing_variable (you never need to actually do this)
#log_virus <- missing_variable(CHAIN$log_virus, type = "NN_censored-continuous", 
#                              lower = 0, upper = Inf)
#show(log_virus)
}
\keyword{classes}
\keyword{DirectedTowardDevelopeRs}

\name{06pool}
\alias{06pool}
\Rdversion{1.1}
\docType{class}
\alias{pool}
\alias{pooled-class}
\alias{pooled-methods}
\alias{display,pooled-method}

\title{Estimate a Model Pooling Over the Imputed Datasets}
\description{
This function estimates a chosen model, taking into account the additional
uncertainty that arises due to a finite number of imputations of the missing
data.
}
\usage{
pool(formula, data, m = NULL, FUN = NULL, ...)
}
\arguments{
  \item{formula}{a \code{\link{formula}} in the same syntax as used by \code{\link{glm}}
}
  \item{data}{an object of \code{\link{mi-class}}
}
  \item{m}{number of completed datasets to average over, which if \code{NULL} defaults to
    the number of chains used in \code{\link{mi}}
}
  \item{FUN}{Function to estimate models or \code{NULL} which uses the same function as
    used in the \code{\link{fit_model-methods}} for the dependent variable
}
  \item{\dots}{further arguments passed to \code{FUN}
}
}
\details{
\code{FUN} is estimated on each of the \code{m} completed datasets according to the given
\code{formula} and the results are combined using the Rubin Rules.
}
\value{
An object of class \code{"pooled"} whose definition is subject to change but it has a 
\code{\link{summary}} and \code{\link[arm]{display}} method.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{mi}}
}
\examples{
if(!exists("imputations", env = .GlobalEnv)) {
  imputations <- mi:::imputations # cached from example("mi-package")
}
analysis <- pool(ppvtr.36 ~ first + b.marr + income + momage + momed + momrace, 
                 data = imputations)
display(analysis)
}
\keyword{regression}
\keyword{AimedAtUseRs}

\name{continuous}
\Rdversion{1.1}
\docType{class}
\alias{continuous}
\alias{continuous-class}
\title{Class "continuous"}
\description{
The continuous class inherits from the \code{\link{missing_variable-class}} and is the parent of the following
classes: \code{\link{semi-continuous}}, \code{\link{censored-continuous}}, \code{\link{truncated-continuous}}, 
and \code{\link{bounded-continuous}}. The distinctions
among these subclasses are given on their respective help pages. Aside from these facts, the rest of the 
documentation here is primarily directed toward developers.
}
\section{Objects from the Classes}{Objects can be created that are of class continuous via
  the \code{\link{missing_variable}} generic function by specifying \code{type = "continuous"}
}
\section{Slots}{
  The continuous class inherits from the \code{\link{missing_variable}} class and has the following additional slots:
  \describe{
    \item{transformation}{Object of class \code{"function"} which is passed the \code{raw_data} slot and
      whose returned value is assigned to the \code{data} slot. By default, this function is the 
      \dQuote{standardize} transformation, using the mean and \emph{twice} the standard deviation of the 
      observed values}
    \item{inverse_transformation}{Object of class \code{"function"} which is the inverse of the function
      in the \code{transformation} slot.}
    \item{transformed}{Object of class \code{"logical"} of length one indicating whether the 
      \code{data} slot is in the \dQuote{transformed} state or the \dQuote{untransformed} state}
    \item{known_transformations}{Object of class \code{"character"} indicating which transformations
      are possible for this variable}
  }
  The \code{\link{fit_model}} method for a continuous variable is, by default, a wrapper for
  \code{\link[arm]{bayesglm}} and its \code{family} slot is, by default, \code{\link{gaussian}}

}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_variable}}, \code{\link{semi-continuous-class}}, \code{\link{censored-continuous-class}},
\code{\link{truncated-continuous-class}}, \code{\link{bounded-continuous-class}}
}
\examples{
# STEP 0: GET DATA
data(nlsyV, package = "mi")

# STEP 0.5 CREATE A missing_variable (you never need to actually do this)
income <- missing_variable(nlsyV$income, type = "continuous")
show(income)
}
\keyword{classes}
\keyword{DirectedTowardDevelopeRs}

\name{multilevel_missing_data.frame}
\Rdversion{1.1}
\docType{class}
\alias{multilevel_missing_data.frame}
\alias{multilevel_missing_data.frame-class}
\title{Class "multilevel_missing_data.frame"}
\description{
This class inherits from the \code{\link{missing_data.frame-class}} but is customized for the situation
where the sample has a multilevel structure.
}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("multilevel_missing_data.frame", ...)}.
However, its users almost always will pass a \code{\link{data.frame}} to the 
\code{\link{missing_data.frame}} function and specify the \code{subclass} and \code{groups} arguments.
}
\section{Slots}{
  The multilevel_missing_data.frame class inherits from the \code{\link{missing_data.frame-class}} and
  has two additional slots
  \describe{
    \item{groups}{Object of class \code{\link{character}} indicating which variables define the 
      multilevel structure}
    \item{mdf_list}{Object of class \code{mdf_list} whose elements contain a \code{\link{missing_data.frame}}
      for each group. This slot is filled automatically by the \code{\link{initialize}} method.}
  }
}
\details{
  The \code{\link{fit_model-methods}} for the multilevel_missing_data.frame class will, by default, utilize 
  multilevel modeling techniques that shrink the estimated parameters for each group toward their global
  means.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_data.frame}}
}
\examples{
## Write example
}
\keyword{classes}
\keyword{manip}
\keyword{AimedAtUseRs}

\name{allcategorical_missing_data.frame}
\Rdversion{1.1}
\docType{class}
\alias{allcategorical_missing_data.frame}
\alias{allcategorical_missing_data.frame-class}
\title{Class "allcategorical_missing_data.frame"}
\description{
This class inherits from the \code{\link{missing_data.frame-class}} but is customized for the situation where all the variables are categorical.
}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("allcategorical_missing_data.frame", ...)}.
However, its users almost always will pass a \code{\link{data.frame}} to the 
\code{\link{missing_data.frame}} function and specify the \code{subclass} argument.
}
\section{Slots}{
  The allcategorical_missing_data.frame class inherits from the \code{\link{missing_data.frame-class}} and
  has three additional slots
  \describe{
    \item{Hstar}{Positive integer indicating the maximum number of latent classes}
    \item{parameters}{A list that holds the current realization of the unknown parameters}
    \item{latents}{An object of \code{\link{unordered-categorical-class}} that contains
      the current realization of the latent classes}
  }
}
\details{
  The \code{\link{fit_model-methods}} for the allcategorical_missing_data.frame class 
  implement a Gibbs sampler. However, it does not utilize any ordinal information that 
  may be available. Continuous variables should be made into factors using the
  \code{\link{cut}} command before calling \code{\link{missing_data.frame}}.
}
\author{
Sophie Si for the algorithm and Ben Goodrich for the R implementation
}
\seealso{
\code{\link{missing_data.frame}}
}
\examples{
rdf <- rdata.frame(n_full = 2, n_partial = 2, 
                   restrictions = "stratified", types = "ord")
mdf <- missing_data.frame(rdf$obs, subclass = "allcategorical")
}
\keyword{classes}
\keyword{manip}
\keyword{AimedAtUseRs}

\name{02missing_data.frame}
\Rdversion{1.1}
\docType{class}
\alias{02missing_data.frame}
\alias{missing_data.frame-class}
\alias{missing_data.frame}
\title{Class "missing_data.frame"}
\description{
This class is similar to a \code{\link{data.frame}} but is customized for the situation in 
which variables with missing data are being modeled for multiple imputation. This class primarily 
consists of a list of \code{\link{missing_variable}}s plus slots containing metadata indicating how the
\code{\link{missing_variable}}s relate to each other. Most operations that work for a
\code{\link{data.frame}} also work for a missing_data.frame.
}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("missing_data.frame", ...)}.
However, useRs almost always will pass a \code{\link{data.frame}} to the 
missing_data.frame constructor function to produce an object of missing_data.frame class.
}
\usage{
missing_data.frame(y, ...)
## Hidden arguments not included in the signature
## favor_ordered = TRUE, favor_positive = FALSE, 
## subclass = NA_character_,
## include_missingness = TRUE, skip_correlation_check = FALSE
}
\arguments{
  \item{y}{Usually a \code{\link{data.frame}}, possibly a numeric matrix, 
    possibly a list of \code{\link{missing_variable}}s.}
  \item{\dots}{Hidden arguments. The \code{favor_ordered} and \code{favor_positive}
    arguments are passed to the \code{\link{missing_variable}} function and are 
    documented under the \code{type} argument. Briefly, they affect the heuristics
    that are used to guess what class a variable should be coerced to. The 
    \code{subclass} argument defaults to \code{\link{NA}} and can be used to specify
    that the resulting object should inherit from the missing_data.frame class
    rather than be an object of \code{missing_data.frame} class.

    Any further arguments are passed to the \code{\link{initialize-methods}} for
    a missing_data.frame. They currently are \code{include_missingness}, which 
    defaults to \code{TRUE} and indicates that the missingness pattern of the other
    variables should be included when modeling a particular \code{\link{missing_variable}}, 
    and \code{skip_correlation_check}, which defaults to FALSE and indicates whether
    to skip the default check for whether the observed values of each pair of \code{\link{missing_variable}}s 
    has a perfect absolute Spearman \code{\link{cor}}relation.
}
}
\section{Slots}{
  This section is primarily aimed at developeRs. A missing_data.frame inherits from
  \code{\link{data.frame}} but has the following additional slots:
  \describe{
    \item{\code{variables}:}{Object of class \code{"list"} and each list element
      is an object that inherits from the \code{\link{missing_variable-class}} }
    \item{\code{no_missing}:}{Object of class \code{"logical"}, which is a vector
      whose length is the same as the length of the \bold{variables} slot indicating 
      whether the corresponding \code{\link{missing_variable}} is fully observed }
    \item{\code{patterns}:}{Object of class \code{\link{factor}} whose length is equal
      to the number of observation and whose elements indicate the missingness pattern
      for that observation}
    \item{\code{DIM}:}{Object of class \code{"integer"} of length two indicating
      first the number of observations and second the length of the \bold{variables}
      slot }
    \item{\code{DIMNAMES}:}{Object of class \code{"list"} of length two providing
      the appropriate number rownames and column names }
    \item{\code{postprocess}:}{Object of class \code{"function"} used to create
      additional variables from existing variables, such as interactions between
      two \code{\link{missing_variable}}s once their missing values have been
      imputed. Does not work at the moment}
    \item{\code{index}:}{Object of class \code{"list"} whose length is equal to 
      the number of \code{\link{missing_variable}}s with some missing values. Each
      list element is an integer vector indicating which columns of the \bold{X}
      slot must be dropped when modeling the corresponding \code{\link{missing_variable}} }
    \item{\code{X}:}{Object of \code{\link{MatrixTypeThing-class}} with rows equal to the
      number of observations and is loosely related to a \code{\link{model.matrix}}. Rather 
      than repeatedly parsing a \code{\link{formula}} during the multiple imputation process,
      this \bold{X} matrix is created once and some of its columns are dropped when
      modeling a \code{\link{missing_variable}} utilizing the \bold{index} slot.
      The columns of the \bold{X} matrix consists of numeric representations of the 
      \code{\link{missing_variable}}s plus (by default) the unique missingness patterns }
    \item{\code{weights}:}{Object of class \code{"list"} whose length is equal to one
       or the number of \code{\link{missing_variable}}s with some missing values. Each 
       list element is passed to the corresponding argument of \code{\link[arm]{bayesglm}} 
       and similar functions. In particular, some observations can be given a weight
       of zero, which should drop them when modeling some \code{\link{missing_variable}}s}
    \item{\code{priors}:}{Object of class \code{"list"} whose length is equal to the number
       of \code{\link{missing_variable}}s and whose elements give appropriate values for
       the priors used by the model fitting function wraped by the \code{\link{fit_model-methods}}; 
       see, e.g., \code{\link[arm]{bayesglm}}}
    \item{\code{correlations}:}{Object of class \code{"matrix"} with rows and
        columns equal to the length of the \bold{variables} slot. Its strict upper
        triangle contains Spearman \code{\link{cor}}relations between pairs of
        variables (ignoring missing values), and its strict lower triangle contains
        Squared Multiple Correlations (SMCs) between a variable and all other
        variables (ignoring missing values). If either a Spearman correlation or
        a SMC is very close to unity, there may be difficulty or error messages
        during the multiple imputation process.}
    \item{\code{done}:}{Object of class \code{"logical"} of length one indicating
        whether the missing values have been imputed}
    \item{\code{workpath}:}{Object of class \code{\link{character}} of length one indicating
        the path to a working directory that is used to store some objects}
  }
}
\details{
In most cases, the first step of an analysis is for a useR to call the 
\code{missing_data.frame} function on a \code{\link{data.frame}} whose variables
have some \code{\link{NA}} values, which will call the \code{\link{missing_variable}}
function on each column of the \code{\link{data.frame}} and return the \code{\link{list}}
that fills the \bold{variable} slot. The classes of the list elements will depend on the
nature of the column of the \code{\link{data.frame}} and various fallible heuristics. The
success rate can be enhanced by making sure that columns of the original 
\code{\link{data.frame}} that are intended to be categorical variables are 
(ordered if appropriate) \code{\link{factor}}s with labels. Even in the best case
scenario, it will often be necessary to utlize the \code{\link{change}} function to 
modify various discretionary aspects of the \code{\link{missing_variable}}s in the 
\bold{variables} slot of the missing_data.frame. The \code{\link{show}} method for
a missing_data.frame should be utilized to get a quick overview of the 
\code{\link{missing_variable}}s in a missing_data.frame and recognized what needs
to be \code{\link{change}}d.
}
\section{Methods}{
  There are many methods that are defined for a missing_data.frame, although some
  are primarily intended for developers. The most relevant ones for users are:
  \describe{
    \item{change}{\code{signature(data = "missing_data.frame", y = "ANY", what = "character", to = "ANY")}
      which is used to change discretionary aspects of the \code{\link{missing_variable}}s
      in the \bold{variables} slot of a missing_data.frame}
    \item{hist}{\code{signature(x = "missing_data.frame")} which shows histograms
      of the observed variables that have missingness}
    \item{image}{\code{signature(x = "missing_data.frame")} which plots 
      an image of the \bold{missingness} slot to visualize the pattern of missingness
      when \code{grayscale = FALSE} or the pattern of missingness in light of the
      observed values (\code{grayscale = TRUE}, the default)}
    \item{mi}{\code{signature(y = "missing_data.frame", model = "missing")} which 
      multiply imputes the missing values}
    \item{show}{\code{signature(object = "missing_data.frame")} which gives an overview
      of the salient characteristics of the \code{\link{missing_variable}}s in the 
      \bold{variables} slot of a missing_data.frame }
    \item{summary}{\code{signature(object = "missing_data.frame")} which produces the same
      result as the \code{\link{summary}} method for a \code{\link{data.frame}}}
  }
  There are also S3 methods for the \code{\link{dim}}, \code{\link{dimnames}}, and \code{\link{names}}
  generics, which allow functions like \code{\link{nrow}}, \code{\link{ncol}}, \code{\link{rownames}},
  \code{\link{colnames}}, etc. to work as expected on \code{missing_data.frame}s. Also, accessing
  and changing elements for a \code{missing_data.frame} mostly works the same way as for a
  \code{\link{data.frame}}
}
\value{
The \code{missing_data.frame} constructor function returns an object of class \code{missing_data.frame} 
or that inherits from the \code{missing_data.frame} class.
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{change}}, \code{\link{missing_variable}}, \code{\link{mi}},
\code{\link{experiment_missing_data.frame}}, \code{\link{multilevel_missing_data.frame}}
}
\examples{
# STEP 0: Get data
data(CHAIN, package = "mi")

# STEP 1: Convert to a missing_data.frame
mdf <- missing_data.frame(CHAIN) # warnings about missingness patterns
show(mdf)

# STEP 2: change things
mdf <- change(mdf, y = "log_virus", what = "transformation", to = "identity")

# STEP 3: look deeper
summary(mdf)
hist(mdf)
image(mdf)

# STEP 4: impute
\dontrun{
imputations <- mi(mdf)
}

## An example with subsetting on a fully observed variable
data(nlsyV, package = "mi")
mdfs <- missing_data.frame(nlsyV, favor_positive = TRUE, favor_ordered = FALSE, by = "first")
mdfs <- change(mdfs, y = "momed", what = "type", to = "ord")
show(mdfs)

}
\keyword{classes}
\keyword{manip}
\keyword{AimedAtUseRs}

\name{bounded-continuous-class}
\Rdversion{1.1}
\docType{class}
\alias{bounded-continuous-class}
\alias{bounded-continuous}
\title{Class "bounded-continuous"}
\description{
The bounded-continuous class inherits from the \code{\link{continuous-class}} and is intended for variables whose observations
fall within open intervals that have \emph{known} boundaries. Although proportions satisfy this definition, the 
\code{\link{proportion-class}} should be used in that case. At the moment, a bounded continuous variable is modeled as if it were
simply a continuous variable, but its \code{\link{mi-methods}} impute the missing values from a truncated normal distribution using
the \code{\link[truncnorm]{rtruncnorm}} function in the \pkg{truncnorm} package. Note that the default transformation is the identity
so if another transformation is used, the bounds must be specified on the transformed data. Aside from these facts, the rest of the 
documentation here is primarily directed toward developers.
}
\section{Objects from the Classes}{Objects can be created that are of bounded-continuous class via the 
  the \code{\link{missing_variable}} generic function by specifying \code{type = "bounded-continuous"} 
  as well as \code{lower} and / or \code{upper}
}
\section{Slots}{
  The bounded-continuous class inherits from the continuous class and is intended for variables that are supported on a known
  interval. Its default transformation function is the identity transformation and its \code{imputation_method} must be 
  \code{"ppd"}. It has two additional slots:
  \describe{
    \item{upper}{a numeric vector whose length is either one or the value of the \code{n_total} slot giving the upper bound for
      \emph{every} observation; \code{NA}s are not allowed}
    \item{lower}{a numeric vector whose length is either one or the value of the \code{n_total} slot giving the lower bound for
      \emph{every} observation; \code{NA}s are not allowed}
  }
}
\author{
Ben Goodrich and Jonathan Kropko, for this version, based on earlier versions written by Yu-Sung Su, Masanao Yajima,
Maria Grazia Pittau, Jennifer Hill, and Andrew Gelman.
}
\seealso{
\code{\link{missing_variable}}, \code{\link{continuous-class}}, \code{\link{positive-continuous-class}},
\code{\link{proportion-class}}
}
\examples{
# STEP 0: GET DATA
data(CHAIN, package = "mi")

# STEP 0.5 CREATE A missing_variable (you never need to actually do this)
lo_bound <- 0
hi_bound <- rep(Inf, nrow(CHAIN))
hi_bound[CHAIN$log_virus == 0] <- 6
log_virus <- missing_variable(ifelse(CHAIN$log_virus == 0, NA, CHAIN$log_virus), 
                              type = "bounded-continuous", lower = lo_bound, upper = hi_bound)

show(log_virus)
}
\keyword{classes}
\keyword{DirectedTowardDevelopeRs}

\name{mi-internal}
\alias{mi-internal}
\alias{change,missing_data.frame,ANY,ANY,character-method}
\alias{mi,missing_data.frame,missing-method}
\alias{plot,missing_data.frame,missing-method}
\alias{plot,mi,ANY-method}
\alias{show,missing_data.frame-method}
\alias{show,missing_variable-method}
\alias{summary,missing_data.frame-method}
\alias{summary,mi-method}
\alias{show,mi-method}

\alias{change_family,missing,missing_variable,family-method}
\alias{change_family,missing,proportion,family-method}
\alias{change_family,missing,unordered-categorical,family-method}
\alias{change_imputation_method,missing,missing_variable,character-method}
\alias{change_imputation_method,missing,missing_variable,missing-method}
\alias{change_link,missing,missing_variable,character-method}
\alias{change_link,missing,missing_variable,missing-method}
\alias{change_link,missing_data.frame,character,character-method}
\alias{change_link,missing_data.frame,numeric,character-method}
\alias{change_link,missing_data.frame,logical,character-method}
\alias{change_model,missing,missing_variable,character-method}
\alias{change_model,missing_data.frame,character,character-method}
\alias{change_model,missing_data.frame,numeric,character-method}
\alias{change_model,missing_data.frame,logical,character-method}
\alias{change_size,missing,missing_variable,integer-method}
\alias{change_size,missing,categorical,integer-method}
\alias{change_size,missing,fixed,integer-method}
\alias{change_size,missing_data.frame,missing,integer-method}
\alias{change_type,missing,missing_variable,character-method}
\alias{change_imputation_method,missing_data.frame,logical,character-method}
\alias{change_imputation_method,missing_data.frame,numeric,character-method}
\alias{change_transformation,missing,missing_variable,missing-method}
\alias{change_transformation,missing_data.frame,character,missing-method}
\alias{change_transformation,missing_data.frame,character,function-method}
\alias{change_transformation,missing_data.frame,numeric,function-method}
\alias{change_transformation,missing_data.frame,logical,function-method}
\alias{change_type,missing,missing_variable,missing-method}
\alias{change_type,missing_data.frame,character,missing-method}
\alias{complete,irrelevant,integer-method}
\alias{complete,categorical,integer-method}
\alias{complete,binary,integer-method}
\alias{complete,continuous,integer-method}
\alias{complete,nonnegative-continuous,integer-method}
\alias{complete,SC_proportion,integer-method}
\alias{complete,mi,integer-method}
\alias{complete,missing_data.frame,numeric-method}
\alias{complete,missing_variable,integer-method}
\alias{complete,mi,missing-method}
\alias{complete,mi_list,numeric-method}
\alias{complete,mi_list,missing-method}
\alias{fit_model,missing_variable,missing_data.frame-method}
\alias{fit_model,missing,missing_data.frame-method}
\alias{fit_model,continuous,missing_data.frame-method}
\alias{fit_model,semi-continuous,missing_data.frame-method}
\alias{fit_model,nonnegative-continuous,missing_data.frame-method}
\alias{fit_model,SC_proportion,missing_data.frame-method}
\alias{fit_model,proportion,missing_data.frame-method}
\alias{fit_model,truncated-continuous,missing_data.frame-method}
\alias{fit_model,censored-continuous,missing_data.frame-method}
\alias{fit_model,missing_variable,experiment_missing_data.frame-method}
\alias{fit_model,continuous,experiment_missing_data.frame-method}
\alias{fit_model,missing,multilevel_missing_data.frame-method}
\alias{fit_model,missing,mdf_list-method}
\alias{fit_model,binary,allcategorical_missing_data.frame-method}
\alias{fit_model,missing,allcategorical_missing_data.frame-method}
\alias{fit_model,ordered-categorical,allcategorical_missing_data.frame-method}
\alias{fit_model,unordered-categorical,allcategorical_missing_data.frame-method}
\alias{get_parameters,ANY-method}
\alias{get_parameters,polr-method}
\alias{get_parameters,multinom-method}
\alias{get_parameters,mi-method}
\alias{get_parameters,mi_list-method}
\alias{get_parameters,missing_data.frame-method}
\alias{get_parameters,missing_variable-method}
\alias{hist,mi-method}
\alias{hist,missing_variable-method}
\alias{hist,semi-continuous-method}
\alias{hist,binary-method}
\alias{hist,categorical-method}
\alias{initialize,missing_variable-method}
\alias{image,mi-method}
\alias{image,mi_list-method}
\alias{image,mdf_list-method}
\alias{image,missing_data.frame-method}
\alias{image,dgTMatrix-method}
\alias{mi,character,missing-method}
\alias{mi,missing_variable,ANY-method}
\alias{mi,missing_variable,missing-method}
\alias{mi,semi-continuous,missing-method}
\alias{mi,bounded-continuous,missing-method}
\alias{mi,binary,glm-method}
\alias{mi,grouped-binary,clogit-method}
\alias{mi,continuous,glm-method}
\alias{mi,bounded-continuous,glm-method}
\alias{mi,SC_proportion,betareg-method}
\alias{mi,proportion,betareg-method}
\alias{mi,nonnegative-continuous,glm-method}
\alias{mi,censored-continuous,glm-method}
\alias{mi,semi-continuous,glm-method}
\alias{mi,categorical,missing-method}
\alias{mi,count,glm-method}
\alias{mi,data.frame,missing-method}
\alias{mi,irrelevant,ANY-method}
\alias{mi,interval,glm-method}
\alias{mi,mdf_list,missing-method}
\alias{mi,mi_list,missing-method}
\alias{mi,list,missing-method}
\alias{mi,missing_data.frame,mi-method}
\alias{mi,matrix,missing-method}
\alias{mi,mi,missing-method}
\alias{mi,by,missing-method}
\alias{mi,missing_variable,ANY-method}
\alias{mi,missing_variable,missing-method}
\alias{mi,nonnegative-continuous,missing-method}
\alias{mi,ordered-categorical,polr-method}
\alias{mi,proportion,glm-method}
\alias{mi,unordered-categorical,multinom-method}
\alias{mi,unordered-categorical,RNL-method}
\alias{mi,categorical,matrix-method}
\alias{missing_data.frame,data.frame-method}
\alias{missing_data.frame,list-method}
\alias{missing_data.frame,matrix-method}
\alias{missing_variable,ANY,character-method}
\alias{missing_variable,ANY,missing-method}
\alias{betareg-class}
\alias{clogit-class}
\alias{mdf_list-class}
\alias{mi_list-class}
\alias{family-class}
\alias{multinom-class}
\alias{RNL-class}
\alias{plot,missing_data.frame,missing_variable-method}
\alias{plot,mi_list,ANY-method}
\alias{plot,allcategorical_missing_data.frame,binary-method}
\alias{plot,allcategorical_missing_data.frame,categorical-method}
\alias{change_family,missing_data.frame,character,character-method}
\alias{change_family,missing,missing_variable,missing-method}
\alias{change_family,missing_data.frame,character,family-method}
\alias{change_family,missing_data.frame,character,list-method}
\alias{change_family,missing_data.frame,logical,character-method}
\alias{change_family,missing_data.frame,logical,family-method}
\alias{change_family,missing_data.frame,numeric,character-method}
\alias{change_family,missing_data.frame,numeric,family-method}
\alias{change_family,missing_data.frame,numeric,list-method}
\alias{change_imputation_method,missing_data.frame,character,character-method}
\alias{change_size,missing_data.frame,numeric-method}
\alias{change_transformation,missing,missing_variable,function-method}
\alias{change_transformation,missing_data.frame,character,character-method}
\alias{change_transformation,missing_data.frame,numeric,character-method}
\alias{change_transformation,missing_data.frame,logical,character-method}
\alias{change_type,missing_data.frame,character,character-method}
\alias{change_type,missing_data.frame,logical,character-method}
\alias{change_type,missing_data.frame,numeric,character-method}
\alias{change,missing_data.frame,ANY,numeric,numeric-method}
\alias{change,missing_data.frame,ANY,logical,numeric-method}
\alias{change,missing_data.frame,ANY,character,numeric-method}
\alias{change,missing_data.frame,ANY,logical,character-method}
\alias{change,mdf_list,ANY,ANY,ANY-method}
\alias{coerce,data.frame,missing_data.frame-method}
\alias{coerce,matrix,missing_data.frame-method}
\alias{coerce,missing_data.frame,data.frame-method}
\alias{coerce,missing_data.frame,matrix-method}
\alias{complete,missing_data.frame,integer-method}
\alias{complete,mi,numeric-method}
\alias{fit_model,binary,missing_data.frame-method}
\alias{fit_model,grouped-binary,missing_data.frame-method}
\alias{fit_model,count,missing_data.frame-method}
\alias{fit_model,irrelevant,missing_data.frame-method}
\alias{fit_model,interval,missing_data.frame-method}
\alias{fit_model,missing_variable,missing_data.frame-method}
\alias{fit_model,ordered-categorical,missing_data.frame-method}
\alias{fit_model,unordered-categorical,missing_data.frame-method}
\alias{fit_model,character,mi-method}
\alias{fit_model,missing,mi-method}
\alias{fit_model,missing_data.frame,missing_data.frame-method}
\alias{hist,missing_data.frame-method}
\alias{hist,mdf_list-method}
\alias{hist,mi_list-method}
\alias{initialize,missing_data.frame-method}
\alias{plot,missing_data.frame,binary-method}
\alias{plot,missing_data.frame,categorical-method}
\alias{plot,missing_data.frame,semi-continuous-method}
\alias{plot,missing_data.frame,missing_variable-method}
\alias{plot,mi,missing-method}
\alias{traceplot,mi}
\alias{traceplot,mi_list}
\alias{.prune}
\alias{.possible_missing_variable}
\title{Internal Functions and Methods}
\description{
These functions are not intended to be called directly. In the case of
methods, they documented elsewhere, either with the associated generic
function or with the class of the object that the method is defined for.
}
\keyword{internal}
